﻿/* PhotovoltaicPanel.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Weather;

namespace Popolo.Electric
{
  /// <summary>太陽光発電パネル</summary>
  /// <remarks>
  /// JIS8960:太陽光発電用語
  /// JIS8907:太陽光発電システムの発電電力量推定方法
  /// 湯川元信 他:太陽電池モジュール温度上昇の推定, 電気学会論文誌.B, Vol. 116 (1996) No. 9
  /// </remarks>
  [Serializable]
  public class PhotovoltaicPanel : ImmutablePhotovoltaicPanel
  {

    #region 列挙型定義

    /// <summary>材料</summary>
    public enum MaterialType
    {
      /// <summary>アモルファス</summary>
      Amorphous,
      /// <summary>結晶</summary>
      Crystal
    }

    /// <summary>設置タイプ</summary>
    public enum MountType
    {
      /// <summary>屋根置き形</summary>
      RoofMount,
      /// <summary>屋根材一体形</summary>
      RoofIntegrated,
      /// <summary>架台形（ラック形）</summary>
      MountMode
    }

    #endregion

    #region プロパティ

    /// <summary>インバータ効率[-]</summary>
    private double inverterEfficiency = 0.9;

    /// <summary>インバータ効率[-]を設定・取得する</summary>
    public double InverterEfficiency
    {
      get { return inverterEfficiency; }
      set { inverterEfficiency = Math.Max(0, Math.Min(1, value)); }
    }

    /// <summary>最大出力[W]を取得する</summary>
    public double PeakPower { get; private set; }

    /// <summary>設置タイプを取得する</summary>
    public MountType Mount { get; private set; }

    /// <summary>材料を取得する</summary>
    public MaterialType Material { get; private set; }

    /// <summary>傾斜面を取得する</summary>
    public ImmutableIncline Incline { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="peakPower">最大出力[W]</param>
    /// <param name="mount">設置タイプ</param>
    /// <param name="material">材料</param>
    /// <param name="tiltAngle">設置角[degree]（水平面=0度,垂直面=90度）</param>
    /// <remarks>
    /// 最大出力は傾斜面日射強度1000W/m2、周囲温度25Cにおける性能値。
    /// </remarks>
    public PhotovoltaicPanel
      (double peakPower, MountType mount, MaterialType material, double tiltAngle)
      : this(peakPower, mount, material, new Incline(Weather.Incline.Orientation.S, tiltAngle * Math.PI / 180d))
    { }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="peakPower">最大出力[W]</param>
    /// <param name="mount">設置タイプ</param>
    /// <param name="material">材料</param>
    /// <param name="incline">設置傾斜面</param>
    /// <remarks>
    /// 最大出力は傾斜面日射強度1000W/m2、周囲温度25Cにおける性能値。
    /// </remarks>
    public PhotovoltaicPanel
      (double peakPower, MountType mount, MaterialType material, Incline incline)
    {
      this.PeakPower = peakPower;
      this.Mount = mount;
      this.Material = material;
      this.Incline = incline;
    }

    #endregion

    #region メソッド

    /// <summary>出力[W]を計算する</summary>
    /// <param name="drybulbTemperature">外気乾球温度[C]</param>
    /// <param name="velocity">風速[m/s]</param>
    /// <param name="totalIrradiance">
    /// 傾斜面日射強度[W/m2]（直達日射+拡散日射）</param>
    /// <returns>出力[W]</returns>
    public double GetPower(double drybulbTemperature, double velocity, double totalIrradiance)
    {
      return GetPower
        (drybulbTemperature, velocity, totalIrradiance, PeakPower, InverterEfficiency, Mount, Material);
    }

    /// <summary>出力[W]を計算する</summary>
    /// <param name="drybulbTemperature">外気乾球温度[C]</param>
    /// <param name="velocity">風速[m/s]</param>
    /// <param name="sun">太陽</param>
    /// <returns>出力[W]</returns>
    public double GetPower(double drybulbTemperature, double velocity, ImmutableSun sun)
    {
      double totalIrradiance = Incline.GetDirectSolarRadiationRate(sun) * sun.DirectNormalRadiation
        + Incline.ConfigurationFactorToSky * sun.DiffuseHorizontalRadiation;
      return GetPower(drybulbTemperature, velocity, totalIrradiance);
    }

    /// <summary>パネル面の温度上昇による効率変化係数を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="velocity">風速[m/s]</param>
    /// <param name="totalIrradiance">
    /// 傾斜面日射強度[W/m2]（直達日射+拡散日射）</param>
    /// <param name="mount">設置タイプ</param>
    /// <param name="material">材料</param>
    /// <returns>パネル面の温度上昇による効率変化係数</returns>
    private static double getTemperatureRiseCorrectionFactor
      (double drybulbTemperature, double velocity, double totalIrradiance, MountType mount, MaterialType material)
    {
      double a, b;
      switch (mount)
      {
        case MountType.MountMode:
          a = 46;
          b = 0.41;
          break;
        case MountType.RoofIntegrated:
          a = 50;
          b = 0.33;
          break;
        default:
          a = 50;
          b = 0.38;
          break;
      }
      //パネル温度の計算
      double tp = drybulbTemperature + (a / (b * Math.Pow(velocity, 0.8) + 1) + 2) * totalIrradiance / 1000d - 2d;

      if (material == MaterialType.Amorphous) return 1 - 0.002 * (tp - 25d);
      else return 1 - 0.004 * (tp - 25d);
    }

    /// <summary>出力[kW]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="velocity">風速[m/s]</param>
    /// <param name="totalIrradiance">傾斜面日射強度[W/m2]（直達日射+拡散日射）</param>
    /// <param name="peakPower">最大出力[kW]</param>
    /// <param name="inverterEfficiency">インバータ効率[-]</param>
    /// <param name="mount">設置タイプ</param>
    /// <param name="material">材料</param>
    /// <returns>出力[kW]</returns>
    public static double GetPower
      (double drybulbTemperature, double velocity, double totalIrradiance,
      double peakPower, double inverterEfficiency, MountType mount, MaterialType material)
    {
      double kpt = getTemperatureRiseCorrectionFactor
        (drybulbTemperature, velocity, totalIrradiance, mount, material);
      return totalIrradiance / 1000d * kpt * peakPower * inverterEfficiency;
    }

    #endregion    

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用の太陽光発電パネル</summary>
  public interface ImmutablePhotovoltaicPanel
  {

    /// <summary>インバータ効率[-]を設定・取得する</summary>
    double InverterEfficiency { get; }

    /// <summary>最大出力[W]を取得する</summary>
    double PeakPower { get; }

    /// <summary>設置タイプを取得する</summary>
    PhotovoltaicPanel.MountType Mount { get; }

    /// <summary>材料を取得する</summary>
    PhotovoltaicPanel.MaterialType Material { get; }

    /// <summary>傾斜面を取得する</summary>
    ImmutableIncline Incline { get; }

  }

  #endregion

}
