﻿/* AirHandlingUnit.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;
using Popolo.HVAC.HeatExchanger;
using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.AirConditioner
{
  /// <summary>空気調和機</summary>
  [Serializable]
  public class AirHandlingUnit : ImmutableAirHandlingUnit
  {

    #region 定数宣言

    /// <summary>湿り空気の定圧比熱[J/(kgK)]</summary>
    /// <remarks>24C, 50%, 9.5g/kg</remarks>
    private const double AIR_SPECIFICHEAT = 1005 + 1846 * 0.0095;

    /// <summary>湿り空気の密度[kg/m3]</summary>
    private const double AIR_DENSITY = 1.2;

    /// <summary>大気圧[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    #endregion

    #region 列挙型定義

    /// <summary>外気導入制御方式</summary>
    public enum OutdoorAirCoolingControl
    {
      /// <summary>無し</summary>
      None,
      /// <summary>乾球温度基準</summary>
      DrybulbTemperature,
      /// <summary>絶対湿度基準</summary>
      HumidityRatio,
      /// <summary>比エンタルピー基準</summary>
      Enthalpy
    }

    /// <summary>加湿方式</summary>
    public enum HumidifierType
    {
      /// <summary>無し</summary>
      None,
      /// <summary>蒸気加湿</summary>
      Steam,
      /// <summary>水滴下気化式</summary>
      DropPervaporation,
      /// <summary>水噴霧式</summary>
      Atomizing,
      /// <summary>超音波式</summary>
      UltraSonic,
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>冷却コイル</summary>
    private CrossFinHeatExchanger cCoil;

    /// <summary>加熱コイル</summary>
    private CrossFinHeatExchanger hCoil;

    /// <summary>給気ファン</summary>
    private CentrifugalFan saFan;

    /// <summary>還気ファン</summary>
    private CentrifugalFan raFan;

    /// <summary>全熱交換器</summary>
    private RotaryRegenerator regen = null;

    /// <summary>外気冷房制御方式を設定・取得する</summary>
    public OutdoorAirCoolingControl OutdoorAirCooling
    { get; set; } = OutdoorAirCoolingControl.None;

    /// <summary>加湿方式を取得する</summary>
    public HumidifierType Humidifier { get; private set; }

    /// <summary>冷水コイルを取得する</summary>
    public ImmutableCrossFinHeatExchanger CoolingCoil { get { return cCoil; } }

    /// <summary>温水コイルを取得する</summary>
    public ImmutableCrossFinHeatExchanger HeatingCoil { get { return hCoil; } }

    /// <summary>給気ファンを取得する</summary>
    public ImmutableFluidMachinery SupplyAirFan { get { return saFan; } }

    /// <summary>還気ファンを取得する</summary>
    public ImmutableFluidMachinery ReturnAirFan { get { return raFan; } }

    /// <summary>全熱交換器を取得する</summary>
    public ImmutableRotaryRegenerator Regenerator { get { return regen; } }

    /// <summary>ダクト熱損失率[-]を設定・取得する</summary>
    public double DuctHeatLossRate { get; set; } = 0.08;

    /// <summary>全熱交換器をバイパスさせるか否かを設定・取得する</summary>
    public bool BypassRegenerator { get; set; }

    /// <summary>最大外気量[kg/s]を取得する</summary>
    public double MaxOAFlowRate { get; private set; }

    /// <summary>最小外気量[kg/s]を取得する</summary>
    public double MinOAFlowRate { get; private set; }

    /// <summary>外気量[kg/s]を取得する</summary>
    public double OAFlowRate { get; private set; }

    /// <summary>還気量[kg/s]を取得する</summary>
    public double RAFlowRate { get; private set; }

    /// <summary>給気量[kg/s]を取得する</summary>
    public double SAFlowRate { get; private set; }

    /// <summary>排気量[kg/s]を取得する</summary>
    public double EAFlowRate { get; private set; }

    /// <summary>還気乾球温度[C]を設定・取得する</summary>
    public double RATemperature { get; set; }

    /// <summary>還気絶対湿度[kg/kg]を設定・取得する</summary>
    public double RAHumidityRatio { get; set; }

    /// <summary>外気乾球温度[C]を設定・取得する</summary>
    public double OATemperature { get; set; }

    /// <summary>外気絶対湿度[kg/kg]を設定・取得する</summary>
    public double OAHumidityRatio { get; set; }

    /// <summary>給気乾球温度[C]を取得する</summary>
    public double SATemperature { get; private set; }

    /// <summary>給気絶対湿度[kg/kg]を取得する</summary>
    public double SAHumidityRatio { get; private set; }

    /// <summary>冷水入口温度[C]を設定・取得する</summary>
    public double ChilledWaterInletTemperature { get; set; }

    /// <summary>温水入口温度[C]を設定・取得する</summary>
    public double HotWaterInletTemperature { get; set; }

    /// <summary>加湿用給水量[kg/s]を取得する</summary>
    public double WaterConsumption { get; private set; }

    /// <summary>加湿用蒸気消費量[kg/s]を取得する</summary>
    public double SteamConsumption { get; private set; }

    /// <summary>加湿水の給水効率[-]を設定・取得する</summary>
    public double WaterSupplyCoefficient { get; set; }

    /// <summary>加湿器の最大飽和効率[-]を設定・取得する</summary>
    public double MaxSaturationEfficiency { get; set; }

    /// <summary>加湿器の飽和効率[-]を設定・取得する</summary>
    public double SaturationEfficiency { get; set; }

    /// <summary>過加熱あるいは過冷却が発生してでも風量を最小化するか否か</summary>
    public bool MinimizeAirFlow { get; set; } = true;

    /// <summary>吹き出し上限温度[C]を設定・取得する</summary>
    public double UpperTemperatureLimit { get; set; } = 37;

    /// <summary>吹き出し下限温度[C]を設定・取得する</summary>
    public double LowerTemperatureLimit { get; set; } = 15;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="cCoil">冷却除湿コイル</param>
    /// <param name="hCoil">加熱コイル</param>
    /// <param name="humidType">加湿器のタイプ</param>
    /// <param name="saFan">給気ファン</param>
    /// <param name="raFan">還気ファン</param>
    public AirHandlingUnit
      (CrossFinHeatExchanger cCoil, CrossFinHeatExchanger hCoil,
      HumidifierType humidType, CentrifugalFan saFan, CentrifugalFan raFan) :
      this(cCoil, hCoil, humidType, saFan, raFan, null)
    { }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="cCoil">冷却除湿コイル</param>
    /// <param name="hCoil">加熱コイル</param>
    /// <param name="humidType">加湿器のタイプ</param>
    /// <param name="saFan">給気ファン</param>
    /// <param name="raFan">還気ファン</param>
    /// <param name="regenerator">全熱交換器</param>
    public AirHandlingUnit
      (CrossFinHeatExchanger cCoil, CrossFinHeatExchanger hCoil, HumidifierType humidType,
      CentrifugalFan saFan, CentrifugalFan raFan, RotaryRegenerator regenerator)
    {
      this.cCoil = cCoil;
      this.hCoil = hCoil;
      this.Humidifier = humidType;
      this.saFan = saFan;
      this.raFan = raFan;
      this.regen = regenerator;

      switch (humidType)
      {
        case HumidifierType.DropPervaporation:
          WaterSupplyCoefficient = 0.5;
          MaxSaturationEfficiency = 0.8;
          WaterSupplyCoefficient = 0.5;
          break;
        case HumidifierType.Steam:
          WaterSupplyCoefficient = 0.9;
          MaxSaturationEfficiency = 1.0;
          WaterSupplyCoefficient = 0.9;
          break;
        case HumidifierType.UltraSonic:
          WaterSupplyCoefficient = 0.9;
          MaxSaturationEfficiency = 0.5;
          WaterSupplyCoefficient = 0.9;
          break;
        case HumidifierType.Atomizing:
          WaterSupplyCoefficient = 0.4;
          MaxSaturationEfficiency = 0.3;
          WaterSupplyCoefficient = 0.4;
          break;
      }
    }

    #endregion

    #region 冷却運転

    /// <summary>空気を冷却する（成り行き計算）</summary>
    public void CoolAir() { coolAir(false, 0, 0); }
    
    /// <summary>空気を冷却する（吹出温度制御）</summary>
    /// <param name="setpointTemperature">給気温度設定値[C]</param>
    /// <param name="setpointHumidity">給気絶対湿度設定値[kg/kg]</param>
    /// <returns>制御成功の真偽</returns>
    /// <remarks>絶対湿度設定値は外気冷房制御用。温度基準の場合は不要。</remarks>
    public bool CoolAir(double setpointTemperature, double setpointHumidity)
    { return coolAir(true, setpointTemperature, setpointHumidity); }

    /// <summary>空気を冷却除湿する</summary>
    /// <param name="controlOutletTemp">AHU出口温度を制御するか否か</param>
    /// <param name="spTemp">給気温度設定値[C]</param>
    /// <param name="spHumid">給気絶対湿度設定値[kg/kg]</param>
    /// <returns>制御成功の真偽</returns>
    /// <remarks>絶対湿度設定値は外気冷房制御用。温度基準の場合は不要。</remarks>
    private bool coolAir(bool controlOutletTemp, double spTemp, double spHumid)
    {
      //SA風量が0の場合は停止処理（全外気でRA風量=0はありえる）
      if (SAFlowRate <= 0)
      {
        ShutOff();
        return false;
      }      

      //ファン昇温を計算
      saFan.UpdateState(SAFlowRate / AIR_DENSITY);
      raFan.UpdateState(RAFlowRate / AIR_DENSITY);
      double tRise = saFan.GetElectricConsumption() / (SAFlowRate * AIR_SPECIFICHEAT);

      //ダクト熱損失とファン昇温を考慮して出口温度設定を補正
      double sp2 = Math.Max(LowerTemperatureLimit, spTemp);
      double tdCo = (sp2 - RATemperature * DuctHeatLossRate) / (1 - DuctHeatLossRate) - tRise;

      //外気冷房を反映した外気量調整//成り行き計算の場合は外気量は所与
      if (controlOutletTemp)
      {
        double mOA = MinOAFlowRate;
        if (OutdoorAirCooling == OutdoorAirCoolingControl.DrybulbTemperature)
        {
          if (tdCo < OATemperature && OATemperature < RATemperature)
            mOA = Math.Min(MaxOAFlowRate, SAFlowRate);
          else mOA = RAFlowRate * (RATemperature - tdCo) / (tdCo - OATemperature);
        }
        else if (OutdoorAirCooling == OutdoorAirCoolingControl.HumidityRatio)
        {
          if (spHumid < OAHumidityRatio && OAHumidityRatio < RAHumidityRatio)
            mOA = Math.Min(MaxOAFlowRate, SAFlowRate);
          else mOA = RAFlowRate * (RAHumidityRatio - spHumid) / (spHumid - OAHumidityRatio);
        }
        else if (OutdoorAirCooling == OutdoorAirCoolingControl.Enthalpy)
        {
          double hAHUi =
            MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(RATemperature, RAHumidityRatio);
          double hAHUo =
            MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(tdCo, spHumid);
          double hOA =
            MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(OATemperature, OAHumidityRatio);
          if (hAHUo < hOA && hOA < hAHUi) mOA = Math.Min(MaxOAFlowRate, SAFlowRate);
          else mOA = RAFlowRate * (hAHUi - hAHUo) / (hAHUo - hOA);
        } 
        OAFlowRate = Math.Max(MinOAFlowRate, Math.Min(MaxOAFlowRate, mOA));
        EAFlowRate = RAFlowRate + OAFlowRate - SAFlowRate;
      }

      //全熱交換器による熱回収
      double tdOA = OATemperature;
      double hrOA = OAHumidityRatio;
      if (regen != null && !BypassRegenerator && 0 < OAFlowRate && 0 < EAFlowRate)
      {
        double hRtn = 
          MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(RATemperature, RAHumidityRatio);
        double hOA =
          MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(OATemperature, OAHumidityRatio);
        if (hRtn < hOA)
        {
          const double cf = 3600 / AIR_DENSITY;
          //成り行き計算を試行
          regen.UpdateState(OAFlowRate * cf, EAFlowRate * cf, 1.0, tdOA, hrOA, RATemperature, RAHumidityRatio);
          //過剰処理の場合には給気温度を制御
          if (controlOutletTemp && regen.SupplyAirOutletDrybulbTemperature < tdCo)
            regen.ControlOutletTemperature
              (OAFlowRate * cf, EAFlowRate * cf, 1.0, tdOA, hrOA, RATemperature, RAHumidityRatio, tdCo);
          tdOA = regen.SupplyAirOutletDrybulbTemperature;
          hrOA = regen.SupplyAirOutletHumidityRatio;
        }
        else regen.ShutOff();
      }

      //OAとRAを混合して冷却
      double mr = OAFlowRate / (OAFlowRate + RAFlowRate);
      double tdCi = tdOA * mr + RATemperature * (1 - mr);
      double hrCi = hrOA * mr + RAHumidityRatio * (1 - mr);
      if (controlOutletTemp)
        cCoil.ControlOutletAirTemperature(tdCi, hrCi, ChilledWaterInletTemperature, SAFlowRate, tdCo);
      else
        cCoil.UpdateOutletState(tdCi, hrCi, ChilledWaterInletTemperature, SAFlowRate, cCoil.WaterFlowRate);
      hCoil.ShutOff();
      WaterConsumption = SteamConsumption = 0.0;

      //SAファンによる昇温とダクト熱損失効果を反映
      SATemperature = cCoil.OutletAirTemperature + tRise;
      SAHumidityRatio = cCoil.OutletAirHumidityRatio;
      computeDuctLoss();

      return -1e-4 < spTemp - SATemperature;
    }

    /// <summary>外気冷房による処理熱[kW]を計算する</summary>
    /// <returns>外気冷房による処理熱[kW]</returns>
    public double GetOutdoorCoolingHeat()
    {
      double hSA = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (cCoil.InletAirTemperature, cCoil.InletAirHumidityRatio);
      double hOA = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (OATemperature, OAHumidityRatio);
      return (OAFlowRate - MinOAFlowRate) * (hSA - hOA);
    }

    /// <summary>吹き出し空気にダクトロスを反映する</summary>
    private void computeDuctLoss()
    {
      double dl1 = 1 - DuctHeatLossRate;
      SATemperature = SATemperature * dl1 + RATemperature * DuctHeatLossRate;
      SAHumidityRatio = SAHumidityRatio * dl1 + RAHumidityRatio * DuctHeatLossRate;
    }

    #endregion

    #region 加熱運転

    /// <summary>空気を加熱加湿する（成り行き計算）</summary>
    public void HeatAir() { heatAir(false, 0, 0); } //これだと蒸気加湿の場合に全く加湿されない!!! DEBUG

    /// <summary>空気を加熱加湿する（吹出温湿度制御）</summary>
    /// <param name="setpointTemperature">出口空気温度設定値[C]</param>
    /// <param name="setpointHumidity">出口空気湿度設定値[kg/kg]</param>
    /// <returns>加熱成功の真偽</returns>
    public bool HeatAir(double setpointTemperature, double setpointHumidity)
    { return heatAir(true, setpointTemperature, setpointHumidity); }

    /// <summary>空気を加熱加湿する</summary>
    /// <param name="controlOutletState">給気温度を制御するか</param>
    /// <param name="spTemp">給気温度設定値</param>
    /// <param name="spHumid">給気絶対湿度設定値</param>
    /// <returns>制御成功の真偽</returns>
    private bool heatAir(bool controlOutletState, double spTemp, double spHumid)
    {
      //SA風量が0の場合は停止処理（全外気でRA風量=0はありえる）
      if (SAFlowRate <= 0)
      {
        ShutOff();
        return false;
      }

      //ファン昇温を計算
      saFan.UpdateState(SAFlowRate / AIR_DENSITY);
      raFan.UpdateState(RAFlowRate / AIR_DENSITY);
      double tRise = saFan.GetElectricConsumption() / (SAFlowRate * AIR_SPECIFICHEAT);

      //ダクト熱損失とファン昇温を考慮して出口温湿度設定を補正
      double sp2 = Math.Min(UpperTemperatureLimit, spTemp);
      double tdCo = (sp2 - RATemperature * DuctHeatLossRate) / (1 - DuctHeatLossRate) - tRise;
      double wAHUo = (spHumid - RAHumidityRatio * DuctHeatLossRate) / (1 - DuctHeatLossRate);
      double hCo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(tdCo, wAHUo);

      //外気量は最小値
      double mOA = MinOAFlowRate;
      OAFlowRate = MinOAFlowRate;
      EAFlowRate = RAFlowRate + OAFlowRate - SAFlowRate;

      //全熱交換器による熱回収
      double tdOA = OATemperature;
      double hrOA = OAHumidityRatio;
      if (regen != null && !BypassRegenerator && 0 < OAFlowRate && 0 < EAFlowRate)
      {
        double hRtn =
          MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(RATemperature, RAHumidityRatio);
        double hOA =
          MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(OATemperature, OAHumidityRatio);
        if (hOA < hRtn)
        {
          const double cf = 3600 / AIR_DENSITY;
          //成り行き計算実行
          regen.UpdateState(OAFlowRate * cf, EAFlowRate * cf, 1.0, tdOA, hrOA, RATemperature, RAHumidityRatio);
          //過剰処理の場合には給気比エンタルピーを制御
          double hRego = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
            (regen.SupplyAirOutletDrybulbTemperature, regen.SupplyAirOutletHumidityRatio);
          if (controlOutletState && hCo < hRego)
            regen.ControlOutletEnthalpy
              (OAFlowRate * cf, EAFlowRate * cf, 1.0, tdOA, hrOA, RATemperature, RAHumidityRatio, hCo);
          tdOA = regen.SupplyAirOutletDrybulbTemperature;
          hrOA = regen.SupplyAirOutletHumidityRatio;
        }
        else regen.ShutOff();
      }

      //OAとRAを混合
      double mr = OAFlowRate / (OAFlowRate + RAFlowRate);
      double tdCi = tdOA * mr + RATemperature * (1 - mr);
      double hrCi = hrOA * mr + RAHumidityRatio * (1 - mr);

      //水加湿の場合には温水コイル出口温度を比エンタルピーで調整
      if (hrCi < wAHUo &&
        (Humidifier != HumidifierType.None && Humidifier != HumidifierType.Steam))
        tdCo = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndEnthalpy(hrCi, hCo);

      //加熱
      if (controlOutletState)
        hCoil.ControlOutletAirTemperature(tdCi, hrCi, HotWaterInletTemperature, SAFlowRate, tdCo);
      else
        hCoil.UpdateOutletState(tdCi, hrCi, HotWaterInletTemperature, SAFlowRate, hCoil.WaterFlowRate);
      tdCo = hCoil.OutletAirTemperature;
      cCoil.ShutOff();

      //加湿
      double tSFi, wSFi;
      tSFi = hCoil.OutletAirTemperature;
      wSFi = hCoil.OutletAirHumidityRatio;
      WaterConsumption = SteamConsumption = 0.0;
      //水加湿
      if (Humidifier != HumidifierType.None && Humidifier != HumidifierType.Steam)
      {
        double hSFi =
          MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(tdCo, hCoil.OutletAirHumidityRatio);
        double satW = MoistAir.GetSaturationHumidityRatioFromEnthalpy(hSFi, ATMOSPHERIC_PRESSURE);
        if (controlOutletState)
        {
          if (hCoil.OutletAirHumidityRatio < wAHUo)
          {
            double maxW = (1 - MaxSaturationEfficiency)
            * hCoil.OutletAirHumidityRatio + MaxSaturationEfficiency * satW;
            wAHUo = Math.Min(maxW, wAHUo);
            SaturationEfficiency = (wAHUo - hCoil.OutletAirHumidityRatio) / (satW - hCoil.OutletAirHumidityRatio);
          }
          else
          {
            wAHUo = hCoil.OutletAirHumidityRatio;
            SaturationEfficiency = 0.0;
          }
        }
        else wAHUo = (1 - SaturationEfficiency) * hCoil.OutletAirHumidityRatio + SaturationEfficiency * satW;
        tSFi = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndEnthalpy(wAHUo, hSFi);
        WaterConsumption = (wAHUo - hCoil.OutletAirHumidityRatio) * SAFlowRate / WaterSupplyCoefficient;
      }
      //蒸気加湿//最大飽和効率=100%
      else if (Humidifier == HumidifierType.Steam)
      {
        double satW = MoistAir.GetSaturationHumidityRatioFromDryBulbTemperature
          (hCoil.OutletAirTemperature, ATMOSPHERIC_PRESSURE);
        wAHUo = Math.Min(satW, wAHUo);
        SteamConsumption = (wAHUo - hCoil.OutletAirHumidityRatio) * SAFlowRate / WaterSupplyCoefficient;
      }

      //SAファンによる昇温とダクト熱損失効果を反映
      SATemperature = tSFi + tRise;
      SAHumidityRatio = wAHUo;
      computeDuctLoss();

      return -1e-4 < SATemperature - spTemp;
    }

    #endregion

    #region 換気運転

    /// <summary>換気運転を行う</summary>
    public void Ventilate()
    {
      //SA風量が0の場合は停止処理（全外気でRA風量=0はありえる）
      if (SAFlowRate <= 0)
      {
        ShutOff();
        return;
      }

      //ファン昇温とダクト熱損失の計算
      saFan.UpdateState(SAFlowRate / AIR_DENSITY);
      raFan.UpdateState(SAFlowRate / AIR_DENSITY);
      double tRise = saFan.GetElectricConsumption() / (SAFlowRate * AIR_SPECIFICHEAT);
      SATemperature = RATemperature + tRise * (1 - DuctHeatLossRate);
      SAHumidityRatio = RAHumidityRatio;
    }

    #endregion

    #region その他インスタンスメソッド

    /// <summary>風量[kg/s]を設定する</summary>
    /// <param name="raFlowRate">還気量[kg/s]</param>
    /// <param name="saFlowRate">給気量[kg/s]</param>
    public void SetAirFlowRate(double raFlowRate, double saFlowRate)
    {
      RAFlowRate = Math.Max(0, raFlowRate);
      SAFlowRate = Math.Max(0, saFlowRate);
      EAFlowRate = RAFlowRate + OAFlowRate - SAFlowRate;
    }

    /// <summary>外気量範囲を設定する</summary>
    /// <param name="minOAFlow">最小外気量[kg/s]</param>
    /// <param name="maxOAFlow">最大外気量[kg/s]</param>
    public void SetOutdoorAirFlowRange(double minOAFlow, double maxOAFlow)
    {
      MinOAFlowRate = minOAFlow;
      MaxOAFlowRate = maxOAFlow;
      OAFlowRate = Math.Min(Math.Max(OAFlowRate, MinOAFlowRate), MaxOAFlowRate);
      EAFlowRate = RAFlowRate + OAFlowRate - SAFlowRate;
    }    

    /// <summary>機器を停止させる</summary>
    public void ShutOff()
    {
      cCoil.ShutOff();
      hCoil.ShutOff();
      saFan.ShutOff();
      raFan.ShutOff();
      if (regen != null) regen.ShutOff();
      WaterConsumption = 0.0;
      OAFlowRate = 0;
      SetAirFlowRate(0, 0);
    }
    
    /// <summary>VAV風量を計算する</summary>
    /// <param name="isCooling">冷房運転か否か</param>
    /// <param name="supplyHumiditySP">給気湿度設定値[kg/kg]（暖房時に使用）</param>
    /// <param name="isVAVShutOff">VAV停止状態</param>
    /// <param name="zoneTemps">ゾーン温度[C]</param>
    /// <param name="zoneHumids">ゾーン湿度[kg/kg]</param>
    /// <param name="zoneSLoads">ゾーン顕熱負荷[kW]（暖房が正）</param>
    /// <param name="minSAFlow">最小給気量[kg/s]</param>
    /// <param name="maxSAFlow">最大給気量[kg/s]</param>
    /// <param name="maxRAFlow">最大還気量[kg/s]</param>
    /// <param name="success">制御成功の真偽</param>
    /// <returns>VAV風量リスト[kg/s]</returns>
    public double[] OptimizeVAV
      (bool isCooling, double supplyHumiditySP, bool[] isVAVShutOff, double[] zoneTemps, double[] zoneHumids, 
      double[] zoneSLoads, double[] minSAFlow, double[] maxSAFlow, double[] maxRAFlow, out bool success)
    {
      //AHU吹出温度の上下限値を計算
      double tAHUoMin, tAHUoMax;
      if (isCooling)
      {
        if (MinimizeAirFlow) tAHUoMin = 60;
        else tAHUoMin = -10;
        tAHUoMax = 60;
      }
      else
      {
        if (MinimizeAirFlow) tAHUoMin = -10;
        else tAHUoMin = 60;
        tAHUoMax = -10;
      }
      for (int i = 0; i < zoneTemps.Length; i++)
      {
        if (!isVAVShutOff[i])
        {
          //最大風量での給気温度とAHU吹出温度
          double tMax = zoneTemps[i] + zoneSLoads[i] / (AIR_SPECIFICHEAT * maxSAFlow[i]);
          if (isCooling) tAHUoMax = Math.Min(tAHUoMax, tMax);
          else tAHUoMax = Math.Max(tAHUoMax, tMax);
          //最小風量での給気温度とAHU吹出温度          
          if (0 < minSAFlow[i])
          {
            double tMin = zoneTemps[i] + zoneSLoads[i] / (AIR_SPECIFICHEAT * minSAFlow[i]);
            //過冷却と過加熱が生じても風量を最小化する
            if (MinimizeAirFlow)
            {
              if (isCooling) tAHUoMin = Math.Min(tAHUoMin, tMin);
              else tAHUoMin = Math.Max(tAHUoMin, tMin);
            }
            //過冷却と過加熱を発生させないように風量を絞る
            else
            {
              if (isCooling) tAHUoMin = Math.Max(tAHUoMin, tMin);
              else tAHUoMin = Math.Min(tAHUoMin, tMin);
            }
          }
        }
      }
      if (isCooling)
      {
        tAHUoMin = Math.Max(LowerTemperatureLimit, tAHUoMin);
        tAHUoMax = Math.Max(LowerTemperatureLimit, tAHUoMax);
      }
      else
      {
        tAHUoMin = Math.Min(UpperTemperatureLimit, tAHUoMin);
        tAHUoMax = Math.Min(UpperTemperatureLimit, tAHUoMax);
      }

      //評価関数を定義
      double mSAMax = 0;
      for (int i = 0; i < maxSAFlow.Length; i++) mSAMax += maxSAFlow[i];
      double[] mSAn = new double[zoneTemps.Length];
      bool suc = true;
      bool overLoad = true;
      Minimization.MinimizeFunction mFnc = delegate (double tAHUo)
      {
        //還気条件を計算
        double mSAsum = 0;
        double mRAsum = 0;
        RATemperature = 0;
        RAHumidityRatio = 0;
        suc = true;
        for (int i = 0; i < zoneTemps.Length; i++)
        {
          if (!isVAVShutOff[i])
          {
            if (tAHUo == zoneTemps[i]) mSAn[i] = 0;
            else mSAn[i] = zoneSLoads[i] / (AIR_SPECIFICHEAT * (tAHUo - zoneTemps[i]));  //DEBUG
            if (1e-10 < minSAFlow[i] - mSAn[i] || 1e-10 < mSAn[i] - maxSAFlow[i]) suc = false;
            mSAn[i] = Math.Min(maxSAFlow[i], Math.Max(minSAFlow[i], mSAn[i]));
            double mRAn = mSAn[i] - (maxSAFlow[i] - maxRAFlow[i]);
            mSAsum += mSAn[i];
            mRAsum += mRAn;
            RATemperature += mRAn * zoneTemps[i];
            RAHumidityRatio += mRAn * zoneHumids[i];
          }
          else mSAn[i] = 0;
        }
        if (1e-7 < mRAsum)
        {
          RATemperature /= mRAsum;
          RAHumidityRatio /= mRAsum;
        }
        SetAirFlowRate(mRAsum, mSAsum);

        //誤差を評価
        if (isCooling) overLoad = !CoolAir(tAHUo, supplyHumiditySP);
        else
        {
          //必要な湿度を補正
          double bf = SAFlowRate / mSAMax;
          double hsp = RAHumidityRatio * (1 - bf) + supplyHumiditySP * bf;
          overLoad = !HeatAir(tAHUo, hsp);
        }
        if (overLoad)
        {
          double err = Math.Abs(SATemperature - tAHUo);
          if (isCooling) return RATemperature + err;
          else return -RATemperature + err;
        }
        else
        {
          if (isCooling) return SATemperature;
          else return -SATemperature;
        }
      };

      //最大風量で計算
      mFnc(tAHUoMax);
      //冷暖逆転の場合には最小風量
      if ((isCooling && RATemperature < tAHUoMax) || (!isCooling && tAHUoMax < RATemperature)) mFnc(tAHUoMin);
      //最大風量で過負荷の場合には最大風量
      else if (!overLoad)
      {
        //最小風量でAHUが過負荷になるならば処理可能な風量まで修正
        mFnc(tAHUoMin);
        if (overLoad)
        {
          if (isCooling) Minimization.GoldenSection(ref tAHUoMin, tAHUoMax, mFnc);
          else Minimization.GoldenSection(ref tAHUoMax, tAHUoMin, mFnc);
        }
      }
      success = suc;
      return mSAn;
    }

    #endregion

  }

  #region 読み取り専用AHU

  /// <summary>読み取り専用空気調和機</summary>
  public interface ImmutableAirHandlingUnit
  {
    /// <summary>外気冷房制御方式を取得する</summary>
    AirHandlingUnit.OutdoorAirCoolingControl OutdoorAirCooling { get; }

    /// <summary>加湿方式を取得する</summary>
    AirHandlingUnit.HumidifierType Humidifier { get; }

    /// <summary>冷水コイルを取得する</summary>
    ImmutableCrossFinHeatExchanger CoolingCoil { get; }

    /// <summary>温水コイルを取得する</summary>
    ImmutableCrossFinHeatExchanger HeatingCoil { get; }

    /// <summary>給気ファンを取得する</summary>
    ImmutableFluidMachinery SupplyAirFan { get; }

    /// <summary>還気ファンを取得する</summary>
    ImmutableFluidMachinery ReturnAirFan { get; }

    /// <summary>全熱交換器を取得する</summary>
    ImmutableRotaryRegenerator Regenerator { get; }

    /// <summary>ダクト熱損失率[-]を取得する</summary>
    double DuctHeatLossRate { get; } 

    /// <summary>全熱交換器をバイパスさせるか否かを取得する</summary>
    bool BypassRegenerator { get; }

    /// <summary>最大外気量[kg/s]を取得する</summary>
    double MaxOAFlowRate { get; }

    /// <summary>最小外気量[kg/s]を取得する</summary>
    double MinOAFlowRate { get; }

    /// <summary>外気量[kg/s]を取得する</summary>
    double OAFlowRate { get; }

    /// <summary>還気量[kg/s]を取得する</summary>
    double RAFlowRate { get; }

    /// <summary>給気量[kg/s]を取得する</summary>
    double SAFlowRate { get; }

    /// <summary>排気量[kg/s]を取得する</summary>
    double EAFlowRate { get; }

    /// <summary>還気乾球温度[C]を取得する</summary>
    double RATemperature { get; }

    /// <summary>還気絶対湿度[kg/kg]を取得する</summary>
    double RAHumidityRatio { get; }

    /// <summary>外気乾球温度[C]を取得する</summary>
    double OATemperature { get; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    double OAHumidityRatio { get; }

    /// <summary>給気乾球温度[C]を取得する</summary>
    double SATemperature { get; }

    /// <summary>給気絶対湿度[kg/kg]を取得する</summary>
    double SAHumidityRatio { get; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    double ChilledWaterInletTemperature { get; }

    /// <summary>温水入口温度[C]を取得する</summary>
    double HotWaterInletTemperature { get; }

    /// <summary>加湿用給水量[kg/s]を取得する</summary>
    double WaterConsumption { get; }

    /// <summary>加湿用蒸気消費量[kg/s]を取得する</summary>
    double SteamConsumption { get; }

    /// <summary>加湿水の給水効率[-]を取得する</summary>
    double WaterSupplyCoefficient { get; }

    /// <summary>加湿器の最大飽和効率[-]を取得する</summary>
    double MaxSaturationEfficiency { get; }

    /// <summary>加湿器の飽和効率[-]を取得する</summary>
    double SaturationEfficiency { get; }

    /// <summary>過加熱あるいは過冷却が発生してでも風量を最小化するか否か</summary>
    bool MinimizeAirFlow { get; }

    /// <summary>吹き出し上限温度[C]を取得する</summary>
    double UpperTemperatureLimit { get; }

    /// <summary>吹き出し下限温度[C]を取得する</summary>
    double LowerTemperatureLimit { get; }
  }

  #endregion

}
