﻿/* CentrifugalFan.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit
{
  /// <summary>遠心ファン</summary>
  [Serializable]
  public class CentrifugalFan : FluidMachinery
  {

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nomPressure">定格（最大効率点）全圧[kPa]</param>
    /// <param name="nomFlowRate">定格（最大効率点）風量[m3/s]</param>
    /// <param name="designPressure">設計全圧[kPa]</param>
    /// <param name="designFlowRate">設計風量[m3/s]</param>
    /// <param name="number">番手[-]</param>
    /// <param name="hasInverter">インバータ機か否か</param>
    public CentrifugalFan(double nomPressure, double nomFlowRate, double designPressure,
      double designFlowRate, double number, bool hasInverter) :
      base(nomPressure, nomFlowRate, designPressure, designFlowRate, 0.02, hasInverter)
    {
      //効率・圧力特性係数を計算
      efficiencyCoefficient = new double[4];
      pressureCoefficient = new double[3];
      getGeneralParameters
        (nomFlowRate, nomPressure, number, ref efficiencyCoefficient, ref pressureCoefficient);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nomPressure">定格全圧[kPa]</param>
    /// <param name="nomFlowRate">定格風量[m3/s]</param>
    /// <param name="designPressure">設計全圧[kPa]</param>
    /// <param name="designFlowRate">設計風量[m3/s]</param>
    /// <param name="number">番手[-]</param>
    /// <param name="dynamicPressure">設計動圧[kPa]</param>
    /// <param name="hasInverter">インバータ機か否か</param>
    public CentrifugalFan(double nomPressure, double nomFlowRate, double designPressure,
      double designFlowRate, double number, double dynamicPressure, bool hasInverter) :
      this(nomPressure, nomFlowRate, designPressure, designFlowRate, dynamicPressure, hasInverter)
    { }

    #endregion

    #region インスタンスメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    public void UpdateState(double flowRate)
    {
      VolumetricFlowRate = flowRate;

      //停止判定
      if (VolumetricFlowRate <= 0)
      {
        ShutOff();
        return;
      }

      //抵抗曲線上の圧力
      double ps = VolumetricFlowRate * VolumetricFlowRate * resistanceCoefficient + ActualHead;

      //インバータ有り
      if (HasInverter)
      {
        updateWithFlowRateAndPressure(VolumetricFlowRate, ps);
        //回転数上限の場合には最大回転数で成り行き計算
        if (1.0 < RotationRatio)
        {
          RotationRatio = 1.0;
          VolumetricFlowRate =
            GetFlowRate(RotationRatio, pressureCoefficient, resistanceCoefficient, ActualHead, 1, DesignFlowRate);
          Pressure = VolumetricFlowRate * VolumetricFlowRate * resistanceCoefficient + ActualHead;
        }
        //回転数下限の場合には最小回転数でダンパ制御
        else if (RotationRatio < MinimumRotationRatio)
          updateWithFlowRateAndRotationRatio(VolumetricFlowRate, MinimumRotationRatio);
      }
      //インバータ無し
      else
      {
        //過負荷判定
        updateWithFlowRateAndRotationRatio(VolumetricFlowRate, 1.0);
        if (Pressure < ps)
        {
          RotationRatio = 1.0;
          VolumetricFlowRate =
            GetFlowRate(1.0, pressureCoefficient, resistanceCoefficient, ActualHead, 1, DesignFlowRate);
          Pressure = VolumetricFlowRate * VolumetricFlowRate * resistanceCoefficient + ActualHead;
        }
      }
    }

    #endregion

    #region staticメソッド

    /// <summary>汎用ファンの特性係数を計算する</summary>
    /// <param name="flowRate">風量[m3/s]</param>
    /// <param name="pressure">全圧[kPa]</param>
    /// <param name="number">ファンの番手</param>
    /// <param name="efficiencyCoef">出力：効率特性係数</param>
    /// <param name="pressureCoef">出力：揚程特性係数</param>
    private static void getGeneralParameters(double flowRate, double pressure,
      double number, ref double[] efficiencyCoef, ref double[] pressureCoef)
    {
      //最大効率近似係数
      double[][] aj = new double[3][];
      aj[0] = new double[] { 1.397E-01, 1.303E-01, -1.179E-02, 3.219E-04 };
      aj[1] = new double[] { 7.649E-01, -9.105E-02, -5.345E-03, 1.006E-03 };
      aj[2] = new double[] { -7.141E-01, 1.325E-01, -2.639E-03, -5.690E-04 };

      //最大効率[-]の計算
      double[] act = new double[3];
      for (int i = 0; i < aj.Length; i++)
        for (int j = aj[0].Length - 1; 0 <= j; j--)
          act[i] = (number * act[i] + aj[i][j]);
      double maxEff = act[0] + pressure * (act[1] + act[2] * pressure);

      //効率特性
      double m2 = flowRate * flowRate;
      double m3 = m2 * flowRate;
      efficiencyCoef[0] = 0.337 * maxEff / m3;
      efficiencyCoef[1] = -1.700 * maxEff / m2;
      efficiencyCoef[2] = 2.350 * maxEff / flowRate;
      efficiencyCoef[3] = 0;

      //圧力特性//2017.01.05.係数修正 E.Togashi
      pressureCoef[0] = -0.203 * pressure / m2;
      pressureCoef[1] = 0.219 * pressure / flowRate;
      pressureCoef[2] = 0.984 * pressure;
    }

    #endregion

  }

}
