﻿/* SimpleCircuitBranch.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit
{

  /// <summary>圧力損失が体積流量の2乗に比例する単純な流路</summary>
  [Serializable]
  public class SimpleCircuitBranch : ICircuitBranch
  {
    
    /// <summary>上流節点を設定・取得する</summary>
    public CircuitNode UpStreamNode { get; set; }

    /// <summary>下流節点を設定・取得する</summary>
    public CircuitNode DownStreamNode { get; set; }

    /// <summary>体積流量[m3/s]を設定・取得する</summary>
    public double VolumetricFlowRate { get; set; }

    /// <summary>流動抵抗係数[kPa/(m/s)2]を設定・取得する</summary>
    public double Resistance { get; set; }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="pressure">圧力損失[kPa]</param>
    public SimpleCircuitBranch(double flowRate, double pressure)
    { Resistance = pressure / (flowRate * flowRate); }

    /// <summary>圧力損失[kPa]を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <returns>圧力損失[kPa]</returns>
    public double GetPressureDrop(double flowRate)
    {
      VolumetricFlowRate = flowRate;
      return flowRate * flowRate * Resistance;
    }

    /// <summary>前後差圧から体積流量[m3/s]を計算する</summary>
    /// <returns>体積流量[m3/s]</returns>
    public void UpdateFlowRateFromNodePressureDifference()
    {
      double dp = UpStreamNode.Pressure - DownStreamNode.Pressure;
      VolumetricFlowRate = Math.Sign(dp) * Math.Sqrt((Math.Abs(dp) / Resistance));
    }

  }
}
