﻿/* WaterPipe.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.Circuit
{
  /// <summary>水配管</summary>
  [Serializable]
  public class WaterPipe : ImmutableWaterPipe, ICircuitBranch
  {

    #region 定数宣言

    /// <summary>大気圧は101.325kPa（海抜0m）で一定とする</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    #endregion

    #region 列挙型の定義

    /// <summary>配管材料</summary>
    public enum Material
    {
      /// <summary>炭素鋼管</summary>
      CarbonSteel = 0,
      /// <summary>ビニル管</summary>
      Plastic = 1,
      /// <summary>ステンレス</summary>
      StainlessSteel = 2
    }

    /// <summary>断熱材料</summary>
    public enum Insulator
    {
      /// <summary>無し</summary>
      None = 0,
      /// <summary>ロックウール</summary>
      RockWool = 1,
      /// <summary>グラスウール</summary>
      GlassWool = 2,
      /// <summary>ポリスチレン</summary>
      Polystyrene = 3
    }

    #endregion

    #region インスタンス変数

    /// <summary>線熱通過率（水と空気の対流熱伝達率は含まない）[W/(mK)]</summary>
    private double linearThermalTransmittance = 0;

    /// <summary>配管長[m]</summary>
    private double length = 1;

    /// <summary>管内表面表面の粗度[m]</summary>
    private double roughness;

    #endregion

    #region プロパティ

    /// <summary>内径[m]を取得する</summary>
    public double InnerDiameter { get; private set; }

    /// <summary>外径[m]を取得する</summary>
    public double OuterDiameter { get; private set; }

    /// <summary>入口水温[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>周囲の乾球温度[C]を取得する</summary>
    public double AmbientTemperature { get; private set; }

    /// <summary>周囲の絶対湿度[kg/kg]を取得する</summary>
    public double AmbientHumidityRatio { get; private set; }

    /// <summary>線熱通過率[W/(mK)]を取得する</summary>
    public double LinearThermalTransmittance { get; private set; }

    /// <summary>配管長[m]を設定・取得する</summary>
    public double Length
    {
      get { return length; }
      set { if (0 < value) length = value; }
    }

    /// <summary>管内表面の粗度[m]を設定・取得する</summary>
    public double Roughness
    {
      get { return roughness; }
      set { if (0 < value) roughness = value; }
    }

    /// <summary>熱損失[kW]を取得する</summary>
    public double HeatLoss { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double OutletWaterTemperauture { get; private set; }

    /// <summary>配管の熱伝導率[W/(mK)]を取得する</summary>
    public double PipeThermalConductivity { get; private set; }

    /// <summary>断熱材の熱伝導率[W/(mK)]を取得する</summary>
    public double InsulatorThermalConductivity { get; private set; }

    /// <summary>断熱材厚[m]を取得する</summary>
    public double InsulatorThickness { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="length">長さ[m]</param>
    /// <param name="innerDiameter">内径[m]</param>
    /// <param name="roughness">配管粗度[m]</param>
    /// <param name="thickness">配管厚み[m]</param>
    public WaterPipe(double length, double innerDiameter, double roughness, double thickness)
    {
      InnerDiameter = innerDiameter;
      Length = length;
      Roughness = roughness;
      OuterDiameter = InnerDiameter + 2 * thickness;

      //断熱材無しとする
      RemoveInsulator();
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="length">長さ[m]</param>
    /// <param name="innerDiameter">内径[m]</param>
    /// <param name="material">配管材料</param>
    public WaterPipe(double length, double innerDiameter, Material material)
    {
      InnerDiameter = innerDiameter;
      Length = length;
      double dth;
      switch (material)
      {
        case Material.CarbonSteel:
          Roughness = 0.03e-3;
          dth = 1.0796e-3 * Math.Pow(1000 * innerDiameter, 0.3181);
          break;
        case Material.Plastic:
          Roughness = 0.03e-3;
          dth = 0.5196e-3 * Math.Pow(1000 * innerDiameter, 0.5746);
          break;
        default://SUS
          Roughness = 0.015e-3;
          dth = 0.7250e-3 * Math.Pow(1000 * innerDiameter, 0.4581);
          break;
      }
      OuterDiameter = InnerDiameter + 2 * dth;

      //断熱材無しとする
      RemoveInsulator();
    }

    /// <summary>初期化処理</summary>
    private void initialize()
    {
      //線熱通過率[W/(mK)]を初期化する
      linearThermalTransmittance = GetPipeLinearThermalTransmittance
        (InnerDiameter, InsulatorThermalConductivity, InsulatorThickness, PipeThermalConductivity,
        0.5 * (OuterDiameter - InnerDiameter));

      //熱流を初期化する//水量は水速2.0m/s相当
      UpdateHeatFlow(7, InnerDiameter * InnerDiameter / 4d * Math.PI * 2 * 1000, 25, 0.02);
    }

    #endregion

    #region 熱流に関わるインスタンスメソッド

    /// <summary>熱流を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[m3/s]</param>
    /// <param name="ambientTemperature">周囲の乾球温度[C]</param>
    /// <param name="ambientHumidityRatio">周囲の絶対湿度[C]</param>
    public void UpdateHeatFlow(double inletWaterTemperature,
      double waterFlowRate, double ambientTemperature, double ambientHumidityRatio)
    {
      VolumetricFlowRate = waterFlowRate;
      InletWaterTemperature = inletWaterTemperature;
      AmbientTemperature = ambientTemperature;
      AmbientHumidityRatio = ambientHumidityRatio;

      //体積流量[m3/s]を質量流量[kg/s]に換算
      double mw = waterFlowRate / Water.GetLiquidDensity(inletWaterTemperature);

      //水側の対流熱伝達率[W/(m2K)]を計算
      double aw = GetInsideHeatTransferCoefficient(InletWaterTemperature, InnerDiameter, mw);

      //配管表面温度は水温と空気温度の平均と仮定
      double ts = 0.5 * (InletWaterTemperature + AmbientTemperature);

      double err = 1;
      int iterNum = 0;
      while (0.01 < err)
      {
        if (20 < iterNum) throw new Exception("WaterPipe iteration error");

        //空気側の対流熱伝達率[W/(m2K)]を計算
        double aa = GetOutSideHeatTransferCoefficient
          (AmbientTemperature, AmbientHumidityRatio, OuterDiameter, ts);

        //線熱通過率[W/(mK)]を更新する
        LinearThermalTransmittance = GetAirToWaterLinearThermalTransmittance
          (aw, aa, InnerDiameter, OuterDiameter, linearThermalTransmittance);

        //線表面温度[C]を更新して誤差を評価
        double tsOld = ts;
        ts = AmbientTemperature - LinearThermalTransmittance
          * (AmbientTemperature - InletWaterTemperature) / aa;
        err = Math.Abs(ts - tsOld);
        iterNum++;
      }

      //熱損失と出口水温を更新
      HeatLoss = LinearThermalTransmittance * Length
        * (AmbientTemperature - InletWaterTemperature) / 1000d;
      double cpw = Water.GetLiquidIsobaricSpecificHeat(InletWaterTemperature);
      OutletWaterTemperauture = HeatLoss / (mw * cpw) + InletWaterTemperature;
    }

    /// <summary>断熱材を設定する</summary>
    /// <param name="thickness">断熱材厚み[m]</param>
    /// <param name="insulator">断熱材種別</param>
    public void SetInsulator(double thickness, Insulator insulator)
    {
      InsulatorThickness = thickness;
      switch (insulator)
      {
        case Insulator.RockWool:
          SetInsulator(thickness, 0.044);
          break;
        case Insulator.GlassWool:
          SetInsulator(thickness, 0.043);
          break;
        case Insulator.Polystyrene:
          SetInsulator(thickness, 0.043);
          break;
        default:
          throw new Exception("Invalid Insulator");
      }
    }

    /// <summary>断熱材を設定する</summary>
    /// <param name="thickness">断熱材厚み[m]</param>
    /// <param name="thermalConductivity">断熱材熱伝導率[W/(mK)]</param>
    public void SetInsulator(double thickness, double thermalConductivity)
    {
      InsulatorThickness = thickness;
      InsulatorThermalConductivity = thermalConductivity;
      initialize();
    }

    /// <summary>断熱材を取り外す</summary>
    public void RemoveInsulator()
    {
      InsulatorThickness = 0.0d;
      InsulatorThermalConductivity = 0.04d;
      initialize();
    }

    /// <summary>配管サイズを変更する</summary>
    /// <param name="innerDiameter">内径[m]</param>
    /// <param name="thickness">厚み[m]</param>
    public void ChangeSize(double innerDiameter, double thickness)
    {
      if ((0 < innerDiameter) && (0 < thickness))
      {
        InnerDiameter = innerDiameter;
        OuterDiameter = InnerDiameter + 2 * thickness;
        initialize();
      }
    }

    /// <summary>配管熱伝導率を設定する</summary>
    /// <param name="thermalConductivity">熱伝導率[W/(mK)]</param>
    public void SetPipeThermalConductivity(double thermalConductivity)
    {
      if (0 < thermalConductivity) PipeThermalConductivity = thermalConductivity;
      initialize();
    }

    #endregion

    #region ICircuitBranch実装

    /// <summary>流量[m3/s]を設定・取得する</summary>
    public double VolumetricFlowRate { get; set; }

    /// <summary>上流節点を設定・取得する</summary>
    public CircuitNode UpStreamNode { get; set; }

    /// <summary>下流節点を設定・取得する</summary>
    public CircuitNode DownStreamNode { get; set; }

    /// <summary>前後差圧から体積流量[m3/s]を計算する</summary>
    /// <returns>体積流量[m3/s]</returns>
    public void UpdateFlowRateFromNodePressureDifference()
    {
      //水物性（動粘性係数[m2/s], 熱拡散率[m2/s]）を計算
      double v = Water.GetLiquidDynamicViscosity(InletWaterTemperature);
      double rho = Water.GetLiquidDensity(InletWaterTemperature);

      //流路面積[m2]と前後差圧[kPa]
      double fArea = InnerDiameter * InnerDiameter / 4d * Math.PI;
      double dp = (UpStreamNode.Pressure - DownStreamNode.Pressure) * 1000;

      //流速[m/s]を収束計算
      Roots.ErrorFunction eFnc = delegate (double vel)
      {
        double reNumber = vel * InnerDiameter / v;
        double ff = Conduit.GetFrictionFactor(reNumber, Roughness / InnerDiameter);
        return Conduit.GetVelocity(ff, rho, Length, InnerDiameter, dp) - vel;
      };
      //初期値はレイノルズ数（流速大）が大きめの点とすると収束がうまくいく
      VolumetricFlowRate = Roots.Newton(eFnc, 5, 0.0001, 1e-10, 1e-9, 30) * fArea;
    }

    #endregion

    /// <summary>圧力損失[kPa]を取得する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <returns>圧力損失[kPa]</returns>
    public double GetPressureDrop(double flowRate)
    {
      double v = Water.GetLiquidDynamicViscosity(InletWaterTemperature);
      double rho = Water.GetLiquidDensity(InletWaterTemperature);
      double fArea = InnerDiameter * InnerDiameter / 4d * Math.PI;
      double vel = flowRate / fArea;
      double reNumber = vel * InnerDiameter / v;
      double ff = Conduit.GetFrictionFactor(reNumber, Roughness / InnerDiameter);
      return 0.001 * Conduit.GetPressureDrop(ff, rho, Length, InnerDiameter, vel);
    }

    #region クラスメソッド

    /// <summary>配管内表面の対流熱伝達率[W/(m2·K)]を計算する</summary>
    /// <param name="waterTemperature">水温[C]</param>
    /// <param name="diameter">直径[m]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <returns>対流熱伝達率[W/(m2·K)]</returns>
    /// <remarks>乱流かつ発達した流れの場合</remarks>
    public static double GetInsideHeatTransferCoefficient
      (double waterTemperature, double diameter, double waterFlowRate)
    {
      //水の物性を計算
      //動粘性係数[m2/s], 熱拡散率[m2/s], 熱伝導率[W/(m·K)], 密度[kg/m3]
      double v = Water.GetLiquidDynamicViscosity(waterTemperature);
      double a = Water.GetLiquidThermalDiffusivity(waterTemperature);
      double lam = Water.GetLiquidThermalConductivity(waterTemperature);
      double rho = Water.GetLiquidDensity(waterTemperature);

      //配管内流速[m/s]からヌセルト数を計算
      double u = waterFlowRate / (rho * diameter * diameter / 4 * Math.PI);
      double reNumber = u * diameter / v;
      double prNumber = v / a;
      double nuNumber = 0.023 * Math.Pow(reNumber, 0.8) * Math.Pow(prNumber, 0.4);

      //ヌセルト数から対流熱伝達率[W/m2K]を計算
      return nuNumber * lam / diameter;
    }

    /// <summary>配管外表面の自然対流熱伝達率[W/(m2·K)]を計算する</summary>
    /// <param name="drybulbTemperature">湿り空気の乾球温度[C]</param>
    /// <param name="humidityRatio">湿り空気の絶対湿度[kg/kg]</param>
    /// <param name="diameter">直径[m]</param>
    /// <param name="surfaceTemperature">配管表面温度[C]</param>
    /// <returns>配管外表面の自然対流熱伝達率[W/(m2·K)]</returns>
    public static double GetOutSideHeatTransferCoefficient
      (double drybulbTemperature, double humidityRatio, double diameter, double surfaceTemperature)
    {
      //湿り空気の物性を計算
      //動粘性係数[m2/s], 熱拡散率[m2/s], 膨張率[1/K], 熱伝導率[W/(m·K)]
      double v = MoistAir.GetDynamicViscosity(drybulbTemperature, humidityRatio, ATMOSPHERIC_PRESSURE);
      double a = MoistAir.GetThermalDiffusivity(drybulbTemperature, humidityRatio, ATMOSPHERIC_PRESSURE);
      double beta = MoistAir.GetExpansionCoefficient(drybulbTemperature);
      double lam = MoistAir.GetThermalConductivity(drybulbTemperature);

      //グラフホフ数の計算
      double grNumber = 9.8 * Math.Pow(diameter, 3) * beta *
        Math.Abs(surfaceTemperature - drybulbTemperature) / (v * v);

      //プラントル数の計算
      double prNumber = v / a;

      //ヌセルト数の計算
      double nuNumber;
      double grpr = grNumber * prNumber;
      if (grpr < 1e10) nuNumber = 0.53 * Math.Pow(grpr, 0.25);
      else nuNumber = 0.13 * Math.Pow(grpr, 0.333);

      //ヌセルト数から対流熱伝達率[W/(m2K)]を計算
      return nuNumber * lam / diameter;
    }

    /// <summary>水から空気までの線熱通過率[W/(mK)]を計算する</summary>
    /// <param name="insideHeatTransferCoef">水側対流熱伝達率[W/(m2K)]</param>
    /// <param name="outsideHeatTransferCoef">空気側対流熱伝達率[W/(m2K)]</param>
    /// <param name="innerDiameter">内径[m]</param>
    /// <param name="outerDiameter">外径[m]</param>
    /// <param name="thermalConductance">内表面から外表面までの線熱通過率[W/(mK)]</param>
    /// <returns>水から空気までの線熱通過率[W/(mK)]</returns>
    public static double GetAirToWaterLinearThermalTransmittance
      (double insideHeatTransferCoef, double outsideHeatTransferCoef,
      double innerDiameter, double outerDiameter, double thermalConductance)
    {
      double p2 = 2 * Math.PI;
      double kl = 1 / (insideHeatTransferCoef * innerDiameter) +
        1 / (outsideHeatTransferCoef * outerDiameter) + p2 / thermalConductance;
      return p2 / kl;
    }

    /// <summary>配管の内外表面間の線熱通過率[W/(mK)]を計算する</summary>
    /// <param name="innerDiameter">配管の内径[m]</param>
    /// <param name="insulatorThermalConductivity">断熱材の熱伝導率[W/(mK)]</param>
    /// <param name="insulatorThickness">断熱材の厚み[m]</param>
    /// <param name="pipeThermalConductivity">配管の熱伝導率[W/(mK)]</param>
    /// <param name="pipeThickness">配管の厚み[m]</param>
    /// <returns>配管の内外表面間の線熱通過率[W/(mK)]</returns>
    public static double GetPipeLinearThermalTransmittance
      (double innerDiameter, double insulatorThermalConductivity, double insulatorThickness,
      double pipeThermalConductivity, double pipeThickness)
    {
      double d_b = innerDiameter + pipeThickness;
      double d_o = d_b + insulatorThickness;
      double klc = 1 / pipeThermalConductivity * Math.Log(d_b / innerDiameter)
        + 1 / insulatorThermalConductivity * Math.Log(d_o / d_b);
      return Math.PI / klc;
    }

    #endregion

  }

  #region Immutable Interface

  /// <summary>読み取り専用の配管</summary>
  public interface ImmutableWaterPipe
  {
    /// <summary>内径[m]を取得する</summary>
    double InnerDiameter { get; }

    /// <summary>外径[m]を取得する</summary>
    double OuterDiameter { get; }

    /// <summary>入口水温[C]を取得する</summary>
    double InletWaterTemperature { get; }

    /// <summary>周囲の乾球温度[C]を取得する</summary>
    double AmbientTemperature { get; }

    /// <summary>周囲の絶対湿度[kg/kg]を取得する</summary>
    double AmbientHumidityRatio { get; }

    /// <summary>線熱通過率[W/(mK)]を取得する</summary>
    double LinearThermalTransmittance { get; }

    /// <summary>配管長[m]を取得する</summary>
    double Length { get; }

    /// <summary>水量[m3/s]を取得する</summary>
    double VolumetricFlowRate { get; }

    /// <summary>熱損失[W]を取得する</summary>
    double HeatLoss { get; }

    /// <summary>出口水温[C]を取得する</summary>
    double OutletWaterTemperauture { get; }
  }

  #endregion

}
