﻿/* HVACSystemModel.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Weather;
using Popolo.ThermalLoad;
using Popolo.HVAC.SubSystem;
using Popolo.Numerics;
using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC
{
  /// <summary>建物・熱源・空調設備連成モデル</summary>
  [Serializable]
  public class HVACSystemModel
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ
    
    /// <summary>空調機系統の倍数</summary>
    private int[] acFactor;

    /// <summary>現在の日時を取得する</summary>
    public DateTime CurrentDateTime { get; private set; }

    /// <summary>計算時間間隔[sec]</summary>
    private double timeStep = 3600;

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set
      {
        timeStep = value;
        bModel.TimeStep = value;
        foreach (IAirConditioningSystemModel acs in acModels) acs.TimeStep = value;
        hsModel.TimeStep = value;
      }
    }

    /// <summary>建物熱負荷計算モデル</summary>
    private BuildingThermalModel bModel;

    /// <summary>二次側空調システムモデルリスト</summary>
    private IAirConditioningSystemModel[] acModels;

    /// <summary>熱源システムモデル</summary>
    private HeatSourceSystemModel hsModel;

    /// <summary>建物熱負荷計算モデルを取得する</summary>
    public ImmutableBuildingThermalModel BuildingThermalModel { get { return bModel; } }

    /// <summary>二次側空調システムモデルリストを取得する</summary>
    public ImmutableAirConditioningSystemModel[] AirConditioningSystemModel { get { return acModels; } }

    /// <summary>熱源システムモデルを取得する</summary>
    public ImmutableHeatSourceSystemModel HeatSourceSystemModel { get { return hsModel; } }

    /// <summary>温水下限温度[C]を設定・取得する</summary>
    public double HotWaterLowerLimitTemperature { get; set; } = 35;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint
    {
      get { return hsModel.HotWaterSupplyTemperatureSetpoint; }
      set { hsModel.HotWaterSupplyTemperatureSetpoint = value; }
    }

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水還温度[C]を取得する</summary>
    public double HotWaterReturnTemperature { get; private set; } = 40;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>冷水上限温度[C]を設定・取得する</summary>
    public double ChilledWaterUpperLimitTemperature { get; set; } = 15;

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint
    {
      get { return hsModel.ChilledWaterSupplyTemperatureSetpoint; }
      set { hsModel.ChilledWaterSupplyTemperatureSetpoint = value; }
    }

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水還温度[C]を取得する</summary>
    public double ChilledWaterReturnTemperature { get; private set; } = 12;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>未処理加熱負荷[kW]を取得する</summary>
    public double RemainingHeatingLoad { get; private set; }

    /// <summary>未処理冷却負荷[kW]を取得する</summary>
    public double RemainingCoolingLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="bModel">建物熱負荷計算モデル</param>
    /// <param name="acModels">二次側空調システムモデルリスト</param>
    /// <param name="hsModel">熱源システムモデル</param>
    public HVACSystemModel
      (BuildingThermalModel bModel, IAirConditioningSystemModel[] acModels, HeatSourceSystemModel hsModel)
    {
      this.bModel = bModel;
      this.acModels = acModels;
      this.hsModel = hsModel;
      acFactor = new int[acModels.Length];
      for (int i = 0; i < acFactor.Length; i++) acFactor[i] = 1;
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>状態を更新する</summary>
    public void Update()
    {
      //未処理負荷初期化
      RemainingCoolingLoad = RemainingHeatingLoad = 0;

      //二次側からの還温度を計算
      calcReturnWaterState(ChilledWaterSupplyTemperatureSetpoint, HotWaterSupplyTemperatureSetpoint);

      //一次側過負荷判定
      hsModel.ForecastSupplyWaterTemperature
        (ChilledWaterFlowRate, ChilledWaterReturnTemperature, HotWaterFlowRate, HotWaterReturnTemperature);

      //冷却能力不足の場合
      if (hsModel.IsOverLoad_C)
      {
        //往温度上限値で計算
        calcReturnWaterState(ChilledWaterUpperLimitTemperature, HotWaterSupplyTemperatureSetpoint);
        hsModel.ForecastSupplyWaterTemperature
          (ChilledWaterFlowRate, ChilledWaterReturnTemperature, HotWaterFlowRate, HotWaterReturnTemperature);

        //未処理負荷発生
        if (ChilledWaterUpperLimitTemperature < hsModel.ChilledWaterSupplyTemperature)
          RemainingCoolingLoad = ChilledWaterFlowRate * WATER_SPECIFIC_HEAT *
            (hsModel.ChilledWaterSupplyTemperature - ChilledWaterUpperLimitTemperature);
        //往温度上昇
        else
        {
          Roots.ErrorFunction eFnc = delegate (double tcSply)
          {
            calcReturnWaterState(tcSply, HotWaterSupplyTemperatureSetpoint);
            hsModel.ForecastSupplyWaterTemperature
            (ChilledWaterFlowRate, ChilledWaterReturnTemperature, HotWaterFlowRate, HotWaterReturnTemperature);
            return tcSply - hsModel.ChilledWaterSupplyTemperature;
          };
          ChilledWaterSupplyTemperature = Roots.Bisection
            (eFnc, ChilledWaterSupplyTemperatureSetpoint, ChilledWaterUpperLimitTemperature, 0.01, 0.01, 10);
        }
      }
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;

      //加熱能力不足の場合
      if (hsModel.IsOverLoad_H)
      {
        //往温度下限値で計算
        calcReturnWaterState(ChilledWaterSupplyTemperature, HotWaterLowerLimitTemperature);
        hsModel.ForecastSupplyWaterTemperature
          (ChilledWaterFlowRate, ChilledWaterReturnTemperature, HotWaterFlowRate, HotWaterReturnTemperature);

        //未処理負荷発生
        if (hsModel.HotWaterSupplyTemperature < HotWaterLowerLimitTemperature)
          RemainingHeatingLoad = HotWaterFlowRate * WATER_SPECIFIC_HEAT *
            (HotWaterLowerLimitTemperature - hsModel.HotWaterSupplyTemperature);
        //往温度低下
        else
        {
          Roots.ErrorFunction eFnc = delegate (double tcSply)
          {
            calcReturnWaterState(ChilledWaterSupplyTemperature, tcSply);
            hsModel.ForecastSupplyWaterTemperature
            (ChilledWaterFlowRate, ChilledWaterReturnTemperature, HotWaterFlowRate, HotWaterReturnTemperature);
            return tcSply - hsModel.HotWaterSupplyTemperature;
          };
          HotWaterSupplyTemperature = Roots.Bisection
            (eFnc, HotWaterLowerLimitTemperature, HotWaterSupplyTemperatureSetpoint, 0.01, 0.01, 10);
        }
      }
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;

      //状態確定
      hsModel.FixState();
      foreach (IAirConditioningSystemModel ac in acModels) ac.FixState();
      bModel.FixState();
    }

    /// <summary>冷温水の還温度[C]を計算する</summary>
    /// <param name="tcwSP">冷水往温度</param>
    /// <param name="thwSP">温水往温度</param>
    private void calcReturnWaterState(double tcwSP, double thwSP)
    {
      ChilledWaterFlowRate = HotWaterFlowRate = ChilledWaterReturnTemperature = HotWaterReturnTemperature = 0;      
      for (int i = 0; i < acModels.Length; i++)
      {
        IAirConditioningSystemModel acm = acModels[i];
        acm.ForecastReturnWaterTemperature(tcwSP, thwSP);
        ChilledWaterReturnTemperature += acm.ChilledWaterReturnTemperature * acm.ChilledWaterFlowRate * acFactor[i];
        HotWaterReturnTemperature += acm.HotWaterReturnTemperature * acm.HotWaterFlowRate * acFactor[i];
        ChilledWaterFlowRate += acm.ChilledWaterFlowRate * acFactor[i];
        HotWaterFlowRate += acm.HotWaterFlowRate * acFactor[i];
      }
      if (ChilledWaterFlowRate == 0) ChilledWaterReturnTemperature = tcwSP;
      else ChilledWaterReturnTemperature /= ChilledWaterFlowRate;
      if (HotWaterFlowRate == 0) HotWaterReturnTemperature = thwSP;
      else HotWaterReturnTemperature /= HotWaterFlowRate;
    }

    /// <summary>外気条件を更新する</summary>
    /// <param name="dTime">現在の日時</param>
    /// <param name="sun">太陽</param>
    /// <param name="temperature">外気乾球温度[C]</param>
    /// <param name="humidityRatio">外気絶対湿度[kg/kg]</param>
    /// <param name="nocRadiation">夜間放射[W/m2]</param>
    public void UpdateOutdoorCondition
      (DateTime dTime, ImmutableSun sun, double temperature, double humidityRatio, double nocRadiation)
    {
      MoistAir outdoorAir = new MoistAir(temperature, humidityRatio);
      CurrentDateTime = dTime;
      bModel.UpdateOutdoorCondition(dTime, sun, temperature, humidityRatio, nocRadiation);
      foreach (IAirConditioningSystemModel acs in acModels)
      {
        acs.CurrentDateTime = dTime;
        acs.OutdoorAir = outdoorAir;
      }
      hsModel.CurrentDateTime = dTime;
      hsModel.OutdoorAir = outdoorAir;
    }

    /// <summary>空調システムの倍数を設定する</summary>
    /// <param name="airConditioningSystemIndex">空調システム系統番号</param>
    /// <param name="factor">倍数[-]</param>
    public void SetACFactor(int airConditioningSystemIndex, int factor)
    { acFactor[airConditioningSystemIndex] = factor; }

    #endregion

  }
}
