﻿/* AirToAirFlatPlateHeatExchanger.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.HeatExchanger
{

  /// <summary>空気対空気の静止型熱交換器クラス</summary>
  [Serializable]
  public class AirToAirFlatPlateHeatExchanger
  {

    #region 定数宣言

    /// <summary>大気圧[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    #endregion

    #region 列挙型

    /// <summary>空気の流れ</summary>
    public enum AirFlow
    {
      /// <summary>向流</summary>
      CounterFlow,
      /// <summary>直交流</summary>
      CrossFlow
    }

    #endregion

    #region インスタンス変数

    /// <summary>顕熱熱貫流率[kW/K]</summary>
    private double sensibleHeatTransferCoefficient;

    /// <summary>潜熱熱貫流率[kg/(kg/kg)]</summary>
    private double latentHeatTransferCoefficient;

    #endregion

    #region プロパティ

    /// <summary>全熱交換器か否かを取得する</summary>
    public bool IsTotalHeatExchanger { get; private set; }

    /// <summary>給気風量[CMH]を取得する</summary>
    public double SupplyAirFlowVolume { get; private set; }

    /// <summary>排気風量[CMH]を取得する</summary>
    public double ExhaustAirFlowVolume { get; private set; }

    /// <summary>給気入口乾球温度[C]を取得する</summary>
    public double SupplyAirInletDrybulbTemperature { get; private set; }

    /// <summary>排気入口乾球温度[C]を取得する</summary>
    public double ExhaustAirInletDrybulbTemperature { get; private set; }

    /// <summary>給気出口乾球温度[C]を取得する</summary>
    public double SupplyAirOutletDrybulbTemperature { get; private set; }

    /// <summary>排気出口乾球温度[C]を取得する</summary>
    public double ExhaustAirOutletDrybulbTemperature { get; private set; }

    /// <summary>給気入口絶対湿度[kg/kg]を取得する</summary>
    public double SupplyAirInletHumidityRatio { get; private set; }

    /// <summary>排気入口絶対湿度[kg/kg]を取得する</summary>
    public double ExhaustAirInletHumidityRatio { get; private set; }

    /// <summary>給気出口絶対湿度[kg/kg]を取得する</summary>
    public double SupplyAirOutletHumidityRatio { get; private set; }

    /// <summary>排気出口絶対湿度[kg/kg]を取得する</summary>
    public double ExhaustAirOutletHumidityRatio { get; private set; }

    /// <summary>顕熱交換効率[-]を取得する</summary>
    public double SensibleEfficiency { get; private set; }

    /// <summary>潜熱交換効率[-]を取得する</summary>
    public double LatentEfficiency { get; private set; }

    /// <summary>空気の流れのタイプ</summary>
    public AirFlow Flow { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="sensibleEfficiency">顕熱交換効率[-]</param>
    /// <param name="latentEfficiency">潜熱交換効率[-]</param>
    /// <param name="flow">空気の流れのタイプ</param>
    /// <param name="isTotalHeatExchanger">全熱交換器の場合にはtrue</param>
    public AirToAirFlatPlateHeatExchanger
      (double supplyAirFlowVolume, double exhaustAirFlowVolume,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio,
      double sensibleEfficiency, double latentEfficiency,
      AirFlow flow, bool isTotalHeatExchanger)
    {
      //機器情報を保存
      IsTotalHeatExchanger = isTotalHeatExchanger;
      Flow = flow;

      //空気の質量流量を計算
      double svSA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletSADrybulbTemperature, inletSAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double svEA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletEADrybulbTemperature, inletEAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double mSA = supplyAirFlowVolume / (3600 * svSA);
      double mEA = supplyAirFlowVolume / (3600 * svEA);

      //顕熱貫流率を計算
      sensibleHeatTransferCoefficient = GetSensibleHeatTransferCoefficient
        (mSA, mEA, inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio, sensibleEfficiency, flow);

      //全熱交換器の場合には潜熱貫流率[kg/(kg/kg)]を計算
      if (IsTotalHeatExchanger)
      {
        latentHeatTransferCoefficient = GetLatentHeatTransferCoefficient
          (mSA, mEA, inletSAHumidityRatio, inletEAHumidityRatio, latentEfficiency, flow);
      }

      //定格条件で成り行き計算
      UpdateState(supplyAirFlowVolume, exhaustAirFlowVolume, inletSADrybulbTemperature,
        inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>出口状態を更新する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    public void UpdateState
      (double supplyAirFlowVolume, double exhaustAirFlowVolume,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio)
    {
      //風量を保存
      SupplyAirFlowVolume = supplyAirFlowVolume;
      ExhaustAirFlowVolume = exhaustAirFlowVolume;

      //入口空気状態を保存
      SupplyAirInletDrybulbTemperature = inletSADrybulbTemperature;
      SupplyAirInletHumidityRatio = inletSAHumidityRatio;
      ExhaustAirInletDrybulbTemperature = inletEADrybulbTemperature;
      ExhaustAirInletHumidityRatio = inletEAHumidityRatio;

      //風量が0の場合
      if (supplyAirFlowVolume <= 0 || exhaustAirFlowVolume <= 0)
      {
        SensibleEfficiency = 0;
        SupplyAirOutletDrybulbTemperature = SupplyAirInletDrybulbTemperature;
        ExhaustAirOutletDrybulbTemperature = ExhaustAirInletDrybulbTemperature;
        SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio;
        ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio;
        return;
      }

      //空気の質量流量を計算
      double svSA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletSADrybulbTemperature, inletSAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double svEA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletEADrybulbTemperature, inletEAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double mSA = supplyAirFlowVolume / (3600 * svSA);
      double mEA = exhaustAirFlowVolume / (3600 * svEA);

      //熱通過有効度[-]を計算
      double effectiveness, mcMin, capacityRate;
      bool isMcMinSA;
      GetSensibleEffectiveness
        (mSA, mEA, inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio, sensibleHeatTransferCoefficient, Flow,
        out effectiveness, out mcMin, out capacityRate, out isMcMinSA);

      //熱交換効率[-]を計算
      double eff2;
      if (isMcMinSA)
      {
        SensibleEfficiency = effectiveness;
        eff2 = effectiveness * capacityRate;
      }
      else
      {
        SensibleEfficiency = effectiveness * capacityRate;
        eff2 = effectiveness;
      }

      //出口空気状態の計算
      SupplyAirOutletDrybulbTemperature =
        SupplyAirInletDrybulbTemperature - SensibleEfficiency *
        (SupplyAirInletDrybulbTemperature - ExhaustAirInletDrybulbTemperature);
      ExhaustAirOutletDrybulbTemperature = ExhaustAirInletDrybulbTemperature -
        eff2 * (ExhaustAirInletDrybulbTemperature - SupplyAirInletDrybulbTemperature);

      //水分交換
      if (IsTotalHeatExchanger)
      {
        //熱通過有効度[-]を計算
        GetLatentEffectiveness
          (mSA, mEA, inletSAHumidityRatio, inletEAHumidityRatio,
          latentHeatTransferCoefficient, Flow,
          out effectiveness, out mcMin, out capacityRate);

        //熱交換効率[-]を計算
        if (mcMin == mSA)
        {
          LatentEfficiency = effectiveness;
          eff2 = effectiveness * capacityRate;
        }
        else
        {
          LatentEfficiency = effectiveness * capacityRate;
          eff2 = effectiveness;
        }

        SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio -
          LatentEfficiency * (SupplyAirInletHumidityRatio - ExhaustAirInletHumidityRatio);
        ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio -
          eff2 * (ExhaustAirInletHumidityRatio - SupplyAirInletHumidityRatio);
      }
      else
      {
        SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio;
        ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio;
      }
    }

    /// <summary>全熱交換効率[-]を計算する</summary>
    /// <returns>全熱交換効率[-]</returns>
    public double GetTotalEfficiency()
    {
      //空気の出入口エンタルピーの計算
      double hSAi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirInletDrybulbTemperature, SupplyAirInletHumidityRatio);
      double hSAo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirOutletDrybulbTemperature, SupplyAirOutletHumidityRatio);
      double hEAi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (ExhaustAirInletDrybulbTemperature, ExhaustAirInletHumidityRatio);

      return (hSAi - hSAo) / (hSAi - hEAi);
    }

    #endregion

    #region staticメソッド

    /// <summary>顕熱貫流率[kW/K]を計算する</summary>
    /// <param name="supplyAirMassFlowRate">給気質量流量[kg/s]</param>
    /// <param name="exhaustAirMassFlowRate">排気質量流量[kg/s]</param>
    /// <param name="supplyAirDrybulbTemperature">給気乾球温度[C]</param>
    /// <param name="supplyAirHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="exhaustAirDrybulbTemperature">排気乾球温度[C]</param>
    /// <param name="exhaustAirHumitidyRatio">排気絶対湿度[kg/kg]</param>
    /// <param name="efficiency">顕熱熱交換効率[-]</param>
    /// <param name="flow">空気の流れのタイプ</param>
    /// <returns>顕熱貫流率[kW/K]</returns>
    public static double GetSensibleHeatTransferCoefficient
      (double supplyAirMassFlowRate, double exhaustAirMassFlowRate,
      double supplyAirDrybulbTemperature, double supplyAirHumidityRatio,
      double exhaustAirDrybulbTemperature, double exhaustAirHumitidyRatio,
      double efficiency, AirFlow flow)
    {
      //熱容量流量[kW/K]の計算
      double cpSA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (supplyAirDrybulbTemperature, supplyAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      double cpEA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (exhaustAirDrybulbTemperature, exhaustAirHumitidyRatio, ATMOSPHERIC_PRESSURE);
      double mcSA = supplyAirMassFlowRate * cpSA;
      double mcEA = exhaustAirMassFlowRate * cpEA;
      double mcMin = Math.Min(mcSA, mcEA);

      //熱容量流量比[-]の計算
      double capacityRatio = mcMin / Math.Max(mcSA, mcEA);

      //熱通過有効度[-]を計算
      double effectiveness;
      if (mcSA < mcEA) effectiveness = efficiency;
      else effectiveness = efficiency / capacityRatio;

      //移動単位数を計算
      HeatExchange.FlowType fType;
      if (flow == AirFlow.CounterFlow) fType = HeatExchange.FlowType.CounterFlow;
      else fType = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
      double ntu = HeatExchange.GetNTU(effectiveness, capacityRatio, fType);

      return ntu * mcMin;
    }

    /// <summary>潜熱貫流率[kg/(kg/kg)]を計算する</summary>
    /// <param name="supplyAirMassFlowRate">給気質量流量[kg/s]</param>
    /// <param name="exhaustAirMassFlowRate">排気質量流量[kg/s]</param>
    /// <param name="supplyAirHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="exhaustAirHumitidyRatio">排気絶対湿度[kg/kg]</param>
    /// <param name="efficiency">潜熱熱交換効率[-]</param>
    /// <param name="flow">空気の流れのタイプ</param>
    /// <returns>潜熱貫流率[kg/(kg/kg)]</returns>
    public static double GetLatentHeatTransferCoefficient
      (double supplyAirMassFlowRate, double exhaustAirMassFlowRate,
      double supplyAirHumidityRatio, double exhaustAirHumitidyRatio,
      double efficiency, AirFlow flow)
    {
      //質量流量比[-]の計算
      double mcMin = Math.Min(supplyAirMassFlowRate, exhaustAirMassFlowRate);
      double massFlowRatio = mcMin / Math.Max
        (supplyAirMassFlowRate, exhaustAirMassFlowRate);

      //熱通過有効度[-]を計算
      double effectiveness;
      if (supplyAirMassFlowRate < exhaustAirMassFlowRate) effectiveness = efficiency;
      else effectiveness = efficiency / massFlowRatio;

      //移動単位数を計算
      HeatExchange.FlowType fType;
      if (flow == AirFlow.CounterFlow) fType = HeatExchange.FlowType.CounterFlow;
      else fType = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
      double ntu = HeatExchange.GetNTU(effectiveness, massFlowRatio, fType);

      return ntu * mcMin;
    }

    /// <summary>顕熱熱通過有効度[-]を計算する</summary>
    /// <param name="supplyAirMassFlowRate">給気質量流量[kg/s]</param>
    /// <param name="exhaustAirMassFlowRate">排気質量流量[kg/s]</param>
    /// <param name="supplyAirDrybulbTemperature">給気乾球温度[C]</param>
    /// <param name="supplyAirHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="exhaustAirDrybulbTemperature">排気乾球温度[C]</param>
    /// <param name="exhaustAirHumitidyRatio">排気絶対湿度[kg/kg]</param>
    /// <param name="heatTransferCoefficient">顕熱貫流率[kW/K]</param>
    /// <param name="flow">空気の流れのタイプ</param>
    /// <param name="effectiveness">顕熱熱通過有効度[-]</param>
    /// <param name="mcMin">小さい側の熱容量流量[kg/s]</param>
    /// <param name="capacityRatio">質量流量比[-]</param>
    /// <param name="isMcMinSASide">給気側の熱容量流量が小さい場合にtrue</param>
    public static void GetSensibleEffectiveness
      (double supplyAirMassFlowRate, double exhaustAirMassFlowRate,
      double supplyAirDrybulbTemperature, double supplyAirHumidityRatio,
      double exhaustAirDrybulbTemperature, double exhaustAirHumitidyRatio,
      double heatTransferCoefficient, AirFlow flow, out double effectiveness,
      out double mcMin, out double capacityRatio, out bool isMcMinSASide)
    {
      //熱容量流量[kW/K]の計算
      double cpSA = MoistAir.GetSpecificHeat(supplyAirHumidityRatio);
      double cpEA = MoistAir.GetSpecificHeat(exhaustAirHumitidyRatio);
      double mcSA = supplyAirMassFlowRate * cpSA;
      double mcEA = exhaustAirMassFlowRate * cpEA;
      mcMin = Math.Min(mcSA, mcEA);
      isMcMinSASide = (mcSA == mcMin);

      //熱容量流量比[-]の計算
      capacityRatio = mcMin / Math.Max(mcSA, mcEA);

      //熱通過有効度[-]の計算
      double ntu = heatTransferCoefficient / mcMin;
      HeatExchange.FlowType fType;
      if (flow == AirFlow.CounterFlow) fType = HeatExchange.FlowType.CounterFlow;
      else fType = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
      effectiveness = HeatExchange.GetEffectiveness(ntu, capacityRatio, fType);
    }

    /// <summary>潜熱熱通過有効度[-]を計算する</summary>
    /// <param name="supplyAirMassFlowRate">給気質量流量[kg/s]</param>
    /// <param name="exhaustAirMassFlowRate">排気質量流量[kg/s]</param>
    /// <param name="supplyAirHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="exhaustAirHumitidyRatio">排気絶対湿度[kg/kg]</param>
    /// <param name="heatTransferCoefficient">潜熱貫流率[kg/(kg/kg)]</param>
    /// <param name="flow">空気の流れのタイプ</param>
    /// <param name="effectiveness">潜熱熱通過有効度[-]</param>
    /// <param name="mMin">小さい側の質量流量[kg/s]</param>
    /// <param name="capacityRatio">質量流量比[-]</param>
    public static void GetLatentEffectiveness
     (double supplyAirMassFlowRate, double exhaustAirMassFlowRate,
     double supplyAirHumidityRatio, double exhaustAirHumitidyRatio,
     double heatTransferCoefficient, AirFlow flow, out double effectiveness,
     out double mMin, out double capacityRatio)
    {
      //質量流量比[-]の計算
      mMin = Math.Min(supplyAirMassFlowRate, exhaustAirMassFlowRate);
      capacityRatio = mMin / Math.Max(supplyAirMassFlowRate, exhaustAirMassFlowRate);

      //熱通過有効度[-]の計算
      double ntu = heatTransferCoefficient / mMin;
      HeatExchange.FlowType fType;
      if (flow == AirFlow.CounterFlow) fType = HeatExchange.FlowType.CounterFlow;
      else fType = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
      effectiveness = HeatExchange.GetEffectiveness(ntu, capacityRatio, fType);
    }

    #endregion

  }
}
