﻿/* CrossFinCondensor.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>プレートフィン付管形凝縮器</summary>
  [Serializable]
  public class CrossFinCondensor : ImmutableCrossFinCondensor
  {

    #region 定数宣言

    /// <summary>大気圧[kPa]は固定:海抜0m</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>コイルの熱貫流率[kW/(m2K)]計算係数</summary>
    private const double CF_A = 0.0688;
    private const double CF_B = 0.3187;

    #endregion

    #region プロパティ

    /// <summary>水噴霧の温度低減効果率[-]</summary>
    /// <remarks>0.4～0.5程度の値。0は水噴霧無し</remarks>
    private double sprayEffectiveness = 0.0;

    /// <summary>伝熱面積[m2]を取得する</summary>
    public double SurfaceArea { get; private set; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    public double NominalAirFlowRate { get; private set; }

    /// <summary>風量[kg/s]を取得する</summary>
    public double AirFlowRate { get; private set; }

    /// <summary>凝縮温度[C]を取得する</summary>
    public double CondensingTemperature { get; private set; }

    /// <summary>入口空気乾球温度[C]を設定・取得する</summary>
    public double InletAirTemperature { get; set; }

    /// <summary>入口空気絶対温度[kg/kg]を設定・取得する</summary>
    public double InletAirHumidityRatio { get; set; }

    /// <summary>出口空気乾球温度[C]を取得する</summary>
    public double OutletAirTemperature { get; set; }

    /// <summary>出口空気絶対温度[kg/kg]を取得する</summary>
    public double OutletAirHumidityRatio { get; set; }

    /// <summary>交換熱量[kW]を取得する</summary>
    public double HeatTransfer { get; private set; }

    /// <summary>水噴霧の温度低減効果率[-]を設定・取得する</summary>
    public double SprayEffectiveness
    {
      get { return sprayEffectiveness; }
      set { sprayEffectiveness = Math.Max(0, Math.Min(value, 1.0)); }
    }

    /// <summary>水噴霧による水使用量[kg/s]を取得する</summary>
    public double WaterSupply { get; private set; }

    /// <summary>停止しているか否か</summary>
    public bool IsShutOff { get; private set; }

    /// <summary>水噴霧を用いるか否か</summary>
    public bool UseWaterSpray { get; set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="cndTemperature">凝縮温度[C]</param>
    /// <param name="heatTransfer">熱交換能力[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    public CrossFinCondensor
      (double cndTemperature, double heatTransfer, double airFlowRate,
      double inletAirTemperature, double inletAirHumidityRatio)
    {
      //プロパティ初期化
      NominalAirFlowRate = airFlowRate;
      AirFlowRate = airFlowRate;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;
      shutOff();

      //伝熱面積を初期化する
      SurfaceArea = GetSurfaceArea
        (cndTemperature, heatTransfer, airFlowRate, airFlowRate, inletAirTemperature, inletAirHumidityRatio);
    }

    /// <summary>伝熱面積[m2]を計算する</summary>
    /// <param name="cndTemperature">凝縮温度[C]</param>
    /// <param name="heatTransfer">熱交換能力[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <returns>伝熱面積[m2]</returns>
    public static double GetSurfaceArea
      (double cndTemperature, double heatTransfer, double airFlowRate,
      double nominalAirFlowRate, double inletAirTemperature, double inletAirHumidityRatio)
    {
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = cpma * airFlowRate;
      double epsilon = heatTransfer / (mca * (cndTemperature - inletAirTemperature));
      double kCnd = CF_A * Math.Pow(airFlowRate / nominalAirFlowRate, CF_B);
      return -Math.Log(1 - epsilon) * mca / kCnd;
    }

    /// <summary>機器を停止させる</summary>
    private void shutOff()
    {
      AirFlowRate = 0;
      OutletAirTemperature = InletAirTemperature;
      OutletAirHumidityRatio = InletAirHumidityRatio;
      CondensingTemperature = InletAirTemperature;
      WaterSupply = 0;
      IsShutOff = true;
    }

    /// <summary>水噴霧による水消費量[kg/s]を計算する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="sprayEffectiveness">水噴霧の温度低減効果率[-]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    private static double getWaterSupply
      (ref double inletAirTemperature, ref double inletAirHumidityRatio,
      double sprayEffectiveness, double airFlowRate)
    {
      double twb = MoistAir.GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
          (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      double ts = MoistAir.GetDryBulbTemperatureFromWetBulbTemperatureAndRelativeHumidity
        (twb, 100, ATMOSPHERIC_PRESSURE);
      double ws = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (ts, 100, ATMOSPHERIC_PRESSURE);
      inletAirTemperature -= sprayEffectiveness * (inletAirTemperature - ts);
      inletAirHumidityRatio += sprayEffectiveness * (ws - inletAirHumidityRatio);
      return airFlowRate * sprayEffectiveness * (ws - inletAirHumidityRatio);
    }

    #endregion

    #region 交換熱量計算処理

    /// <summary>交換熱量[kW]を計算する</summary>
    /// <param name="cndTemperature">凝縮温度[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <returns>交換熱量[kW]</returns>
    public double GetHeatTransfer
      (double cndTemperature, double airFlowRate, double inletAirTemperature, double inletAirHumidityRatio)
    {
      //プロパティ設定
      CondensingTemperature = cndTemperature;
      AirFlowRate = airFlowRate;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;

      //運転判定
      if (airFlowRate <= 0 || cndTemperature <= inletAirTemperature)
      {
        shutOff();
        return 0;
      }

      double ht, to, wo, ws, sp;
      if (UseWaterSpray) sp = sprayEffectiveness;
      else sp = 0;
      GetHeatTransfer(cndTemperature, airFlowRate, NominalAirFlowRate, SurfaceArea, 
        inletAirTemperature, inletAirHumidityRatio, sp, out ht, out to, out wo, out ws);
      OutletAirTemperature = to;
      OutletAirHumidityRatio = wo;
      HeatTransfer = ht;
      WaterSupply = ws;
      return ht;
    }

    /// <summary>交換熱量[kW]を計算する</summary>
    /// <param name="cndTemperature">凝縮温度[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="sprayEffectiveness">水噴霧の温度低減効果率[-]</param>
    /// <param name="heatTransfer">出力：交換熱量[kW]</param>
    /// <param name="outletAirTemperature">出力：出口空気乾球温度[C]</param>
    /// <param name="outletAirHumidityRatio">出力：出口空気絶対湿度[kg/kg]</param>
    /// <param name="waterSupply">出力：水消費量[kg/s]</param>
    public static void GetHeatTransfer
      (double cndTemperature, double airFlowRate, double nominalAirFlowRate, double surfaceArea,
      double inletAirTemperature, double inletAirHumidityRatio, double sprayEffectiveness,
      out double heatTransfer, out double outletAirTemperature,
      out double outletAirHumidityRatio, out double waterSupply)
    {
      //水噴霧がある場合
      if (0 < sprayEffectiveness)
        waterSupply = getWaterSupply
         (ref inletAirTemperature, ref inletAirHumidityRatio, sprayEffectiveness, airFlowRate);
      //水噴霧がない場合
      else waterSupply = 0;

      //熱通過率[kW/m2K]
      double kCnd = CF_A * Math.Pow(airFlowRate / nominalAirFlowRate, CF_B);
      //湿り空気比熱[kJ/kgK]
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = cpma * airFlowRate;

      double epsilon = 1 - Math.Exp(-kCnd * surfaceArea / mca);
      double q = epsilon * mca * (cndTemperature - inletAirTemperature);
      outletAirTemperature = inletAirTemperature + q / mca;
      outletAirHumidityRatio = inletAirHumidityRatio;
      heatTransfer = q;
    }

    #endregion

    #region 凝縮温度計算処理

    /// <summary>凝縮温度[C]を計算する</summary>
    /// <param name="heatTransfer">交換熱量[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <returns>凝縮温度[C]</returns>
    public double GetCondensingTemperature
      (double heatTransfer, double airFlowRate, double inletAirTemperature, double inletAirHumidityRatio)
    {
      //プロパティ設定
      HeatTransfer = heatTransfer;
      AirFlowRate = airFlowRate;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;

      //運転判定
      if (airFlowRate <= 0 || heatTransfer <= 0)
      {
        shutOff();
        return 0;
      }

      double tc, to, wo, ws;
      GetCondensingTemperature(heatTransfer, airFlowRate, NominalAirFlowRate, SurfaceArea,
        inletAirTemperature, inletAirHumidityRatio, sprayEffectiveness, out tc, out to, out wo, out ws);
      OutletAirTemperature = to;
      OutletAirHumidityRatio = wo;
      WaterSupply = ws;
      return tc;
    }

    /// <summary>凝縮温度[C]を計算する</summary>
    /// <param name="heatTransfer">交換熱量[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="sprayEffectiveness">水噴霧の温度低減効果率[-]</param>
    /// <param name="condensingTemperature">出力：凝縮温度[C]</param>
    /// <param name="outletAirTemperature">出力：出口空気乾球温度[C]</param>
    /// <param name="outletAirHumidityRatio">出力：出口空気絶対湿度[kg/kg]</param>
    /// <param name="waterSupply">出力：水消費量[kg/s]</param>
    public static void GetCondensingTemperature
      (double heatTransfer, double airFlowRate, double nominalAirFlowRate,
      double surfaceArea, double inletAirTemperature, double inletAirHumidityRatio,
      double sprayEffectiveness, out double condensingTemperature,
      out double outletAirTemperature, out double outletAirHumidityRatio,
      out double waterSupply)
    {
      //水噴霧がある場合
      if (0 < sprayEffectiveness)
        waterSupply = getWaterSupply
         (ref inletAirTemperature, ref inletAirHumidityRatio, sprayEffectiveness, airFlowRate);
      //水噴霧がない場合
      else waterSupply = 0;

      //熱通過率[kW/m2K]
      double kCnd = CF_A * Math.Pow(airFlowRate / nominalAirFlowRate, CF_B);
      //湿り空気比熱[kJ/kgK]
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = cpma * airFlowRate;

      outletAirTemperature = inletAirTemperature + heatTransfer / mca;
      outletAirHumidityRatio = inletAirHumidityRatio;
      double epsilon = 1 - Math.Exp(-kCnd * surfaceArea / mca);
      condensingTemperature = inletAirTemperature + heatTransfer / (epsilon * mca);
    }

    #endregion

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用のプレートフィン付管形凝縮器</summary>
  public interface ImmutableCrossFinCondensor
  {
    /// <summary>伝熱面積[m2]を取得する</summary>
    double SurfaceArea { get; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    double NominalAirFlowRate { get; }

    /// <summary>風量[kg/s]を取得する</summary>
    double AirFlowRate { get; }

    /// <summary>凝縮温度[C]を取得する</summary>
    double CondensingTemperature { get; }

    /// <summary>入口空気乾球温度[C]を取得する</summary>
    double InletAirTemperature { get; }

    /// <summary>入口空気絶対温度[kg/kg]を取得する</summary>
    double InletAirHumidityRatio { get; }

    /// <summary>出口空気乾球温度[C]を取得する</summary>
    double OutletAirTemperature { get; }

    /// <summary>出口空気絶対温度[kg/kg]を取得する</summary>
    double OutletAirHumidityRatio { get; }

    /// <summary>交換熱量[kW]を取得する</summary>
    double HeatTransfer { get; }

    /// <summary>水噴霧の温度低減効果率[-]を取得する</summary>
    double SprayEffectiveness { get; }

    /// <summary>水噴霧による水使用量[kg/s]を取得する</summary>
    double WaterSupply { get; }

    /// <summary>停止しているか否か</summary>
    bool IsShutOff { get; }

    /// <summary>水噴霧を用いるか否か</summary>
    bool UseWaterSpray { get; }
  }

  #endregion  

}