﻿/* CrossFinHeatExchanger.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>プレートフィン付管形熱交換器</summary>
  [Serializable]
  public class CrossFinHeatExchanger : ImmutableCrossFinHeatExchanger
  {

    #region 定数宣言

    /// <summary>湿り空気の比重量[kg/m3]</summary>
    private const double AIR_SPECIFIC_WEIGHT = 1.2;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>大気圧[kPa]は固定:海抜0m</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    #endregion

    #region 列挙型定義

    /// <summary>フロータイプ</summary>
    public enum WaterFlowType
    {
      /// <summary>ハーフフロー</summary>
      HalfFlow,
      /// <summary>シングルフロー</summary>
      SingleFlow,
      /// <summary>ダブルフロー</summary>
      DoubleFlow,
      /// <summary>トリプルフロー</summary>
      TripleFlow
    }

    #endregion

    #region インスタンス変数

    /// <summary>詳細モデルか否か</summary>
    private readonly bool isDetailedModel;

    /// <summary>詳細モデルのコイル仕様</summary>
    private readonly int rowNumber, columnNumber;
    private readonly double airWaterSurfaceRatio, coreArea, equivalentFinRadius, equivalentDiameter,
      waterPath, finThickness, thermalConductivity, innerDiameter, outerDiameter;

    /// <summary>簡易モデルのコイル仕様</summary>
    private readonly double ratedVelocity, ratedWaterSpeed;

    #endregion

    #region プロパティ

    /// <summary>乾湿境界の空気相対湿度[%]を取得する</summary>
    public double BorderRelativeHumidity { get; private set; }

    /// <summary>最大水量[kg/s]を取得する</summary>
    public double MaxWaterFlowRate { get; private set; }

    /// <summary>定格水量[kg/s]を取得する</summary>
    public double RatedWaterFlowRate { get; private set; }

    /// <summary>水量[kg/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>風量[kg/s]を取得する</summary>
    public double AirFlowRate { get; private set; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    public double RatedAirFlowRate { get; private set; }

    /// <summary>入口空気の乾球温度[C]を取得する</summary>
    public double InletAirTemperature { get; private set; }

    /// <summary>入口空気の絶対湿度[kg/kg]を取得する</summary>
    public double InletAirHumidityRatio { get; private set; }

    /// <summary>出口空気の乾球温度[C]を取得する</summary>
    public double OutletAirTemperature { get; private set; }

    /// <summary>出口空気の絶対湿度[kg/kg]を取得する</summary>
    public double OutletAirHumidityRatio { get; private set; }

    /// <summary>入口水温[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double OutletWaterTemperature { get; private set; }

    /// <summary>伝熱面積[m2]を取得する</summary>
    public double SurfaceArea { get; private set; }

    /// <summary>乾きコイルの比率[-]を取得する</summary>
    public double DryRate { get; private set; }

    /// <summary>乾きコイルの熱貫流率[kW/(m2K)]を取得する</summary>
    public double DryHeatTransferCoefficient { get; private set; }

    /// <summary>湿りコイルの熱貫流率[kW/(m2(kJ/kg))]を取得する</summary>
    public double WetHeatTransferCoefficient { get; private set; }

    /// <summary>表面積補正係数[-]を設定・取得する</summary>
    public double CorrectionFactor { get; set; }

    /// <summary>交換熱量[kW]を取得する</summary>
    public double HeatTransfer
    {
      get
      {
        return (OutletWaterTemperature - InletWaterTemperature) * WaterFlowRate * WATER_SPECIFIC_HEAT;
      }
    }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する（詳細モデル）</summary>
    /// <param name="depth">奥行き[m]</param>
    /// <param name="width">幅[m]</param>
    /// <param name="height">高さ[m]</param>
    /// <param name="rowNumber">列数[列]</param>
    /// <param name="columnNumber">段数[段]</param>
    /// <param name="finPitch">フィンピッチ[m]</param>
    /// <param name="finThickness">フィン厚み[m]</param>
    /// <param name="thermalConductivity">フィン材料の熱伝導率[W/(mK)]</param>
    /// <param name="innerDiameter">管の内径[m]</param>
    /// <param name="outerDiameter">管の外径[m]</param>
    /// <param name="ratedAirFlowRate">定格風量[kg/s]</param>
    /// <param name="ratedInletAirTemperature">定格入口空気乾球温度[C]</param>
    /// <param name="ratedInletAirHumidityRatio">定格入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="ratedWaterFlowRate">定格水量[kg/s]</param>
    /// <param name="maxWaterFlowRate">最大水量[kg/s]</param>
    /// <param name="ratedInletWaterTemperature">定格入口水温[C]</param>
    /// <param name="flowType">フロータイプ</param>
    /// <param name="heatTransfer">定格能力[kW]</param>
    /// <param name="useCorrectionFactor">修正係数を用いるか否か</param>
    public CrossFinHeatExchanger(double depth, double width, double height, int rowNumber, int columnNumber,
      double finPitch, double finThickness, double thermalConductivity, double innerDiameter,
      double outerDiameter, double ratedAirFlowRate, double ratedInletAirTemperature,
      double ratedInletAirHumidityRatio, double borderRelativeHumidity, double ratedWaterFlowRate,
      double maxWaterFlowRate, double ratedInletWaterTemperature, WaterFlowType flowType,
      double heatTransfer, bool useCorrectionFactor)
      : this(depth, width, height, rowNumber, columnNumber, finPitch, finThickness, thermalConductivity,
          innerDiameter, outerDiameter, ratedAirFlowRate, ratedInletAirTemperature,
          ratedInletAirHumidityRatio, borderRelativeHumidity, ratedWaterFlowRate, maxWaterFlowRate,
          ratedInletWaterTemperature, getFlowFactor(flowType), heatTransfer, useCorrectionFactor)
    { }

    /// <summary>インスタンスを初期化する（詳細モデル：自動初期化）</summary>
    /// <param name="width">幅[m]</param>
    /// <param name="height">高さ[m]</param>
    /// <param name="rowNumber">列数[列]</param>
    /// <param name="columnNumber">段数[段]</param>
    /// <param name="ratedAirFlowRate">定格風量[kg/s]</param>
    /// <param name="ratedInletAirTemperature">定格入口空気乾球温度[C]</param>
    /// <param name="ratedInletAirHumidityRatio">定格入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="ratedWaterFlowRate">定格水量[kg/s]</param>
    /// <param name="maxWaterFlowRate">最大水量[kg/s]</param>
    /// <param name="ratedInletWaterTemperature">定格入口水温[C]</param>
    /// <param name="flowType">フロータイプ</param>
    /// <param name="heatTransfer">定格能力[kW]</param>
    /// <param name="useCorrectionFactor">修正係数を用いるか否か</param>
    public CrossFinHeatExchanger(
      double width, double height, int rowNumber, int columnNumber, double ratedAirFlowRate,
      double ratedInletAirTemperature, double ratedInletAirHumidityRatio, double borderRelativeHumidity,
      double ratedWaterFlowRate, double maxWaterFlowRate, double ratedInletWaterTemperature,
      WaterFlowType flowType, double heatTransfer, bool useCorrectionFactor)
      : this(rowNumber * 0.0329, width, height, rowNumber, columnNumber, 0.0029, 0.0002, 237, 0.0146, 0.0158,
          ratedAirFlowRate, ratedInletAirTemperature, ratedInletAirHumidityRatio, borderRelativeHumidity,
          ratedWaterFlowRate, maxWaterFlowRate, ratedInletWaterTemperature, getFlowFactor(flowType),
          heatTransfer, useCorrectionFactor)
    { }

    /// <summary>フロータイプからフロー倍率を計算する</summary>
    /// <param name="wType">フロータイプ</param>
    /// <returns>フロー倍率</returns>
    private static double getFlowFactor(WaterFlowType wType)
    {
      if (wType == WaterFlowType.HalfFlow) return 0.5;
      else if (wType == WaterFlowType.SingleFlow) return 1.0;
      else if (wType == WaterFlowType.DoubleFlow) return 2.0;
      else return 3.0;
    }

    /// <summary>インスタンスを初期化する（詳細モデル）</summary>
    /// <param name="depth">奥行き[m]</param>
    /// <param name="width">幅[m]</param>
    /// <param name="height">高さ[m]</param>
    /// <param name="rowNumber">列数[列]</param>
    /// <param name="columnNumber">段数[段]</param>
    /// <param name="finPitch">フィンピッチ[m]</param>
    /// <param name="finThickness">フィン厚み[m]</param>
    /// <param name="thermalConductivity">フィン材料の熱伝導率[W/(mK)]</param>
    /// <param name="innerDiameter">管の内径[m]</param>
    /// <param name="outerDiameter">管の外径[m]</param>
    /// <param name="ratedAirFlowRate">定格風量[kg/s]</param>
    /// <param name="ratedInletAirTemperature">定格入口空気乾球温度[C]</param>
    /// <param name="ratedInletAirHumidityRatio">定格入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="ratedWaterFlowRate">定格水量[kg/s]</param>
    /// <param name="maxWaterFlowRate">最大水量[kg/s]</param>
    /// <param name="ratedInletWaterTemperature">定格入口水温[C]</param>
    /// <param name="flowFactor">フロー倍率</param>
    /// <param name="heatTransfer">定格能力[kW]</param>
    /// <param name="useCorrectionFactor">修正係数を用いるか否か</param>
    public CrossFinHeatExchanger(double depth, double width, double height, int rowNumber, int columnNumber,
      double finPitch, double finThickness, double thermalConductivity, double innerDiameter,
      double outerDiameter, double ratedAirFlowRate, double ratedInletAirTemperature,
      double ratedInletAirHumidityRatio, double borderRelativeHumidity, double ratedWaterFlowRate,
      double maxWaterFlowRate, double ratedInletWaterTemperature, double flowFactor, double heatTransfer,
      bool useCorrectionFactor)
    {
      //詳細モデルによる初期化
      isDetailedModel = true;

      //コイルの幾何学形状を計算
      double asr, car, eqr, eqd, asa;
      GetGeometricCompfigulation(depth, width, height, rowNumber, columnNumber, finPitch, finThickness,
        innerDiameter, outerDiameter, out asr, out car, out eqr, out eqd, out asa);

      //コイル仕様を保存
      this.airWaterSurfaceRatio = asr;
      this.coreArea = car;
      this.equivalentFinRadius = eqr;
      this.equivalentDiameter = eqd;
      this.waterPath = flowFactor * columnNumber;
      this.rowNumber = rowNumber;
      this.columnNumber = columnNumber;
      this.finThickness = finThickness;
      this.thermalConductivity = thermalConductivity;
      this.innerDiameter = innerDiameter;
      this.outerDiameter = outerDiameter;

      //その他のコイル仕様を保存
      this.RatedAirFlowRate = ratedAirFlowRate;
      this.RatedWaterFlowRate = ratedWaterFlowRate;
      this.MaxWaterFlowRate = maxWaterFlowRate;

      //乾湿境界での空気の相対湿度を保存
      this.BorderRelativeHumidity = borderRelativeHumidity;

      if (useCorrectionFactor)
      {
        //熱貫流率を計算する
        double kd, kw;
        GetHeatTransferCoefficient(airWaterSurfaceRatio, coreArea, equivalentFinRadius, equivalentDiameter,
          waterPath, finThickness, thermalConductivity, innerDiameter, outerDiameter, RatedAirFlowRate,
          ratedInletAirTemperature, ratedInletAirHumidityRatio, borderRelativeHumidity, RatedWaterFlowRate,
          ratedInletWaterTemperature, out kd, out kw);
        DryHeatTransferCoefficient = kd;
        WetHeatTransferCoefficient = kw;

        //伝熱面積[m2]を取得する
        SurfaceArea = GetSurfaceArea(ratedInletAirTemperature, ratedInletAirHumidityRatio,
          borderRelativeHumidity, ratedInletWaterTemperature, ratedAirFlowRate, ratedWaterFlowRate,
          heatTransfer, kd, kw);
        CorrectionFactor = SurfaceArea / (asa * rowNumber);
      }
      else
      {
        SurfaceArea = asa * rowNumber;
        CorrectionFactor = 1.0d;
      }
    }

    /// <summary>インスタンスを初期化する（簡易モデル）</summary>
    /// <param name="ratedAirFlowRate">定格風量[kg/s]</param>
    /// <param name="ratedVelocity">定格前面風速[m/s]</param>
    /// <param name="ratedInletAirTemperature">定格入口空気乾球温度[C]</param>
    /// <param name="ratedInletAirHumidityRatio">定格入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="ratedWaterFlowRate">定格水量[kg/s]</param>
    /// <param name="ratedWaterSpeed">定格水速[m/s]</param>
    /// <param name="maxWaterFlowRate">最大水量[kg/s]</param>
    /// <param name="ratedInletWaterTemperature">定格入口水温[C]</param>
    /// <param name="heatTransfer">定格能力[kW]</param>
    public CrossFinHeatExchanger(double ratedAirFlowRate, double ratedVelocity,
      double ratedInletAirTemperature, double ratedInletAirHumidityRatio, double borderRelativeHumidity,
      double ratedWaterFlowRate, double ratedWaterSpeed, double maxWaterFlowRate,
      double ratedInletWaterTemperature, double heatTransfer)
    {
      //簡易モデルによる初期化
      isDetailedModel = false;

      //簡易モデルのコイル仕様を保存
      this.ratedVelocity = ratedVelocity;
      this.ratedWaterSpeed = ratedWaterSpeed;

      //その他のコイル仕様を保存
      this.RatedAirFlowRate = ratedAirFlowRate;
      this.RatedWaterFlowRate = ratedWaterFlowRate;
      this.MaxWaterFlowRate = maxWaterFlowRate;

      //乾湿境界での空気の相対湿度を保存
      this.BorderRelativeHumidity = borderRelativeHumidity;

      //熱貫流率を計算する
      double kd, kw;
      GetHeatTransferCoefficient(ratedWaterSpeed, ratedVelocity, out kd, out kw);
      DryHeatTransferCoefficient = kd;
      WetHeatTransferCoefficient = kw;

      //伝熱面積[m2]を取得する
      SurfaceArea = GetSurfaceArea(ratedInletAirTemperature, ratedInletAirHumidityRatio,
        borderRelativeHumidity, ratedInletWaterTemperature, ratedAirFlowRate, ratedWaterFlowRate,
        heatTransfer, kd, kw);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>出口状態を計算する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    public void UpdateOutletState(double inletAirTemperature, double inletAirHumidityRatio,
      double inletWaterTemperature, double airFlowRate, double waterFlowRate)
    {
      //入力値を保存
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;
      InletWaterTemperature = inletWaterTemperature;
      AirFlowRate = airFlowRate;
      WaterFlowRate = waterFlowRate;

      //熱媒が流れていない場合
      if (AirFlowRate <= 0 || WaterFlowRate <= 0)
      {
        OutletAirTemperature = InletAirTemperature;
        OutletAirHumidityRatio = InletAirHumidityRatio;
        OutletWaterTemperature = InletWaterTemperature;
        DryRate = 1.0;
        return;
      }

      //熱貫流率を計算
      if (isDetailedModel)
      {
        //詳細モデル
        double kd, kw;
        GetHeatTransferCoefficient(airWaterSurfaceRatio, coreArea, equivalentFinRadius, equivalentDiameter,
          waterPath, finThickness, thermalConductivity, innerDiameter, outerDiameter, AirFlowRate,
          InletAirTemperature, InletAirHumidityRatio, BorderRelativeHumidity, WaterFlowRate,
          InletWaterTemperature, out kd, out kw);
        DryHeatTransferCoefficient = kd;
        WetHeatTransferCoefficient = kw;
      }
      else
      {
        //簡易モデル
        double kd, kw;
        //水速と風速を計算//風量と水量に比例
        double velocity = (AirFlowRate / RatedAirFlowRate) * ratedVelocity;
        double waterSpeed = (WaterFlowRate / RatedWaterFlowRate) * ratedWaterSpeed;
        GetHeatTransferCoefficient(waterSpeed, velocity, out kd, out kw);
        DryHeatTransferCoefficient = kd;
        WetHeatTransferCoefficient = kw;
      }

      //出口状態を計算
      double ta, xa, tw, dr;
      GetOutletState(InletAirTemperature, InletAirHumidityRatio, BorderRelativeHumidity,
        InletWaterTemperature, AirFlowRate, WaterFlowRate, DryHeatTransferCoefficient,
        WetHeatTransferCoefficient, SurfaceArea, out ta, out xa, out tw, out dr);
      OutletAirTemperature = ta;
      OutletAirHumidityRatio = xa;
      OutletWaterTemperature = tw;
      DryRate = dr;
    }

    /// <summary>出口空気温度を制御する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="outletAirTemperatureSetpoint">出口空気乾球温度設定値[C]</param>
    /// <returns>制御可能か否か</returns>
    public bool ControlOutletAirTemperature(double inletAirTemperature, double inletAirHumidityRatio,
      double inletWaterTemperature, double airFlowRate, double outletAirTemperatureSetpoint)
    {
      //入力値を保存
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;
      InletWaterTemperature = inletWaterTemperature;
      AirFlowRate = airFlowRate;

      //冷却・加熱の判定
      bool isCooling = (inletWaterTemperature < inletAirTemperature);

      //冷却・加熱不要の場合
      if (isCooling && inletAirTemperature < outletAirTemperatureSetpoint + 1e-3 ||
        !isCooling && outletAirTemperatureSetpoint < inletAirTemperature + 1e-3)
      {
        ShutOff();
        return false;
      }

      //最大水量で成り行き出口温度を計算
      UpdateOutletState
        (InletAirTemperature, InletAirHumidityRatio, InletWaterTemperature, AirFlowRate, MaxWaterFlowRate);

      //過負荷の場合には最大能力での成り行き状態を出力
      if ((isCooling && (outletAirTemperatureSetpoint < OutletAirTemperature))
        || (!isCooling && (OutletAirTemperature < outletAirTemperatureSetpoint)))
        return false;

      //負荷が処理可能な場合は水量をBrent法で収束計算
      //誤差関数を定義
      Roots.ErrorFunction eFnc = delegate (double wFlow)
      {
        UpdateOutletState
        (InletAirTemperature, InletAirHumidityRatio, InletWaterTemperature, AirFlowRate, wFlow);
        return outletAirTemperatureSetpoint - OutletAirTemperature;
      };
      double wf = Roots.Brent(0, MaxWaterFlowRate, MaxWaterFlowRate * 0.001, eFnc);
      UpdateOutletState(InletAirTemperature, InletAirHumidityRatio, InletWaterTemperature, AirFlowRate, wf);
      OutletAirTemperature = outletAirTemperatureSetpoint;
      return true;
    }

    /// <summary>停止させる</summary>
    public void ShutOff()
    {
      OutletAirTemperature = InletAirTemperature;
      OutletAirHumidityRatio = InletAirHumidityRatio;
      OutletWaterTemperature = InletWaterTemperature;
      WaterFlowRate = 0;
      DryRate = 1;
    }

    #endregion

    #region staticメソッド

    /// <summary>出口状態を計算する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="dryHeatTransferCoefficient">乾き部分の熱伝達率[kW/(m2K)]</param>
    /// <param name="wetHeatTransferCoefficient">湿り部分の熱伝達率[kW/(m2(kJ/kg))]</param>
    /// <param name="surfaceArea">表面積[m2]</param>
    /// <param name="outletAirTemperature">出力：出口乾球温度[C]</param>
    /// <param name="outletAirHumidityRatio">出力：出口絶対湿度[kg/kg]</param>
    /// <param name="outletWaterTemperature">出力：出口水温[C]</param>
    /// <param name="dryRate">出力：乾き部分の割合[-]</param>
    public static void GetOutletState
      (double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity,
      double inletWaterTemperature, double airFlowRate, double waterFlowRate,
      double dryHeatTransferCoefficient, double wetHeatTransferCoefficient, double surfaceArea,
      out double outletAirTemperature, out double outletAirHumidityRatio,
      out double outletWaterTemperature, out double dryRate)
    {
      //熱媒流量が0の場合は出口状態=入口状態
      if (airFlowRate <= 0 || waterFlowRate <= 0 ||
        inletWaterTemperature == inletAirTemperature)
      {
        outletAirTemperature = inletAirTemperature;
        outletAirHumidityRatio = inletAirHumidityRatio;
        outletWaterTemperature = inletWaterTemperature;
        dryRate = 1;
        return;
      }

      //水と湿り空気の熱容量流量[kW/s]の計算
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = airFlowRate * cpma;
      double mcw = waterFlowRate * WATER_SPECIFIC_HEAT;

      //加熱コイルの場合
      if (inletAirTemperature < inletWaterTemperature)
      {
        dryRate = 1.0;

        double mcMin = Math.Min(mcw, mca);
        double mcMax = Math.Max(mcw, mca);
        double ntu = dryHeatTransferCoefficient * surfaceArea / mcMin;

        //対向流の熱通過有効度[-]の計算
        double eff = HeatExchange.GetEffectiveness(ntu, mcMin / mcMax, HeatExchange.FlowType.CounterFlow);

        //交換熱量・出口状態を計算
        double q = eff * mcMin * (inletWaterTemperature - inletAirTemperature);
        outletAirTemperature = inletAirTemperature + q / mca;
        outletWaterTemperature = inletWaterTemperature - q / mcw;
        outletAirHumidityRatio = inletAirHumidityRatio;
      }
      //冷却コイルの場合
      else
      {
        //乾湿境界での空気温度[C]の計算
        double ba = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
          (inletAirHumidityRatio, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);

        //入口空気のエンタルピー[kJ/kg]の計算
        double iAirEnthalpy = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (inletAirTemperature, inletAirHumidityRatio);

        //エンタルピー近似係数の計算
        double a, b;
        getSaturationEnthalpyCoefficients(inletWaterTemperature, out a, out b);

        double xd = 1 / mca;
        double yd = -1 / mcw;
        double xw = 1 / airFlowRate;
        double yw = -a / mcw;

        double v2, v3, v4, bWaterTemp, bAirTemp;
        v2 = v3 = v4 = bWaterTemp = bAirTemp = 0;

        Roots.ErrorFunction eFnc = delegate (double dRate)
        {
          double zd = Math.Exp(dryHeatTransferCoefficient * surfaceArea * dRate * (xd + yd));
          double wd = zd * xd + yd;
          double v1 = xd * (zd - 1) / wd;
          v2 = zd * (xd + yd) / wd;

          double zw = Math.Exp(wetHeatTransferCoefficient * surfaceArea * (1 - dRate) * (xw + yw));
          double ww = zw * xw + yw;
          v3 = (xw + yw) / ww;
          v4 = xw * (zw - 1) / ww;
          double v5 = zw * (xw + yw) / ww;
          double v6 = yw * (1 - zw) / ww / a;

          //乾湿境界の水温[C]の計算
          bWaterTemp = (v5 * inletWaterTemperature
          + v6 * (iAirEnthalpy - v1 * cpma * inletAirTemperature - b)) / (1 - v1 * v6 * cpma);
          //乾湿境界での空気状態の計算
          bAirTemp = inletAirTemperature - v1 * (inletAirTemperature - bWaterTemp);

          //誤差の評価
          return ba - bAirTemp;
        };
        //結露が生じる場合には乾きコイル面積比を収束計算
        dryRate = 1.0;
        if (0 < eFnc(dryRate)) dryRate = Roots.Brent(0, 1, 0.0001, eFnc);

        //出口水温[C]の計算
        outletWaterTemperature = inletAirTemperature - v2 * (inletAirTemperature - bWaterTemp);
        double bAirEnthalpy = cpma * (bAirTemp - inletAirTemperature) + iAirEnthalpy;
        //出口空気状態の計算
        double iWaterEnthalpy = a * inletWaterTemperature + b;
        double oAirEnthalpy = v3 * bAirEnthalpy + v4 * iWaterEnthalpy;
        if (dryRate < 1.0)
          outletAirHumidityRatio = MoistAir.GetHumidityRatioFromEnthalpyAndRelativeHumidity
            (oAirEnthalpy, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        else outletAirHumidityRatio = inletAirHumidityRatio;
        outletAirTemperature = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndEnthalpy
          (outletAirHumidityRatio, oAirEnthalpy);
      }
    }

    /// <summary>必要な冷温水流量[kg/s]を計算する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="velocity">風速[m/s]</param>
    /// <param name="ratedWaterSpeed">定格水速[m/s]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="ratedWaterFlowRate">定格冷温水量[kg/s]</param>
    /// <param name="maxWaterFlowRate">最大冷温水量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="outletAirTemperatureSetpoint">出口空気乾球温度設定値[C]</param>
    /// <returns>冷温水流量[kg/s]</returns>
    public static double GetWaterFlowRate(
      double inletAirTemperature, double inletAirHumidityRatio,
      double borderRelativeHumidity, double inletWaterTemperature,
      double velocity, double ratedWaterSpeed,
      double airFlowRate, double ratedWaterFlowRate, double maxWaterFlowRate,
      double surfaceArea, double outletAirTemperatureSetpoint)
    {
      //冷却・加熱の判定
      bool isCooling = (inletWaterTemperature < inletAirTemperature);

      //冷却・加熱不要の場合
      if (isCooling && inletAirTemperature < outletAirTemperatureSetpoint
        || !isCooling && outletAirTemperatureSetpoint < inletAirTemperature) return 0;

      double wc = ratedWaterSpeed / ratedWaterFlowRate;

      //最大水量で成り行き出口温度を計算
      double oat, oah, owt, dr, kd, kw;
      GetHeatTransferCoefficient(wc * maxWaterFlowRate, velocity, out kd, out kw);
      GetOutletState(inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity,
        inletWaterTemperature, airFlowRate, maxWaterFlowRate, kd, kw, surfaceArea,
        out oat, out oah, out owt, out dr);

      //過負荷の場合には最大水量を出力
      if ((isCooling && (outletAirTemperatureSetpoint < oat))
        || (!isCooling && (oat < outletAirTemperatureSetpoint)))
        return maxWaterFlowRate;

      //負荷が処理可能な場合は水量をBrent法で収束計算
      //誤差関数を定義
      Roots.ErrorFunction eFnc = delegate (double wFlow)
      {
        GetHeatTransferCoefficient(wc * wFlow, velocity, out kd, out kw);
        GetOutletState(inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity,
          inletWaterTemperature, airFlowRate, wFlow, kd, kw, surfaceArea,
          out oat, out oah, out owt, out dr);
        return outletAirTemperatureSetpoint - oat;
      };
      return Roots.Brent(0, maxWaterFlowRate, 0.01, eFnc);
    }

    /// <summary>必要な冷温水流量[kg/s]を計算する</summary>
    /// <param name="airWaterSurfaceRatio">空気側・水側伝熱面積比[-]</param>
    /// <param name="coreArea">コア面積[m2]</param>
    /// <param name="equivalentFinRadius">環状フィンの相当半径[m]</param>
    /// <param name="equivalentDiameter">等価直径[m]</param>
    /// <param name="waterPath">配管の流路数[-]</param>
    /// <param name="finThickness">フィン厚み[m]</param>
    /// <param name="thermalConductivity">フィン材料の熱伝導率[W/(mK)]</param>
    /// <param name="innerDiameter">管の内径[m]</param>
    /// <param name="outerDiameter">管の外径[m]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="maxWaterFlowRate">最大冷温水量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="outletAirTemperatureSetpoint">出口空気乾球温度設定値[C]</param>
    /// <returns>冷温水流量[kg/s]</returns>
    public static double GetWaterFlowRate(
      double airWaterSurfaceRatio, double coreArea, double equivalentFinRadius,
      double equivalentDiameter, double waterPath, double finThickness,
      double thermalConductivity, double innerDiameter, double outerDiameter,
      double airFlowRate, double inletAirTemperature, double inletAirHumidityRatio,
      double borderRelativeHumidity, double waterFlowRate, double inletWaterTemperature,
      double maxWaterFlowRate, double surfaceArea, double outletAirTemperatureSetpoint)
    {
      //冷却・加熱の判定
      bool isCooling = (inletWaterTemperature < inletAirTemperature);

      //冷却・加熱不要の場合
      if (isCooling && inletAirTemperature < outletAirTemperatureSetpoint
        || !isCooling && outletAirTemperatureSetpoint < inletAirTemperature) return 0;

      //最大水量で成り行き出口温度を計算
      double oat, oah, owt, dr, kd, kw;
      GetHeatTransferCoefficient(airWaterSurfaceRatio, coreArea, equivalentFinRadius,
        equivalentDiameter, waterPath, finThickness, thermalConductivity,
        innerDiameter, outerDiameter, airFlowRate, inletAirHumidityRatio,
        inletAirHumidityRatio, borderRelativeHumidity, maxWaterFlowRate,
        inletWaterTemperature, out kd, out kw);
      GetOutletState(inletAirTemperature, inletAirHumidityRatio,
        borderRelativeHumidity, inletWaterTemperature,
        airFlowRate, maxWaterFlowRate, kd, kw, surfaceArea,
        out oat, out oah, out owt, out dr);

      //過負荷の場合には最大水量を出力
      if ((isCooling && (outletAirTemperatureSetpoint < oat))
        || (!isCooling && (oat < outletAirTemperatureSetpoint)))
        return maxWaterFlowRate;

      //負荷が処理可能な場合は水量をBrent法で収束計算
      //誤差関数を定義
      Roots.ErrorFunction eFnc = delegate (double wFlow)
      {
        GetHeatTransferCoefficient(airWaterSurfaceRatio, coreArea, equivalentFinRadius,
          equivalentDiameter, waterPath, finThickness, thermalConductivity,
          innerDiameter, outerDiameter, airFlowRate, inletAirHumidityRatio,
          inletAirHumidityRatio, borderRelativeHumidity, wFlow,
          inletWaterTemperature, out kd, out kw);
        GetOutletState(inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity,
          inletWaterTemperature, airFlowRate, wFlow, kd, kw, surfaceArea,
          out oat, out oah, out owt, out dr);
        return outletAirTemperatureSetpoint - oat;
      };
      return Roots.Brent(0, maxWaterFlowRate, 0.01, eFnc);
    }

    /// <summary>乾球温度[C]による飽和エンタルピー[kJ/kg]近似式の係数を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="a">乾球温度[C]の係数</param>
    /// <param name="b">切片</param>
    private static void getSaturationEnthalpyCoefficients
      (double drybulbTemperature, out double a, out double b)
    {
      const double DELTA = 0.001;
      double hws1 = MoistAir.GetSaturationEnthalpyFromDrybulbTemperature
        (drybulbTemperature, ATMOSPHERIC_PRESSURE);
      double hws2 = MoistAir.GetSaturationEnthalpyFromDrybulbTemperature
        (drybulbTemperature + DELTA, ATMOSPHERIC_PRESSURE);
      a = (hws2 - hws1) / DELTA;
      b = hws1 - a * drybulbTemperature;
    }

    /// <summary>伝熱面積[m2]を計算する</summary>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="heatTransfer">熱交換量能力[kW]</param>
    /// <param name="dryHeatTransferCoefficient">乾き部分の熱貫流率[kW/(m2K)]</param>
    /// <param name="wetHeatTransferCoefficient">湿り部分の熱貫流率[kW/(m2(kJ/kg))]</param>
    /// <returns>伝熱面積[m2]</returns>
    public static double GetSurfaceArea
      (double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity,
      double inletWaterTemperature, double airFlowRate, double waterFlowRate, double heatTransfer,
      double dryHeatTransferCoefficient, double wetHeatTransferCoefficient)
    {

      //水と湿り空気の熱容量流量[kW/s]の計算
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = airFlowRate * cpma;
      double mcw = waterFlowRate * WATER_SPECIFIC_HEAT;

      //加熱コイルの場合
      if (inletAirTemperature < inletWaterTemperature)
      {
        //NTU値の計算
        double mcMin = Math.Min(mcw, mca);
        double mcMax = Math.Max(mcw, mca);

        //熱通過有効度[-]の計算
        double eff = heatTransfer / mcMin / (inletWaterTemperature - inletAirTemperature);
        double ntu = HeatExchange.GetNTU(eff, mcMin / mcMax, HeatExchange.FlowType.CounterFlow);

        return ntu * mcMin / dryHeatTransferCoefficient;
      }
      //冷却コイルの場合
      else
      {
        //冷水出口温度[C]の計算
        double oWaterTemp = inletWaterTemperature + heatTransfer / mcw;

        //空気出入口エンタルピー[kJ/kg]の計算
        double iAirEnthalpy = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (inletAirTemperature, inletAirHumidityRatio);
        double oAirEnthalpy = iAirEnthalpy - heatTransfer / airFlowRate;

        //コイルの乾湿境界の計算
        double oAirHumidRatio = MoistAir.GetHumidityRatioFromEnthalpyAndRelativeHumidity
          (oAirEnthalpy, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);

        //乾きコイルのみの場合
        if (inletAirHumidityRatio < oAirHumidRatio)
        {
          double oAirTemp = inletAirTemperature - heatTransfer / mca;
          double d1 = inletAirTemperature - oWaterTemp;
          double d2 = oAirTemp - inletWaterTemperature;
          double lmtd = (d1 - d2) / Math.Log(d1 / d2);
          return heatTransfer / lmtd / dryHeatTransferCoefficient;
        }
        //乾き+湿りコイルの場合
        else
        {
          //境界点での湿り空気状態の計算
          double bAirTemp =
            MoistAir.GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
            (inletAirHumidityRatio, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
          double bAirEnthalpy =
            MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(bAirTemp, inletAirHumidityRatio);

          //境界点での水温[C]の計算
          double htWet = (bAirEnthalpy - oAirEnthalpy) * airFlowRate;
          double bWaterTemp = inletWaterTemperature + htWet / mcw;

          //水温と等しい温度の空気の飽和エンタルピー[kJ/(kg)]の計算
          double iWaterEnthalpy =
            MoistAir.GetSaturationEnthalpyFromDrybulbTemperature(inletWaterTemperature, ATMOSPHERIC_PRESSURE);
          double bWaterEnthalpy =
            MoistAir.GetSaturationEnthalpyFromDrybulbTemperature(bWaterTemp, ATMOSPHERIC_PRESSURE);

          //乾きコイルの表面積[m2]の計算
          double dt1 = inletAirTemperature - oWaterTemp;
          double dt2 = bAirTemp - bWaterTemp;
          double lmtd = (dt1 - dt2) / Math.Log(dt1 / dt2);
          double sd = (heatTransfer - htWet) / lmtd / dryHeatTransferCoefficient;

          double dh1 = bAirEnthalpy - bWaterEnthalpy;
          double dh2 = oAirEnthalpy - iWaterEnthalpy;
          double lmhd = (dh1 - dh2) / Math.Log(dh1 / dh2);
          double sw = htWet / lmhd / wetHeatTransferCoefficient;

          return sd + sw;
        }
      }
    }

    /// <summary>熱貫流率を計算する</summary>
    /// <param name="waterSpeed">水速[m/s]</param>
    /// <param name="velocity">前面風速[m/s]</param>
    /// <param name="dryHeatTransferCoefficient">乾き部分の熱貫流率[kW/(m2K)]</param>
    /// <param name="wetHeatTransferCoefficient">湿り部分の熱貫流率[kW/(m2(kJ/kg))]</param>
    public static void GetHeatTransferCoefficient
      (double waterSpeed, double velocity,
      out double dryHeatTransferCoefficient, out double wetHeatTransferCoefficient)
    {
      dryHeatTransferCoefficient =
        1 / (4.72 + 4.91 * Math.Pow(waterSpeed, -0.8) + 26.7 * Math.Pow(velocity, -0.64));
      wetHeatTransferCoefficient =
        1 / (10.044 + 10.44 * Math.Pow(waterSpeed, -0.8) + 39.6 * Math.Pow(velocity, -0.64));
    }

    /// <summary>幾何学形状を計算する</summary>
    /// <param name="depth">奥行き[m]</param>
    /// <param name="width">幅[m]</param>
    /// <param name="height">高さ[m]</param>
    /// <param name="rowNumber">列数[列]</param>
    /// <param name="columnNumber">段数[段]</param>
    /// <param name="finPitch">フィンピッチ[m]</param>
    /// <param name="finThickness">フィン厚み[m]</param>
    /// <param name="innerDiameter">管の内径[m]</param>
    /// <param name="outerDiameter">管の外径[m]</param>
    /// <param name="airWaterSurfaceRatio">空気側・水側伝熱面積比[-]</param>
    /// <param name="coreArea">コア面積[m2]</param>
    /// <param name="equivalentFinRadius">環状フィンの相当半径[m]</param>
    /// <param name="equivalentDiameter">等価直径[m]</param>
    /// <param name="surfaceArea">空気側伝熱面積[m2]</param>
    public static void GetGeometricCompfigulation
      (double depth, double width, double height, int rowNumber, int columnNumber,
      double finPitch, double finThickness, double innerDiameter, double outerDiameter,
      out double airWaterSurfaceRatio, out double coreArea, out double equivalentFinRadius,
      out double equivalentDiameter, out double surfaceArea)
    {
      //空気側伝熱面積[m2]の計算
      double sf = 2 * (height * depth / rowNumber 
        - outerDiameter * outerDiameter / 4 * Math.PI * columnNumber) * width / finPitch;
      double sto = outerDiameter * Math.PI * columnNumber * width * (1 - finThickness / finPitch);
      surfaceArea = sf + sto;

      //水側伝熱面積[m2]の計算
      double wSurface = innerDiameter * Math.PI * columnNumber * width;

      //空気側・水側伝熱面積比[-]の計算
      airWaterSurfaceRatio = surfaceArea / wSurface;

      //コア面積[m2]の計算
      coreArea = (width * height - outerDiameter * width * columnNumber) * (1 - finThickness / finPitch);

      //環状フィンの相当半径[m]の計算
      equivalentFinRadius = Math.Sqrt((depth / rowNumber) * (height / columnNumber) / Math.PI);

      //等価直径[m]の計算
      equivalentDiameter = 4 * coreArea / (surfaceArea * rowNumber / depth);
    }

    /// <summary>熱貫流率を計算する</summary>
    /// <param name="airWaterSurfaceRatio">空気側・水側伝熱面積比[-]</param>
    /// <param name="coreArea">コア面積[m2]</param>
    /// <param name="equivalentFinRadius">環状フィンの相当半径[m]</param>
    /// <param name="equivalentDiameter">等価直径[m]</param>
    /// <param name="waterPath">配管の流路数[-]</param>
    /// <param name="finThickness">フィン厚み[m]</param>
    /// <param name="thermalConductivity">フィン材料の熱伝導率[W/(mK)]</param>
    /// <param name="innerDiameter">管の内径[m]</param>
    /// <param name="outerDiameter">管の外径[m]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界での空気相対湿度[%]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="dryHeatTransferCoefficient">乾き部分の熱貫流率[kW/(m2K)]</param>
    /// <param name="wetHeatTransferCoefficient">湿り部分の熱貫流率[kW/(m2(kJ/kg))]</param>
    public static void GetHeatTransferCoefficient
      (double airWaterSurfaceRatio, double coreArea, double equivalentFinRadius,
      double equivalentDiameter, double waterPath, double finThickness,
      double thermalConductivity, double innerDiameter, double outerDiameter,
      double airFlowRate, double inletAirTemperature, double inletAirHumidityRatio,
      double borderRelativeHumidity, double waterFlowRate, double inletWaterTemperature,
      out double dryHeatTransferCoefficient, out double wetHeatTransferCoefficient)
    {
      //湿り空気物性の計算//比熱[kJ/kgK]・動粘性係数[m2/s]・熱伝導率[W/(mK)]
      //比体積[kg/m3]・拡散係数[m2/s]
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double dVis = MoistAir.GetDynamicViscosity
        (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      double tCond = MoistAir.GetThermalConductivity(inletAirTemperature);
      double sVol = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      double difc = MoistAir.GetThermalDiffusivity
        (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);

      //実風速の計算[m/s]
      double coreVelocity = airFlowRate * AIR_SPECIFIC_WEIGHT / coreArea;

      //レイノルズ数[-]の計算
      double re = coreVelocity * equivalentDiameter / dVis;

      //水側の対流熱伝達率[W/(m2K)]の計算
      double wfCoefficient = Circuit.WaterPipe.GetInsideHeatTransferCoefficient
        (inletWaterTemperature, innerDiameter, waterFlowRate / waterPath);

      //乾き部分の計算////
      //空気側対流熱伝達率[W/(m2K)]の計算
      double afd = 0.129 * tCond / equivalentDiameter * Math.Pow(re, 0.64);

      //フィン効率[-]の計算
      double fEfficiencyD = HeatExchange.GetCircularFinEfficiency
        (outerDiameter / 2, equivalentFinRadius, finThickness, afd, thermalConductivity);

      //熱貫流率[kW/(m2K)]の計算
      dryHeatTransferCoefficient = 0.001 / (airWaterSurfaceRatio / wfCoefficient 
        + 1 / (afd * (fEfficiencyD + 1 / airWaterSurfaceRatio)));

      //湿り部分の計算////
      //フィン表面の物質移動係数[W/(m2(kJ/kg))]の計算
      double kf = 37.2 * difc / (sVol * equivalentDiameter) * Math.Pow(re, 0.8);

      //エンタルピー近似係数の計算
      double a, b;
      getSaturationEnthalpyCoefficients(inletWaterTemperature, out a, out b);

      //フィン効率[-]の計算
      double lewis = 3.19 * Math.Pow(re, -0.16);
      double afw = a / (cpma * lewis) * afd;
      double fEfficiencyW = HeatExchange.GetCircularFinEfficiency
        (outerDiameter / 2, equivalentFinRadius, finThickness, afw, thermalConductivity);

      //熱貫流率[kW/(m2(kJ/kg))]の計算
      wetHeatTransferCoefficient = 0.001 / (a * airWaterSurfaceRatio / wfCoefficient 
        + 1 / (kf * (fEfficiencyW + 1 / airWaterSurfaceRatio)));
    }

    #endregion

  }

  #region Immutableインターフェース

  /// <summary>読み取り専用プレートフィン熱交換器</summary>
  public interface ImmutableCrossFinHeatExchanger
  {

    /// <summary>乾湿境界の空気相対湿度[%]を取得する</summary>
    double BorderRelativeHumidity { get; }

    /// <summary>最大水量[kg/s]を取得する</summary>
    double MaxWaterFlowRate { get; }

    /// <summary>定格水量[kg/s]を取得する</summary>
    double RatedWaterFlowRate { get; }

    /// <summary>水量[kg/s]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>風量[kg/s]を取得する</summary>
    double AirFlowRate { get; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    double RatedAirFlowRate { get; }

    /// <summary>入口空気の乾球温度[C]を取得する</summary>
    double InletAirTemperature { get; }

    /// <summary>入口空気の絶対湿度[kg/kg]を取得する</summary>
    double InletAirHumidityRatio { get; }

    /// <summary>出口空気の乾球温度[C]を取得する</summary>
    double OutletAirTemperature { get; }

    /// <summary>出口空気の絶対湿度[kg/kg]を取得する</summary>
    double OutletAirHumidityRatio { get; }

    /// <summary>入口水温[C]を取得する</summary>
    double InletWaterTemperature { get; }

    /// <summary>出口水温[C]を取得する</summary>
    double OutletWaterTemperature { get; }

    /// <summary>伝熱面積[m2]を取得する</summary>
    double SurfaceArea { get; }

    /// <summary>乾きコイルの比率[-]を取得する</summary>
    double DryRate { get; }

    /// <summary>乾きコイルの熱貫流率[kW/(m2K)]を取得する</summary>
    double DryHeatTransferCoefficient { get; }

    /// <summary>湿りコイルの熱貫流率[kW/(m2(kJ/kg))]を取得する</summary>
    double WetHeatTransferCoefficient { get; }

    /// <summary>表面積補正係数[-]を取得する</summary>
    double CorrectionFactor { get; }

    /// <summary>交換熱量[kW]を取得する</summary>
    double HeatTransfer { get; }
  }

  #endregion

}
