﻿/* HeatExchange.cs
 * 
 * Copyright (C) 2013 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>熱交換に関する静的メソッドを提供する</summary>
  public static class HeatExchange
  {

    #region 列挙型定義

    /// <summary>流れのタイプ</summary>
    public enum FlowType
    {
      /// <summary>向流型</summary>
      CounterFlow,
      /// <summary>並流型</summary>
      ParallelFlow,
      /// <summary>直交流型（両流体混合）</summary>
      CrossFlow_BothFluidsUnmixed,
      /// <summary>直交流型（熱容量流量大の流体が混合）</summary>
      CrossFlow_CmaxMixed,
      /// <summary>直交流型（熱容量流量小の流体が混合）</summary>
      CrossFlow_CminMixed,
      /// <summary>直交流型（両流体非混合）</summary>
      CrossFlow_BothFluidMixed
    }

    #endregion

    #region 熱通過有効度の計算

    /// <summary>熱通過有効度を計算する</summary>
    /// <param name="ntu">移動単位数NTU[-]</param>
    /// <param name="heatCapacityRatio">熱容量流量比[-]</param>
    /// <param name="flowType">流れタイプ</param>
    /// <returns>熱通過有効度[-]</returns>
    public static double GetEffectiveness(double ntu, double heatCapacityRatio, FlowType flowType)
    {
      double rMC = heatCapacityRatio;

      //熱容量流量比が0の場合
      if (rMC <= 0) return 1 - Math.Exp(-ntu);

      //NTUが0の場合
      if (ntu <= 0) return 0;

      double eps;
      switch (flowType)
      {
        case FlowType.CounterFlow:
          if (rMC < 1) return (1 - Math.Exp((rMC - 1) * ntu)) / (1 - rMC * Math.Exp((rMC - 1) * ntu));
          else return ntu / (1 + ntu);

        case FlowType.ParallelFlow:
          if (rMC < 1) return (1 - Math.Exp(-ntu * (rMC + 1))) / (1 + rMC);
          else return 0.5 * (1 - Math.Exp(-2 * ntu));

        case FlowType.CrossFlow_BothFluidMixed:
          eps = ntu / (1 - Math.Exp(-ntu)) + rMC * ntu / (1 - Math.Exp(-rMC * ntu));
          return ntu / (eps - 1);

        case FlowType.CrossFlow_CminMixed:
          return 1 - Math.Exp((Math.Exp(-ntu * rMC) - 1) / rMC);

        case FlowType.CrossFlow_CmaxMixed:
          return (1 - Math.Exp((Math.Exp(-ntu) - 1) * rMC)) / rMC;

        case FlowType.CrossFlow_BothFluidsUnmixed:
          eps = (Math.Exp(-rMC * Math.Pow(ntu, 0.78)) - 1) / (rMC * Math.Pow(ntu, -0.22));
          return 1 - Math.Exp(eps);
      }
      return 0;
    }

    /// <summary>移動単位数NTU[-]を計算する</summary>
    /// <param name="effectiveness">熱通過有効度[-]</param>
    /// <param name="heatCapacityRatio">熱容量流量比[-]</param>
    /// <param name="flowType">流れのタイプ</param>
    /// <returns>移動単位数NTU[-]</returns>
    public static double GetNTU(double effectiveness, double heatCapacityRatio, FlowType flowType)
    {
      Roots.ErrorFunction eFnc = delegate (double ntu)
      { return effectiveness - GetEffectiveness(ntu, heatCapacityRatio, flowType); };
      return Roots.Newton(eFnc, 0, 1e-4, 1e-6, 1e-6, 20);
    }

    #endregion

    #region 平均温度差Tmの計算

    /// <summary>出入口流体温度にもとづいて平均温度差[C]を計算する</summary>
    /// <param name="hotInletTemperature">高温流体入口温度[C]</param>
    /// <param name="coldInletTemperature">低温流体入口温度[C]</param>
    /// <param name="hotOutletTemperature">高温流体出口温度[C]</param>
    /// <param name="coldOutletTemperature">低温流体出口温度[C]</param>
    /// <param name="flowType">流れのタイプ</param>
    /// <returns>平均温度差[C]</returns>
    public static double GetMeanTemperatureDifference
      (double hotInletTemperature, double coldInletTemperature,
      double hotOutletTemperature, double coldOutletTemperature, FlowType flowType)
    {
      //向流の場合
      if (flowType == FlowType.CounterFlow)
      {
        double dt1 = hotInletTemperature - coldOutletTemperature;
        double dt2 = hotOutletTemperature - coldInletTemperature;
        if (dt1 == dt2) return dt1;
        if (dt1 <= 0 || dt2 <= 0) return 0;
        else return (dt1 - dt2) / Math.Log(dt1 / dt2);
      }
      //並流の場合
      else if (flowType == FlowType.ParallelFlow)
      {
        double dt1 = hotInletTemperature - coldInletTemperature;
        double dt2 = hotOutletTemperature - coldOutletTemperature;
        if (dt1 <= 0 || dt2 <= 0) return 0;
        return (dt1 - dt2) / Math.Log(dt1 / dt2);
      }
      else
      {
        double p =
          (hotInletTemperature - hotOutletTemperature) / (hotInletTemperature - coldInletTemperature);
        double q =
          (coldOutletTemperature - coldInletTemperature) / (hotInletTemperature - coldInletTemperature);
        double r = 0;

        //直交流（片側混合）の場合
        if (flowType == FlowType.CrossFlow_CminMixed || flowType == FlowType.CrossFlow_CmaxMixed)
        {
          double bf = 1 - q / p * Math.Log(1 / (1 - p));
          if (bf <= 0) r = 0;
          else r = q / Math.Log(1 / bf);
        }
        //直交流（両側混合・非混合）の場合
        else if (flowType == FlowType.CrossFlow_BothFluidMixed
          || flowType == FlowType.CrossFlow_BothFluidsUnmixed)
        {
          //rの上下限値を計算して解の範囲を特定
          double rMax;  //上限値（対向流のr）
          if (Math.Abs(p - q) < 1e-8) rMax = 1;
          else rMax = (p - q) / Math.Log((1 - q) / (1 - p));
          r = Math.Max(1e-4, 1 - (p + q));  //下限値（出口温度差）

          //極小値を黄金探索
          Minimization.MinimizeFunction mFnc = delegate (double x)
          {
            if (flowType == FlowType.CrossFlow_BothFluidMixed) return Math.Abs(eFncBothMixedR(p, q, x));
            else return Math.Abs(eFncBothUnMixedR(p, q, x));
          };
          if (1e-3 < Minimization.GoldenSection(ref r, rMax, mFnc)) r = 0;
        }
        return r * (hotInletTemperature - coldInletTemperature);
      }
      throw new Exception("Not implemented");
    }

    /// <summary>両流体混合の場合の誤差関数</summary>
    /// <param name="p">p</param>
    /// <param name="q">q</param>
    /// <param name="r">r</param>
    /// <returns>誤差</returns>
    private static double eFncBothMixedR(double p, double q, double r)
    { return (p / (1 - Math.Exp(-p / r)) + q / (1 - Math.Exp(-q / r)) - 1) - r; }

    /// <summary>両流体非混合の場合の誤差関数</summary>
    /// <param name="p">p</param>
    /// <param name="q">q</param>
    /// <param name="r">r</param>
    /// <returns>誤差</returns>
    private static double eFncBothUnMixedR(double p, double q, double r)
    {
      const int UVMAX = 50;
      double rSum = 1;
      double pr = p / r;
      double qr = q / r;
      for (int u = 1; u <= UVMAX; u++)
      {
        double dR = 0;
        double pqr = 1;
        for (int i = 1; i <= u; i++) pqr *= pr / (i + 1);
        if (u % 2 == 0) dR += pqr;
        else dR -= pqr;

        for (int v = 1; v <= u; v++)
        {
          pqr *= qr * (u + v) / (v * (v + 1));
          if ((u + v) % 2 == 0) dR += pqr;
          else dR -= pqr;
        }

        for (int i = u; 0 < i; i--)
        {
          pqr *= i * (i + 1) / (pr * (u + i));
          if ((i + u) % 2 == 0) dR -= pqr;
          else dR += pqr;
        }

        rSum += dR;
        if (Math.Abs(dR) < 1e-6) return rSum - r;
      }
      return (1 - (p + q)) - r;
    }

    #endregion

    #region フィン効率の計算

    /// <summary>環状フィンのフィン効率[-]を計算する</summary>
    /// <param name="tubeRadius">管の半径[m]</param>
    /// <param name="finRadius">フィンの半径[m]</param>
    /// <param name="thickness">フィンの厚み[m]</param>
    /// <param name="filmCoefficient">フィンの対流熱伝導率[W/(m2K)]</param>
    /// <param name="thermalConductivity">フィン材料の熱伝導率[W/(mK)]</param>
    /// <returns>フィン効率[-]</returns>
    public static double GetCircularFinEfficiency
      (double tubeRadius, double finRadius, double thickness, 
      double filmCoefficient, double thermalConductivity)
    {
      double w = finRadius - tubeRadius;
      double xexb = finRadius / tubeRadius;

      double ws = w * Math.Sqrt(filmCoefficient / thermalConductivity / thickness);
      //多くの図表はwsを横軸、フィン効率を縦軸に取る
      double ub = ws / (xexb - 1);
      double ue = ub * xexb;

      //ベッセル関数の計算
      double i0ub = besI0(ub);
      double i1ue = besI1(ue);
      double i1ub = besI1(ub);
      double k0ub = besK0(ub, i0ub);
      double k1ue = besK1(ue, i1ue);
      double k1ub = besK1(ub, i1ub);

      double phi = (k1ue * i1ub - i1ue * k1ub) / (k1ue * i0ub + i1ue * k0ub);
      return 2 * phi / ub / (1 - Math.Pow(ue / ub, 2));
    }

    #endregion

    #region ベッセル関数の計算

    /// <summary>ベッセル関数I0(u)を計算する</summary>
    /// <param name="u">引数</param>
    /// <returns>ベッセル関数I0(u)</returns>
    private static double besI0(double u)
    {

      if (u < 3.75)
      {
        double uu = u / 3.75;
        uu *= uu;
        return 1 + uu * (3.5156229 + uu * (3.0899424 + uu * (1.2067492 + uu * (0.2659732
          + uu * (0.0360768 + uu * 0.0045813)))));
      }
      else
      {
        double uu = 3.75 / u;
        return (0.398942280 + uu * (0.013285917 + uu * (0.002253187 + uu * (-0.001575649 
          + uu * (0.009162808 + uu * (-0.020577063 + uu * (0.026355372 + uu * (-0.016476329 
          + uu * 0.003923767)))))))) * (Math.Exp(u) / Math.Sqrt(u));
      }
    }

    /// <summary>ベッセル関数I1(u)を計算する</summary>
    /// <param name="u">引数</param>
    /// <returns>ベッセル関数I1(u)</returns>
    private static double besI1(double u)
    {

      if (u < 3.75)
      {
        double uu = u / 3.75;
        uu *= uu;
        return (0.5 + uu * (0.87890594 + uu * (0.51498869 + uu * (0.15084934 
          + uu * (0.02658733 + uu * (0.00301532 + uu * 0.00032411)))))) * u;
      }
      else
      {
        double uu = 3.75 / u;
        return (0.398942280 + uu * (-0.039880242 + uu * (-0.003620183 + uu * (0.001638014 
          + uu * (-0.010315550 + uu * (0.022829673 + uu * (-0.028953121 + uu * (0.017876535
          + uu * (-0.004200587))))))))) * (Math.Exp(u) / Math.Sqrt(u));
      }
    }

    /// <summary>ベッセル関数K0(u)を計算する</summary>
    /// <param name="u">引数</param>
    /// <param name="i0">ベッセル関数I0(u)の出力</param>
    /// <returns>ベッセル関数K0(u)</returns>
    private static double besK0(double u, double i0)
    {
      if (u < 2)
      {
        double uu = u / 2;
        uu *= uu;
        return -0.57721566 + uu * (0.42278420 + uu * (0.23069756 + uu * (0.03488590
          + uu * (0.00262698 + uu * (0.00010750 + uu * 0.00000740))))) - Math.Log(0.5 * u) * i0;
      }
      else
      {
        double uu = 2 / u;
        return (1.25331414 + uu * (-0.07832358 + uu * (0.02189568 + uu * (-0.01062446 + uu 
          * (0.00587872 + uu * (-0.00251540 + uu * 0.00053208)))))) / (Math.Exp(u) * Math.Sqrt(u));
      }
    }

    /// <summary>ベッセル関数K0(u)を計算する</summary>
    /// <param name="u">引数</param>
    /// <param name="i1">ベッセル関数I1(u)の出力</param>
    /// <returns>ベッセル関数K0(u)</returns>
    private static double besK1(double u, double i1)
    {
      if (u < 2)
      {
        double uu = u / 2;
        uu *= uu;
        return (1 + uu * (0.15443144 + uu * (-0.67278579 + uu * (-0.18156897 + uu
          * (-0.01919402 + uu * (-0.00110404 + uu * (-0.00004686))))))) / u + Math.Log(0.5 * u) * i1;
      }
      else
      {
        double uu = 2 / u;
        return (1.25331414 + uu * (0.23498619 + uu * (-0.03655620 + uu * (0.01504268 + uu 
          * (-0.00780353 + uu * (0.00325614 + uu * (-0.00068245))))))) / (Math.Exp(u) * Math.Sqrt(u));
      }
    }

    #endregion

    #region 熱交換量の計算

    /// <summary>熱交換量[kW]を計算する</summary>
    /// <param name="highTemperature">高温側温度[C]</param>
    /// <param name="lowTemperature">低温側温度[C]</param>
    /// <param name="highHeatCapFlow">高温側熱容量流量[kW/K]</param>
    /// <param name="lowHeatCapFlow">低温側熱容量流量[kW/K]</param>
    /// <param name="heatTransferCoefficient">伝熱係数[kW/K]</param>
    /// <param name="fType">フロータイプ</param>
    /// <returns>熱交換量[kW]</returns>
    public static double GetHeatTransfer(double highTemperature, double lowTemperature, 
      double highHeatCapFlow, double lowHeatCapFlow, double heatTransferCoefficient, FlowType fType)
    {
      double mcMin = Math.Min(highHeatCapFlow, lowHeatCapFlow);
      double mcMax = Math.Max(highHeatCapFlow, lowHeatCapFlow);
      double effectiveness = GetEffectiveness(heatTransferCoefficient / mcMin, mcMin / mcMax, fType);
      return effectiveness * mcMin * (highTemperature - lowTemperature);
    }

    /// <summary>伝熱係数[kW/K]を計算する</summary>
    /// <param name="highTemperature">高温側温度[C]</param>
    /// <param name="lowTemperature">低温側温度[C]</param>
    /// <param name="highHeatCapFlow">高温側熱容量流量[kW/K]</param>
    /// <param name="lowHeatCapFlow">低温側熱容量流量[kW/K]</param>
    /// <param name="heatTransfer">熱交換量[kW]</param>
    /// <param name="fType">フロータイプ</param>
    /// <returns>伝熱係数[kW/K]</returns>
    public static double GetHeatTransferCoefficient(double highTemperature, double lowTemperature,
      double highHeatCapFlow, double lowHeatCapFlow, double heatTransfer, FlowType fType)
    {
      double mcMin = Math.Min(highHeatCapFlow, lowHeatCapFlow);
      double mcMax = Math.Max(highHeatCapFlow, lowHeatCapFlow);
      double effectiveness = Math.Min(0.9999, heatTransfer / (mcMin * (highTemperature - lowTemperature)));
      return HeatExchange.GetNTU(effectiveness, mcMin / mcMax, HeatExchange.FlowType.CounterFlow) * mcMin;
    }

    #endregion

  }
}
