﻿/* AirHeatSourceHeatPump.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>空気熱源モジュールヒートポンプ</summary>
  [Serializable]
  public class AirHeatSourceModularChillers: ImmutableAirHeatSourceModularChillers
  {

    #region 定数宣言

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region 列挙型定義

    /// <summary>運転モード</summary>
    public enum OperatingMode
    {
      /// <summary>加熱運転</summary>
      Heating,
      /// <summary>冷却運転</summary>
      Cooling,
      /// <summary>停止</summary>
      ShutOff
    }

    #endregion

    #region 特性係数

    /// <summary>最大能力比近似係数（冷房運転）</summary>
    private readonly double[] aMax_C = new double[] { -4.1557e-04, 1.0518e-01, 1.5448e-01, -3.8823e01 };

    /// <summary>最大能力比近似係数（暖房運転）</summary>
    private readonly double[] aMax_H = new double[] { 9.2525e-06, 1.7117e-02, -6.7223e-03, -2.4744 };

    /// <summary>COP比近似係数（冷房運転）</summary>
    private readonly double[] aCop_C = new double[] { 0.60824, -2.4486, 2.8067 };

    /// <summary>COP比近似係数（暖房運転）</summary>
    private readonly double[] aCop_H = new double[] { 0.088163, -1.3885, 2.2981 };

    /// <summary>部分負荷特性係数</summary>
    private readonly double[] a_PL = new double[] { -1.4118, 1.3611, 1.0507 };

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>定格理論COP[-]</summary>
    private readonly double copFLRT_C, copFLRT_H;

    /// <summary>風量[(kg/s)/台]</summary>
    private double coolingAirFlowRate, heatingAirFlowRate;

    /// <summary>補機消費電力[kW]</summary>
    private double auxElec;
    
    /// <summary>運転モードを設定・取得する</summary>
    public OperatingMode Mode { get; set; }

    /// <summary>運転効率[-]を最大化するか否か</summary>
    public bool MaximizeEfficiency { get; set; } = true;

    /// <summary>モジュール数[台]を取得する</summary>
    public int NumberOfUnits { get; private set; }

    /// <summary>稼働台数[台]を取得するs</summary>
    public int OperatingNumber { get; private set; }

    /// <summary>ヒートポンプ機か否かを取得する</summary>
    public bool IsHeatPumpModel { get; private set; }

    /// <summary>定格冷却能力[kW/台]を取得する</summary>
    public double NominalCoolingCapacity { get; private set; }

    /// <summary>定格冷却COP[-]を取得する</summary>
    public double NominalCoolingCOP { get; private set; }

    /// <summary>定格加熱能力[kW/台]を取得する</summary>
    public double NominalHeatingCapacity { get; private set; }

    /// <summary>定格加熱COP[-]を取得する</summary>
    public double NominalHeatingCOP { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double WaterOutletTemperature { get; private set; }

    /// <summary>出口水温設定値[C]を設定・取得する</summary>
    public double WaterOutletSetPointTemperature { get; set; }

    /// <summary>入口水温[C]を取得する</summary>
    public double WaterInletTemperature { get; private set; }

    /// <summary>水流量[(kg/s)/台]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>風量[(kg/s)/台]を取得する</summary>
    public double AirFlowRate
    {
      get
      {
        if (Mode == OperatingMode.Cooling) return coolingAirFlowRate;
        else if (Mode == OperatingMode.Heating) return heatingAirFlowRate;
        else return 0;
      }
    }

    /// <summary>消費電力[kW/台]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>補機消費電力[kW/台]を取得する</summary>
    public double AuxiliaryElectricConsumption { get; private set; }

    /// <summary>周囲の空気乾球温度[C]を取得する</summary>
    public double AmbientTemperature { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (ElectricConsumption != 0)
          return Math.Abs(WaterOutletTemperature - WaterInletTemperature)
            * WATER_SPECIFIC_HEAT * WaterFlowRate / (ElectricConsumption * OperatingNumber);
        else return 0;
      }
    }

    /// <summary>冷却量[kW]を取得する</summary>
    public double CoolingLoad
    {
      get
      {
        if (Mode == OperatingMode.Cooling)
          return (WaterInletTemperature - WaterOutletTemperature) * WATER_SPECIFIC_HEAT * WaterFlowRate;
        else return 0;
      }
    }

    /// <summary>加熱量[kW]を取得する</summary>
    public double HeatingLoad
    {
      get
      {
        if (Mode == OperatingMode.Heating)
          return (WaterOutletTemperature - WaterInletTemperature) * WATER_SPECIFIC_HEAT * WaterFlowRate;
        else return 0;
      }
    }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate { get; private set; }

    /// <summary>冷水下限流量[kg/s]を取得する</summary>
    public double MinChilledWaterFlowRate { get; private set; }

    /// <summary>温水下限流量[kg/s]を取得する</summary>
    public double MinHotWaterFlowRate { get; private set; }

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="coolingCapacity">冷却能力[kW/台]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水流量[(kg/s)/台]</param>
    /// <param name="coolingAirTemperature">冷却時空気温度[C]</param>
    /// <param name="coolingAirFlowRate">冷却時風量[(kg/s)/台]</param>
    /// <param name="coolingElectricity">冷却時消費電力[kW/台]</param>
    /// <param name="heatingCapacity">加熱能力[kW/台]</param>
    /// <param name="hotWaterOutletTemperature">温水出口温度[C]</param>
    /// <param name="hotWaterFlowRate">温水流量[(kg/s)/台]</param>
    /// <param name="heatingAirTemperature">加熱時空気温度[C]</param>
    /// <param name="heatingAirFlowRate">加熱時風量[(kg/s)/台]</param>
    /// <param name="heatingElectricity">加熱時消費電力[kW/台]</param>
    /// <param name="numberOfUnits">モジュール数[台]</param>
    /// <param name="auxiliaryElectricConsumption">補機消費電力[kW/台]</param>
    public AirHeatSourceModularChillers(
      double coolingCapacity, double chilledWaterOutletTemperature, double chilledWaterFlowRate, 
      double coolingAirTemperature, double coolingAirFlowRate, double coolingElectricity,
      double heatingCapacity, double hotWaterOutletTemperature, double hotWaterFlowRate, 
      double heatingAirTemperature, double heatingAirFlowRate, double heatingElectricity,
      int numberOfUnits, double auxiliaryElectricConsumption)
    {
      IsHeatPumpModel = true;

      this.coolingAirFlowRate = coolingAirFlowRate;
      this.heatingAirFlowRate = heatingAirFlowRate;
      this.NominalCoolingCapacity = coolingCapacity;
      this.NominalHeatingCapacity = heatingCapacity;
      this.NumberOfUnits = numberOfUnits;
      this.auxElec = auxiliaryElectricConsumption;
      this.MaxChilledWaterFlowRate = chilledWaterFlowRate * numberOfUnits;
      this.MaxHotWaterFlowRate = hotWaterFlowRate * numberOfUnits;
      this.MinChilledWaterFlowRate = chilledWaterFlowRate * 0.4;
      this.MinHotWaterFlowRate = hotWaterFlowRate * 0.4;

      //冷房運転COPの計算
      double mcw = WATER_SPECIFIC_HEAT * chilledWaterFlowRate;
      double mcma = (1.005 + 1.846 * 0.020) * coolingAirFlowRate;
      double tao = coolingAirTemperature + (coolingElectricity + coolingCapacity) / mcma;
      copFLRT_C = (chilledWaterOutletTemperature + 273.15) / (tao - chilledWaterOutletTemperature);
      NominalCoolingCOP = coolingCapacity / coolingElectricity;

      //暖房運転COPの計算
      mcw = WATER_SPECIFIC_HEAT * hotWaterFlowRate;
      mcma = (1.005 + 1.846 * 0.002) * heatingAirFlowRate;
      tao = heatingAirTemperature + (heatingElectricity - heatingCapacity) / mcma;
      copFLRT_H = (hotWaterOutletTemperature + 273.15) / (hotWaterOutletTemperature - tao);
      NominalHeatingCOP = heatingCapacity / heatingElectricity;

      //停止させる
      ShutOff();
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="coolingCapacity">冷却能力[kW/台]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水流量[(kg/s)/台]</param>
    /// <param name="coolingAirTemperature">冷却時空気温度[C]</param>
    /// <param name="coolingAirFlowRate">冷却時風量[(kg/s)/台]</param>
    /// <param name="coolingElectricity">冷却時消費電力[kW/台]</param>
    /// <param name="numberOfUnits">モジュール数[台]</param>
    /// <param name="auxiliaryElectricConsumption">補機消費電力[kW/台]</param>
    public AirHeatSourceModularChillers(
      double coolingCapacity, double chilledWaterOutletTemperature, double chilledWaterFlowRate,
      double coolingAirTemperature, double coolingAirFlowRate, double coolingElectricity,
      int numberOfUnits, double auxiliaryElectricConsumption)
    {
      IsHeatPumpModel = false;

      this.coolingAirFlowRate = coolingAirFlowRate;
      this.NominalCoolingCapacity = coolingCapacity;      
      this.NumberOfUnits = numberOfUnits;
      this.auxElec = auxiliaryElectricConsumption;
      this.heatingAirFlowRate = this.NominalHeatingCapacity = 0;

      //冷房運転COPの計算
      double mcw = WATER_SPECIFIC_HEAT * chilledWaterFlowRate;
      double mcma = (1.005 + 1.846 * 0.020) * coolingAirFlowRate;
      double tao = coolingAirTemperature + (coolingElectricity + coolingCapacity) / mcma;
      copFLRT_C = (chilledWaterOutletTemperature + 273.15) / (tao - chilledWaterOutletTemperature);
      NominalCoolingCOP = coolingCapacity / coolingElectricity;

      //停止させる
      shutOff();
    }

    #endregion

    #region 機器の停止処理

    /// <summary>機器を停止させる</summary>
    public void ShutOff()
    {
      Mode = OperatingMode.ShutOff;
      shutOff();
    }

    /// <summary>機器を停止させる</summary>
    private void shutOff()
    {
      WaterOutletTemperature = WaterInletTemperature;
      ElectricConsumption = 0;
      AuxiliaryElectricConsumption = 0;
      OperatingNumber = 0;
    }

    #endregion

    #region 最大能力計算処理

    /// <summary>最大能力[kW/台]を計算する</summary>
    /// <param name="waterOutletTemperature">出口水温[C]</param>
    /// <param name="ambientTemperature">空気温度[C]</param>
    /// <returns>最大能力[kW/台]</returns>
    public double GetMaxCapacity(double waterOutletTemperature, double ambientTemperature)
    {
      if (Mode == OperatingMode.ShutOff) return 0;
      else if (Mode == OperatingMode.Cooling)
      {
        double two = Math.Min(15, Math.Max(3, waterOutletTemperature)) + 273.15;
        double tai = Math.Min(43, Math.Max(20, ambientTemperature)) + 273.15;
        return NominalCoolingCapacity * (tai * (two * aMax_C[0] + aMax_C[1]) + two * aMax_C[2] + aMax_C[3]);
      }
      else
      {
        double two = Math.Min(55, Math.Max(35, waterOutletTemperature)) + 273.15;
        double tai = Math.Min(21, Math.Max(-15, ambientTemperature)) + 273.15;
        return NominalHeatingCapacity * (tai * (two * aMax_H[0] + aMax_H[1]) + two * aMax_H[2] + aMax_H[3]);
      }
    }

    #endregion

    #region 運転台数計算処理

    /// <summary>運転台数[台]を計算する</summary>
    /// <param name="load">要求負荷[kW]</param>
    /// <param name="mCap">最大能力[kW/台]</param>
    /// <returns>運転台数[台]</returns>
    private int getOperatingNumber(double load, double mCap)
    {
      //部分負荷運転により機器効率を最大化させる場合
      if (MaximizeEfficiency)
      {
        double optCap = mCap * (-0.5 * a_PL[1] / a_PL[0]);
        int optNum = (int)Math.Floor(load / optCap);
        if (NumberOfUnits <= optNum) return NumberOfUnits;
        else
        {
          double plf1 = getPartialLoadFactor(load / optNum);
          double plf2 = getPartialLoadFactor(load / (optNum + 1));
          if (plf1 < plf2) return (optNum + 1);
          else return optNum;
        }
      }
      //最大負荷で運転時間を最小化させる場合（機器寿命重視）
      else return (int)Math.Min(Math.Ceiling(load / mCap), NumberOfUnits);
    }

    /// <summary>部分負荷効率[-]を計算する</summary>
    /// <param name="partialLoad">部分負荷率[-]</param>
    /// <returns>部分負荷効率[-]</returns>
    private double getPartialLoadFactor(double partialLoad)
    {
      double pl = Math.Max(0.2, partialLoad);
      double plf = pl * (a_PL[0] * pl + a_PL[1]) + a_PL[2];
      if (partialLoad < 0.2) return plf * (partialLoad / 0.2);
      else return plf;
    }

    #endregion

    #region 状態更新処理

    /// <summary>状態を更新する</summary>
    /// <param name="waterInletTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    public void Update(double waterInletTemperature, double waterFlowRate, double ambientTemperature)
    {
      //状態値を保存
      this.WaterInletTemperature = waterInletTemperature;
      this.WaterFlowRate = waterFlowRate;
      this.AmbientTemperature = ambientTemperature;

      //必要能力を計算
      double load = 0;  //ShutOffの場合には0になる
      double mcw = WaterFlowRate * WATER_SPECIFIC_HEAT;
      if (Mode == OperatingMode.Cooling)
        load = mcw * (WaterInletTemperature - WaterOutletSetPointTemperature);
      else if (Mode == OperatingMode.Heating && IsHeatPumpModel)
        load = mcw * (WaterOutletSetPointTemperature - WaterInletTemperature);

      //負荷0ならば停止処理
      if (load <= 0)
      {
        shutOff();
        return;
      }

      //過負荷判定//最適運転台数を計算する
      double cap = GetMaxCapacity(WaterOutletSetPointTemperature, AmbientTemperature);
      OperatingNumber = getOperatingNumber(load, cap);
      double qLD = load / OperatingNumber;
      IsOverLoad = cap < qLD;
      if (IsOverLoad) qLD = cap;

      //冷暖切替係数
      double sg, copFLRT, copFLR, mcma;
      double[] aCOP;
      if (Mode == OperatingMode.Cooling)
      {
        sg = -1;
        copFLRT = copFLRT_C;
        copFLR = NominalCoolingCOP;
        aCOP = aCop_C;
        mcma = AirFlowRate * (1.005 + 1.846 * 0.020); //夏季絶対湿度は20g/kgとする
      }
      else
      {
        sg = 1;
        copFLRT = copFLRT_H;
        copFLR = NominalHeatingCOP;
        aCOP = aCop_H;
        mcma = AirFlowRate * (1.005 + 1.846 * 0.002); //冬季絶対湿度は2g/kgとする
      }

      //3次方程式の係数計算
      WaterOutletTemperature = WaterInletTemperature + sg * qLD * OperatingNumber / mcw;
      double abf0 = copFLRT / (WaterOutletTemperature + 273.15);
      double abf1 = 2 * aCOP[0] * copFLRT + sg * aCOP[1];
      double abf2 = copFLRT * (aCOP[0] * copFLRT + sg * aCOP[1]) + aCOP[2];
      double abf3 = mcma * (AmbientTemperature + 273.15) - sg * qLD;
      double[] aTau = new double[4];
      aTau[0] = abf0 * abf0 * aCOP[0] * mcma;
      aTau[1] = -abf0 * (mcma * abf1 + abf0 * aCOP[0] * abf3);
      aTau[2] = abf0 * abf1 * abf3 + mcma * abf2;
      aTau[3] = - abf2 * abf3 - qLD / (getPartialLoadFactor(qLD / cap) * NominalCoolingCOP);

      double x1, x2, x3, tao;
      bool hasMS;
      CubicEquation.Solve(aTau, out x1, out x2, out x3, out hasMS);
      if (hasMS)
      {
        double dt;
        if (Mode == OperatingMode.Cooling) dt = qLD / NominalCoolingCOP;
        else dt = qLD / NominalHeatingCOP;
        double tao2 = AmbientTemperature + (dt - sg * qLD) / mcma + 273.15;
        double x1er = Math.Abs(tao2 - x1);
        double x2er = Math.Abs(tao2 - x2);
        double x3er = Math.Abs(tao2 - x3);
        if (x1er < x2er && x1er < x3er) tao = x1 - 273.15;
        else if (x2er < x3er) tao = x2 - 273.15;
        else tao = x3 - 273.15;
      }
      else tao = x1 - 273.15;

      //出力設定
      ElectricConsumption = mcma * (tao - AmbientTemperature) + sg * qLD;
      AuxiliaryElectricConsumption = auxElec;
    }
    
    #endregion

  }

  #region Immutableインターフェース

  /// <summary>読み取り専用の空気熱源モジュールヒートポンプ</summary>
  public interface ImmutableAirHeatSourceModularChillers
  {
    /// <summary>運転モードを取得する</summary>
    AirHeatSourceModularChillers.OperatingMode Mode { get; }

    /// <summary>運転効率[-]を最大化するか否か</summary>
    bool MaximizeEfficiency { get; }

    /// <summary>モジュール数[台]を取得する</summary>
    int NumberOfUnits { get; }

    /// <summary>稼働台数[台]を取得する</summary>
    int OperatingNumber { get; }

    /// <summary>ヒートポンプ機か否かを取得する</summary>
    bool IsHeatPumpModel { get; }

    /// <summary>定格冷却能力[kW/台]を取得する</summary>
    double NominalCoolingCapacity { get; }

    /// <summary>定格冷却COP[-]を取得する</summary>
    double NominalCoolingCOP { get; }

    /// <summary>定格加熱能力[kW/台]を取得する</summary>
    double NominalHeatingCapacity { get; }

    /// <summary>定格加熱COP[-]を取得する</summary>
    double NominalHeatingCOP { get; }

    /// <summary>出口水温[C]を取得する</summary>
    double WaterOutletTemperature { get; }

    /// <summary>出口水温設定値[C]を取得する</summary>
    double WaterOutletSetPointTemperature { get; }

    /// <summary>入口水温[C]を取得する</summary>
    double WaterInletTemperature { get; }

    /// <summary>水流量[(kg/s)/台]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>風量[(kg/s)/台]を取得する</summary>
    double AirFlowRate { get; }

    /// <summary>消費電力[kW/台]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>補機消費電力[kW/台]を取得する</summary>
    double AuxiliaryElectricConsumption { get; }

    /// <summary>周囲の空気乾球温度[C]を取得する</summary>
    double AmbientTemperature { get; }

    /// <summary>COP[-]を取得する</summary>
    double COP { get; }

    /// <summary>冷却量[kW]を取得する</summary>
    double CoolingLoad { get; }

    /// <summary>加熱量[kW]を取得する</summary>
    double HeatingLoad { get; }

    /// <summary>冷水量上限値[kg/s]を取得する</summary>
    double MaxChilledWaterFlowRate { get; }

    /// <summary>温水量上限値[kg/s]を取得する</summary>
    double MaxHotWaterFlowRate { get; }

    /// <summary>過負荷か否か</summary>
    bool IsOverLoad { get; }
  }

  #endregion

}
