﻿/* HotWaterAbsorptionChiller.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>温水吸収冷凍機</summary>
  [Serializable]
  public class HotWaterAbsorptionChiller
  {

    #region 定数宣言

    /// <summary>再生温度と水溶液温度のアプローチ[C]</summary>
    private const double DESORB_TEMPERATURE_APPROACH = 2;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>蒸発器の伝熱性能[kW/K]</summary>
    private double evaporatorKA;

    /// <summary>凝縮器の伝熱性能[kW/K]</summary>
    private double condensorKA;

    /// <summary>再生器の伝熱性能[kW/K]</summary>
    private double desorborKA;

    /// <summary>溶液熱交換器の伝熱性能[kW/K]</summary>
    private double solutionHexKA;

    /// <summary>冷水最小流量比[kg/s]</summary>
    private double chilledWaterMinimumFLowRatio = 0.4;

    /// <summary>冷却水最小流量比[kg/s]</summary>
    private double coolingWaterMinimumFLowRatio = 0.4;

    /// <summary>温水最小流量比[kg/s]</summary>
    private double hotWaterMinimumFLowRatio = 0.4;

    /// <summary>定格冷媒流量[m3/s]</summary>
    private double nominalSolutionFlowRate;

    /// <summary>熱損失率[-]</summary>
    private double heatLossRate = 0.0;

    /// <summary>冷水出口温度[C]を取得する</summary>
    public double ChilledWaterOutletTemperature { get; private set; }

    /// <summary>冷水出口温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterOutletSetPointTemperature { get; set; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    public double ChilledWaterInletTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>温水出口温度[C]を取得する</summary>
    public double HotWaterOutletTemperature { get; private set; }

    /// <summary>温水入口温度[C]を取得する</summary>
    public double HotWaterInletTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>冷水定格流量[kg/s]を取得する</summary>
    public double NominalChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水定格流量[kg/s]を取得する</summary>
    public double NominalCoolingWaterFlowRate { get; private set; }

    /// <summary>温水定格流量[kg/s]を取得する</summary>
    public double NominalHotWaterFlowRate { get; private set; }

    /// <summary>定格冷凍能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>冷水最小流量比[kg/s]を設定・取得する</summary>
    public double ChilledWaterMinimumFLowRatio
    {
      get { return chilledWaterMinimumFLowRatio; }
      private set { chilledWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>冷却水最小流量比[kg/s]を設定・取得する</summary>
    public double CoolingWaterMinimumFLowRatio
    {
      get { return coolingWaterMinimumFLowRatio; }
      private set { coolingWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>温水最小流量比[kg/s]を設定・取得する</summary>
    public double HotWaterMinimumFLowRatio
    {
      get { return hotWaterMinimumFLowRatio; }
      private set { hotWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>冷凍能力[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (HotWaterFlowRate == 0) return 0;
        else return CoolingLoad / (HotWaterFlowRate * WATER_SPECIFIC_HEAT 
            * (HotWaterInletTemperature - HotWaterOutletTemperature));
      }
    }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingWaterOutletTemperature">冷却水出口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="hotWaterOutletTemperature">温水出口温度[C]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    public HotWaterAbsorptionChiller
      (double chilledWaterInletTemperature, double chilledWaterOutletTemperature, double chilledWaterFlowRate, 
      double coolingWaterInletTemperature, double coolingWaterOutletTemperature, double coolingWaterFlowRate, 
      double hotWaterInletTemperature, double hotWaterOutletTemperature, double hotWaterFlowRate)
    {
      //伝熱係数を計算
      double dsvHL;
      AbsorptionRefrigerationCycle.GetHeatTransferCoefficients
        (chilledWaterInletTemperature, chilledWaterOutletTemperature, chilledWaterFlowRate,
        coolingWaterInletTemperature, coolingWaterOutletTemperature, coolingWaterFlowRate,
        hotWaterInletTemperature, hotWaterFlowRate, DESORB_TEMPERATURE_APPROACH, out evaporatorKA, 
        out condensorKA, out desorborKA, out solutionHexKA, out nominalSolutionFlowRate, out dsvHL);

      //熱損失率[-]を計算    
      double qHW = (hotWaterInletTemperature - hotWaterOutletTemperature) * hotWaterFlowRate * WATER_SPECIFIC_HEAT;
      heatLossRate = (qHW - dsvHL) / qHW;

      this.ChilledWaterOutletSetPointTemperature = chilledWaterOutletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.HotWaterInletTemperature = hotWaterInletTemperature;
      this.NominalCoolingWaterFlowRate = coolingWaterFlowRate;
      this.NominalChilledWaterFlowRate = chilledWaterFlowRate;
      this.NominalHotWaterFlowRate = hotWaterFlowRate;

      //機器を停止
      ShutOff();
    }

    #endregion

    #region publicメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    public void Update
      (double chilledWaterInletTemperature, double chilledWaterFlowRate, double coolingWaterInletTemperature,
      double coolingWaterFlowRate, double hotWaterInletTemperature, double hotWaterFlowRate)
    {
      //状態値を保存
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.HotWaterInletTemperature = hotWaterInletTemperature;
      double rch = chilledWaterFlowRate / NominalChilledWaterFlowRate;
      this.ChilledWaterFlowRate =
        Math.Max(ChilledWaterMinimumFLowRatio, Math.Min(1, rch)) * NominalChilledWaterFlowRate;
      double rcd = coolingWaterFlowRate / NominalCoolingWaterFlowRate;
      this.CoolingWaterFlowRate =
        Math.Max(CoolingWaterMinimumFLowRatio, Math.Min(1, rcd)) * NominalCoolingWaterFlowRate;
      double rht = hotWaterFlowRate / NominalHotWaterFlowRate;
      this.HotWaterFlowRate = Math.Max(HotWaterMinimumFLowRatio, Math.Min(1, rht)) * NominalHotWaterFlowRate;

      //冷却運転
      if (ChilledWaterOutletSetPointTemperature < chilledWaterInletTemperature)
      {
        double cho, cdo, ho;
        //成り行きの出口状態を計算
        AbsorptionRefrigerationCycle.GetOutletTemperatures
          (ChilledWaterInletTemperature, ChilledWaterFlowRate, CoolingWaterInletTemperature,
          CoolingWaterFlowRate, HotWaterInletTemperature, HotWaterFlowRate, evaporatorKA, condensorKA,
          desorborKA, solutionHexKA, nominalSolutionFlowRate, out cho, out cdo, out ho);

        //処理可能の場合
        if (cho < ChilledWaterOutletSetPointTemperature)
        {
          cho = ChilledWaterOutletSetPointTemperature;
          AbsorptionRefrigerationCycle.GetOutletTemperatures
            (ChilledWaterInletTemperature, ChilledWaterFlowRate, CoolingWaterInletTemperature,
            CoolingWaterFlowRate, HotWaterInletTemperature, HotWaterFlowRate, evaporatorKA, condensorKA,
            desorborKA, solutionHexKA, nominalSolutionFlowRate, cho, out cdo, out ho);
        }

        //出口状態設定
        ChilledWaterOutletTemperature = cho;
        CoolingWaterOutletTemperature = cdo;
        HotWaterOutletTemperature = (ho - heatLossRate * HotWaterInletTemperature) / (1 - heatLossRate);
        //処理熱量計算
        CoolingLoad = (ChilledWaterInletTemperature - ChilledWaterOutletTemperature)
          * WATER_SPECIFIC_HEAT * chilledWaterFlowRate;
      }
      //運転停止
      else ShutOff();
    }

    /// <summary>運転を停止させる</summary>
    public void ShutOff()
    {
      ChilledWaterOutletTemperature = ChilledWaterInletTemperature;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      HotWaterOutletTemperature = HotWaterInletTemperature;
      ChilledWaterFlowRate = CoolingWaterFlowRate = HotWaterFlowRate = 0;
      CoolingLoad = 0;
    }

    #endregion

  }
}
