﻿/* SteamBoiler.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>蒸気ボイラ</summary>
  [Serializable]
  public class SteamBoiler
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>空気比[-]</summary>
    private double airRatio = 1.1;

    /// <summary>熱損失係数[kW/K]</summary>
    private double heatLossCoefficient = 0.0;

    /// <summary>定格蒸気温度[C]</summary>
    private double nominalSteamTemperature;

    /// <summary>定格排ガス温度[C]</summary>
    private double nominalSmokeTemperature;

    /// <summary>電力の一次エネルギー換算係数[MJ/kWh]を設定・取得する</summary>
    public double PrimaryEnergyFactor { get; set; } = 9.76;

    /// <summary>燃料種別を取得する</summary>
    public Boiler.Fuel Fuel { get; private set;}

    /// <summary>蒸気圧力[kPa]を設定・取得する</summary>
    public double SteamPressure { get; set; }

    /// <summary>蒸気温度[C]を取得する</summary>
    public double SteamTemperature
    { get { return Water.GetSaturationTemperature(SteamPressure); } }

    /// <summary>温水入口温度[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>蒸気流量[kg/s]を取得する</summary>
    public double SteamFlowRate { get; private set; }

    /// <summary>定格加熱能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>定格燃料消費量[kg/s, Nm3/s]を取得する</summary>
    public double NominalFuelConsumption { get; private set; }
    
    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>燃料消費量[kg/s, Nm3/s]を取得する</summary>
    public double FuelConsumption { get; private set; }

    /// <summary>暖房能力[kW]を取得する</summary>
    public double HeatLoad { get; private set; }

    /// <summary>周囲の温度[C]を設定・取得する</summary>
    public double AmbientTemperature { get; set; }

    /// <summary>空気比を設定・取得する</summary>
    public double AirRatio
    {
      get { return airRatio; }
      set { airRatio = Math.Max(1.0, Math.Min(1.5, value)); }
    }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (HeatLoad == 0) return 0;
        else return 0.001 * HeatLoad / (ElectricConsumption * PrimaryEnergyFactor
            + FuelConsumption * Boiler.GetHeatingValue(Fuel, true));
      }
    }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="inletWaterTemperature">入口温水温度[C]</param>
    /// <param name="steamPressure">蒸気圧力[kPa]</param>
    /// <param name="steamFlowRate">蒸気流量[kg/s]</param>
    /// <param name="fuelConsumption">燃料消費量[kg/s, Nm3/s]</param>
    /// <param name="electricConsumption">消費電力[kW]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="fuel">燃料種別</param>
    /// <param name="smokeTemperature">排ガス温度[C]</param>
    public SteamBoiler(double inletWaterTemperature, double steamPressure, double steamFlowRate, 
      double fuelConsumption, double electricConsumption, double ambientTemperature, double airRatio, 
      Boiler.Fuel fuel, double smokeTemperature)
    {
      //プロパティ保存
      Fuel = fuel;
      AirRatio = airRatio;
      nominalSmokeTemperature = smokeTemperature;
      NominalFuelConsumption = fuelConsumption;
      ElectricConsumption = electricConsumption;
      AmbientTemperature = ambientTemperature;
      SteamPressure = steamPressure;
      SteamFlowRate = steamFlowRate;
      nominalSteamTemperature = Water.GetSaturationTemperature(steamPressure);

      double ts = Water.GetSaturationTemperature(steamPressure);
      double hs = Water.GetSaturatedVaporEnthalpy(ts);
      double hw = Water.GetSaturatedLiquidEnthalpy(inletWaterTemperature);
      NominalCapacity = steamFlowRate * (hs - hw);

      //熱損失係数の計算
      heatLossCoefficient = Boiler.GetHeatLossCoefficient
        (NominalCapacity, ts, AmbientTemperature, fuel, nominalSmokeTemperature, AirRatio,
        NominalFuelConsumption, AmbientTemperature);

      //停止させる
      ShutOff();
    }

    #endregion

    #region publicメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="steamFlowRate">蒸気流量[kg/s]</param>
    public void Update(double inletWaterTemperature, double steamFlowRate)
    {
      //入力条件保存
      InletWaterTemperature = inletWaterTemperature;
      SteamFlowRate = steamFlowRate;

      //流量0、設定圧力 < 入口水温の飽和圧力で停止
      double pwi = Water.GetSaturationPressure(InletWaterTemperature);
      if (SteamFlowRate <= 0) ShutOff();

      //燃料消費量を計算
      double ts = Water.GetSaturationTemperature(SteamPressure);
      double hs = Water.GetSaturatedVaporEnthalpy(ts);
      double hw = Water.GetSaturatedLiquidEnthalpy(inletWaterTemperature);
      double hl = steamFlowRate * (hs - hw);
      FuelConsumption = Boiler.GetFuelConsumption(hl, ts, AmbientTemperature, Fuel,
        nominalSmokeTemperature, airRatio, heatLossCoefficient, AmbientTemperature, nominalSteamTemperature);

      //過負荷の場合には上記流量を調整
      if (NominalFuelConsumption < FuelConsumption)
      {
        FuelConsumption = NominalFuelConsumption;
        double sf;
        Boiler.GetSteamFlowRate(InletWaterTemperature, SteamPressure,
          NominalFuelConsumption, AmbientTemperature, Fuel, nominalSmokeTemperature,
          AirRatio, heatLossCoefficient, AmbientTemperature, nominalSteamTemperature, out sf, out hl);
        SteamFlowRate = sf;
      }
      HeatLoad = hl;
    }

    /// <summary>停止する</summary>
    public void ShutOff()
    {
      SteamPressure = 101.325;
      ElectricConsumption = SteamFlowRate = HeatLoad = FuelConsumption = 0;
    }

    #endregion

  }
}
