﻿/* GasEngineHeatPump.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.MultiplePackagedHeatPump
{
  /// <summary>ガスエンジンヒートポンプ</summary>
  [Serializable]
  public class GasEngineHeatPump
  {

    #region 定数宣言

    /// <summary>廃熱回収機構の有無</summary>
    public static bool HAS_HEATRECOVERY_SYSTEM = false;

    /// <summary>冷媒種類</summary>
    private const Refrigerant.Fluid FLD = Refrigerant.Fluid.R410A;

    private readonly Refrigerant r410a = new Refrigerant(Refrigerant.Fluid.R410A);

    /// <summary>ガスエンジン総合効率[-]</summary>
    private const double TOTAL_EFFICIENCY = 0.8;

    /// <summary>最小部分負荷率[-]</summary>
    private const double MINIMUM_PARTIALLOAD_RATE = 0.05;

    /// <summary>最低凝縮温度[C]</summary>
    private const double MINIMUM_CONDENSING_TEMP = 40;

    /// <summary>最高蒸発温度[C]</summary>
    private const double MAXIMUM_EVAPORATING_TEMP = 10;

    #endregion

    #region 列挙型定義

    /// <summary>運転モード</summary>
    public enum OperatingMode
    {
      /// <summary>冷房</summary>
      Cooling,
      /// <summary>暖房</summary>
      Heating,
      /// <summary>冷暖フリー</summary>
      HeatRecovery,
      /// <summary>停止</summary>
      ShutOff
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>冷房運転定格ヘッド[kW]</summary>
    private double nominalCoolingHead;

    /// <summary>暖房運転定格ヘッド[kW]</summary>
    private double nominalHeatingHead;

    /// <summary>冷房ガス消費調整係数[-]</summary>
    private double coolingAdjustmentFactor = 1.0;

    /// <summary>暖房ガス消費調整係数[-]</summary>
    private double heatingAdjustmentFactor = 1.0;

    /// <summary>室外機熱交換器</summary>
    private OutdoorAirHeatExchanger oHex;

    /// <summary>室内機リスト</summary>
    private IndoorUnit[] iUnits;

    /// <summary>室内機リストを取得する</summary>
    public ImmutableIndoorUnit[] IndoorUnits { get { return iUnits; } }

    /// <summary>室外機空気熱交換器を取得する</summary>
    public ImmutableOutdoorAirHeatExchanger AirHeatExchanger { get { return oHex; } }

    /// <summary>運転モードを取得する</summary>
    public OperatingMode Mode { get; private set; }

    /// <summary>暖房時定格ガス消費量[kW]を取得する</summary>
    public double NominalHeatingGasConsumption { get; private set; }

    /// <summary>冷房時定格ガス消費量[kW]を取得する</summary>
    public double NominalCoolingGasConsumption { get; private set; }

    /// <summary>ガス消費量[kW]を取得する</summary>
    public double GasConsumption { get; private set; }

    /// <summary>凝縮温度[C]を取得する</summary>
    public double CondensingTemperature { get; private set; }

    /// <summary>蒸発温度[C]を取得する</summary>
    public double EvaporatingTemperature { get; private set; }

    /// <summary>屋外空気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir
    {
      get { return oHex.OutdoorAir; }
      set { oHex.OutdoorAir = value; }
    }

    /// <summary>圧縮機負荷率[-]を取得する</summary>
    public double PartialLoadRate
    {
      get; private set;
    }

    /// <summary>圧縮ヘッド[kW]を取得する</summary>
    public double Head
    {
      get; private set;
    }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ（JIS8627条件）</summary>
    /// <param name="evaporatingTemperature">蒸発温度[C]</param>
    /// <param name="condensingTemperature">凝縮温度[C]</param>
    /// <param name="coolingGasConsumption">ガス消費（冷房）[kW]</param>
    /// <param name="heatingGasConsumption">ガス消費（暖房）[kW]</param>
    /// <param name="coolingElectricConsumption">電力消費（冷房）[kW]</param>
    /// <param name="heatingElectricConsumption">電力消費（暖房）[kW]</param>
    /// <param name="auxiliaryElectricConsumption">補機電力消費[kW]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="coolingCapacity">冷房能力[kW]</param>
    /// <param name="heatingCapacity">暖房能力[kW]</param>
    /// <param name="iUnits">室内機リスト</param>
    public GasEngineHeatPump
      (double evaporatingTemperature, double condensingTemperature,
      double coolingGasConsumption, double heatingGasConsumption,
      double coolingElectricConsumption, double heatingElectricConsumption,
      double auxiliaryElectricConsumption, double borderRelativeHumidity,
      double coolingCapacity, double heatingCapacity, IndoorUnit[] iUnits)
    {
      double aFlow = (4.083 * coolingCapacity + 132.1) * 1.2 / 60d;
      //if (!HAS_HEATRECOVERY_SYSTEM) aFlow *= 1.1;

      initialize(evaporatingTemperature, condensingTemperature, coolingGasConsumption,
        heatingGasConsumption, coolingElectricConsumption, heatingElectricConsumption,
        35, 7, 0.0195, 0.0055, aFlow, aFlow, auxiliaryElectricConsumption, 
        borderRelativeHumidity, coolingCapacity, heatingCapacity, iUnits);
    }

    /// <summary>コンストラクタ（JIS8627条件）</summary>
    /// <param name="evaporatingTemperature">蒸発温度[C]</param>
    /// <param name="condensingTemperature">凝縮温度[C]</param>
    /// <param name="coolingGasConsumption">ガス消費（冷房）[kW]</param>
    /// <param name="heatingGasConsumption">ガス消費（暖房）[kW]</param>
    /// <param name="coolingElectricConsumption">電力消費（冷房）[kW]</param>
    /// <param name="heatingElectricConsumption">電力消費（暖房）[kW]</param>
    /// <param name="coolingAirFlowRate">風量（冷房）[kg/s]</param>
    /// <param name="heatingAirFlowRate">風量（暖房）[kg/s]</param>
    /// <param name="auxiliaryElectricConsumption">補機電力消費[kW]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="coolingCapacity">冷房能力[kW]</param>
    /// <param name="heatingCapacity">暖房能力[kW]</param>
    /// <param name="iUnits">室内機リスト</param>
    public GasEngineHeatPump
      (double evaporatingTemperature, double condensingTemperature,
      double coolingGasConsumption, double heatingGasConsumption,
      double coolingElectricConsumption, double heatingElectricConsumption,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double auxiliaryElectricConsumption, double borderRelativeHumidity,
      double coolingCapacity, double heatingCapacity, IndoorUnit[] iUnits)
    {
      initialize(evaporatingTemperature, condensingTemperature, coolingGasConsumption,
        heatingGasConsumption, coolingElectricConsumption, heatingElectricConsumption,
        35, 7, 0.0195, 0.0055, coolingAirFlowRate, heatingAirFlowRate, 
        auxiliaryElectricConsumption, borderRelativeHumidity,
        coolingCapacity, heatingCapacity, iUnits);
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="evaporatingTemperature">蒸発温度[C]</param>
    /// <param name="condensingTemperature">凝縮温度[C]</param>
    /// <param name="coolingGasConsumption">ガス消費（冷房）[kW]</param>
    /// <param name="heatingGasConsumption">ガス消費（暖房）[kW]</param>
    /// <param name="coolingElectricConsumption">電力消費（冷房）[kW]</param>
    /// <param name="heatingElectricConsumption">電力消費（暖房）[kW]</param>
    /// <param name="coolingAirDrybulbTemperature">外気乾球温度（冷房）[C]</param>
    /// <param name="heatingAirDrybulbTemperature">外気乾球温度（暖房）[C]</param>
    /// <param name="coolingAirHumidityRatio">絶対湿度（冷房）[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">絶対湿度（暖房）[kg/kg]</param>
    /// <param name="coolingAirFlowRate">風量（冷房）[kg/s]</param>
    /// <param name="heatingAirFlowRate">風量（暖房）[kg/s]</param>
    /// <param name="auxiliaryElectricConsumption">補機電力消費[kW]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="coolingCapacity">冷房能力[kW]</param>
    /// <param name="heatingCapacity">暖房能力[kW]</param>
    /// <param name="iUnits">室内機リスト</param>
    public GasEngineHeatPump
      (double evaporatingTemperature, double condensingTemperature,
      double coolingGasConsumption, double heatingGasConsumption,
      double coolingElectricConsumption, double heatingElectricConsumption,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double auxiliaryElectricConsumption, double borderRelativeHumidity,
      double coolingCapacity, double heatingCapacity, IndoorUnit[] iUnits)
    {
      initialize(evaporatingTemperature, condensingTemperature, coolingGasConsumption,
        heatingGasConsumption, coolingElectricConsumption, heatingElectricConsumption,
        coolingAirDrybulbTemperature, heatingAirDrybulbTemperature,
        coolingAirHumidityRatio, heatingAirHumidityRatio, coolingAirFlowRate,
        heatingAirFlowRate, auxiliaryElectricConsumption, borderRelativeHumidity,
        coolingCapacity, heatingCapacity, iUnits);
    }

    /// <summary>初期化する</summary>
    /// <param name="evaporatingTemperature">蒸発温度[C]</param>
    /// <param name="condensingTemperature">凝縮温度[C]</param>
    /// <param name="coolingGasConsumption">ガス消費（冷房）[kW]</param>
    /// <param name="heatingGasConsumption">ガス消費（暖房）[kW]</param>
    /// <param name="coolingElectricConsumption">電力消費（冷房）[kW]</param>
    /// <param name="heatingElectricConsumption">電力消費（暖房）[kW]</param>
    /// <param name="coolingAirDrybulbTemperature">外気乾球温度（冷房）[C]</param>
    /// <param name="heatingAirDrybulbTemperature">外気乾球温度（暖房）[C]</param>
    /// <param name="coolingAirHumidityRatio">絶対湿度（冷房）[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">絶対湿度（暖房）[kg/kg]</param>
    /// <param name="coolingAirFlowRate">風量（冷房）[kg/s]</param>
    /// <param name="heatingAirFlowRate">風量（暖房）[kg/s]</param>
    /// <param name="auxiliaryElectricConsumption">補機電力消費[kW]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="coolingCapacity">冷房能力[kW]</param>
    /// <param name="heatingCapacity">暖房能力[kW]</param>
    /// <param name="iUnits">室内機リスト</param>
    public void initialize
      (double evaporatingTemperature, double condensingTemperature,
      double coolingGasConsumption, double heatingGasConsumption,
      double coolingElectricConsumption, double heatingElectricConsumption,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double auxiliaryElectricConsumption, double borderRelativeHumidity,
      double coolingCapacity, double heatingCapacity, IndoorUnit[] iUnits)
    {
      const double DELTA = 0.0001;

      //パラメータを保存
      this.iUnits = iUnits;
      NominalCoolingGasConsumption = coolingGasConsumption;
      NominalHeatingGasConsumption = heatingGasConsumption;
      
      //凝縮・蒸発圧力から蒸発器出入口エンタルピーを計算
      double pCnd, pEvp, sld, svd;
      r410a.GetSaturatedPropertyFromTemperature(condensingTemperature + 273.1, out sld, out svd, out pCnd);
      double hlCnd = r410a.GetEnthalpyFromTemperatureAndDensity(condensingTemperature + 273.15, sld);
      r410a.GetSaturatedPropertyFromTemperature(evaporatingTemperature + 273.15, out sld, out svd, out pEvp);
      double hvEvp = r410a.GetEnthalpyFromTemperatureAndDensity(evaporatingTemperature + 273.15, svd);

      //圧縮ヘッド計算用下処理
      double dhSV = 1 / ((hvEvp - hlCnd) * svd);
      double kappa = r410a.GetSpecificHeatRatioFromTemperatureAndDensity(evaporatingTemperature + 273.15, svd);
      double k1 = (kappa - 1) / kappa;
      k1 = pEvp * (Math.Pow(pCnd / pEvp, k1) - 1) / k1;

      //冷房運転時のヘッド推定
      double vf = coolingCapacity * dhSV;
      nominalCoolingHead = vf * k1;
      double nEff = getCompressionEfficienty(1.0);
      coolingAdjustmentFactor = coolingGasConsumption * nEff / nominalCoolingHead;

      //暖房運転時のヘッド推定（収束計算）
      double nH = 1;
      vf = (heatingCapacity - nH) * dhSV;
      double err1 = vf * k1 - nH;
      while (0.0001 < Math.Abs(err1))
      {
        vf = (heatingCapacity - (nH + DELTA)) * dhSV;
        double err2 = vf * k1 - (nH + DELTA);
        nH -= err1 * DELTA / (err2 - err1);
        vf = (heatingCapacity - nH) * dhSV;
        err1 = vf * k1 - nH;
      }
      nominalHeatingHead = nH;
      heatingAdjustmentFactor = heatingGasConsumption * nEff / nominalHeatingHead;

      //熱交換器伝熱面積を初期化
      double hCap;
      if (HAS_HEATRECOVERY_SYSTEM) hCap = heatingCapacity - heatingGasConsumption * TOTAL_EFFICIENCY;
      else hCap = heatingCapacity - nominalHeatingHead;
      oHex = new OutdoorAirHeatExchanger
        (evaporatingTemperature, condensingTemperature, coolingAirDrybulbTemperature,
        heatingAirDrybulbTemperature, coolingAirHumidityRatio, heatingAirHumidityRatio,
        coolingCapacity + nominalCoolingHead, hCap,
        coolingAirFlowRate, heatingAirFlowRate, auxiliaryElectricConsumption, borderRelativeHumidity);
    }

    #endregion

    #region 更新処理

    /// <summary>状態を更新する</summary>
    public void Update()
    {
      //運転モード（冷専、暖専、冷暖、停止）判定
      bool hasCoolingUnit = false;
      bool hasHeatingUnit = false;
      double totalCoolingLoad = 0;
      double totalHeatingLoad = 0;
      EvaporatingTemperature = 999;
      CondensingTemperature = -999;
      foreach (IndoorUnit iu in iUnits)
      {
        if (iu.IsOperating && iu.IsCooling)
        {
          hasCoolingUnit = true;
          EvaporatingTemperature = Math.Min(EvaporatingTemperature, iu.GetRefrigerantTemperature());
          totalCoolingLoad += Math.Min(-iu.HeatLoad, -iu.MaxCoolingCapacity);
        }
        else if (iu.IsOperating && !iu.IsCooling)
        {
          hasHeatingUnit = true;
          CondensingTemperature = Math.Max(CondensingTemperature, iu.GetRefrigerantTemperature());
          totalHeatingLoad += Math.Min(iu.HeatLoad, iu.MaxHeatingCapacity);
        }
      }
      //冷暖
      if (hasCoolingUnit && hasHeatingUnit)
      {
        Mode = OperatingMode.HeatRecovery;
      }
      //冷専
      else if (hasCoolingUnit)
      {
        EvaporatingTemperature = Math.Min(MAXIMUM_EVAPORATING_TEMP, EvaporatingTemperature);
        Mode = OperatingMode.Cooling;
        oHex.IsCooling = true;
        if (totalCoolingLoad <= 0) shutOff();
        else update_coolingMode(totalCoolingLoad);
      }
      //暖専
      else if (hasHeatingUnit)
      {
        CondensingTemperature = Math.Max(MINIMUM_CONDENSING_TEMP, CondensingTemperature);
        Mode = OperatingMode.Heating;
        oHex.IsCooling = false;
        if (totalHeatingLoad <= 0) shutOff();
        else update_heatingMode(totalHeatingLoad);
      }
      //停止
      else shutOff();
    }

    private void shutOff()
    {
      Mode = OperatingMode.ShutOff;
      GasConsumption = 0.0;
      PartialLoadRate = 0.0;
      Head = 0.0;
      //
    }

    #endregion

    #region privateメソッド（冷房運転）

    /// <summary>運転状態を更新する</summary>
    /// <param name="cLoad">冷房負荷[kW]</param>
    private void update_coolingMode(double cLoad)
    {
      //最大ヘッドを前提に必要ヘッドを再計算
      double mHead = getCompressionEfficienty(1) * NominalCoolingGasConsumption / coolingAdjustmentFactor;
      double head;
      if (EvaporatingTemperature < -10)
      {
        head = mHead + 1;
        EvaporatingTemperature = 2;
      }
      else head = getHead_coolingMode(cLoad, mHead);

      //熱処理可能の場合:処理熱量固定でヘッドを収束計算
      if (head < mHead)
      {
        Roots.ErrorFunction eFnc = delegate (double hd)
        {
          return getHead_coolingMode(cLoad, hd) - hd;
        };
        head = Roots.Brent(0, mHead, 0.0001, eFnc) + 0.0001;
        //室内機の状態を更新
        foreach (IndoorUnit iu in iUnits)
          if (iu.IsOperating && iu.IsCooling) iu.UpdateState(EvaporatingTemperature);
      }
      //過負荷の場合:ヘッド固定で蒸発温度を収束計算
      else
      {
        head = mHead;
        Roots.ErrorFunction eFnc = delegate (double eTemp)
        {
          EvaporatingTemperature = eTemp;
          return getHead_coolingMode(head) - head;
        };
        EvaporatingTemperature = Roots.Brent(EvaporatingTemperature, CondensingTemperature, 0.001, eFnc);
      }

      //ガス消費量[kW]を更新
      GasConsumption = coolingAdjustmentFactor * head
        / getCompressionEfficienty(Math.Max(MINIMUM_PARTIALLOAD_RATE, head / mHead));
      PartialLoadRate = head / mHead;
      Head = head;
    }

    /// <summary>圧縮ヘッド[kW]を計算する</summary>
    /// <param name="head">圧縮ヘッド（仮定値）[kW]</param>
    /// <returns>圧縮ヘッド[kW]</returns>
    private double getHead_coolingMode(double head)
    {
      double cLoad = 0;
      foreach (IndoorUnit iu in iUnits)
      {
        if (iu.IsOperating && iu.IsCooling)
        {
          iu.UpdateState(EvaporatingTemperature);
          cLoad += -iu.HeatLoad + iu.OverLoad;
        }
      }
      return getHead_coolingMode(cLoad, head);
    }

    /// <summary>圧縮ヘッド[kW]を計算する</summary>
    /// <param name="cLoad">冷房負荷[kW]</param>
    /// <param name="head">圧縮ヘッド（仮定値）[kW]</param>
    /// <returns>圧縮ヘッド[kW]</returns>
    private double getHead_coolingMode(double cLoad, double head)
    {
      double qCnd = cLoad + head;
      CondensingTemperature = 
        Math.Max(MINIMUM_CONDENSING_TEMP, oHex.GetRefrigerantTemperature(-qCnd));
      return getHead(EvaporatingTemperature, CondensingTemperature, cLoad, r410a);
    }

    #endregion

    #region privateメソッド（暖房運転）

    /// <summary>運転状態を更新する</summary>
    /// <param name="hLoad">暖房負荷[kW]</param>
    private void update_heatingMode(double hLoad)
    {
      //最大ヘッドを前提に必要ヘッドを再計算
      double mHead = getCompressionEfficienty(1) * NominalHeatingGasConsumption / heatingAdjustmentFactor;
      double heatRecovery = NominalHeatingGasConsumption * TOTAL_EFFICIENCY - mHead;
      double head;
      if (60 < CondensingTemperature) head = mHead + 1;
      else head = getHead_heatingMode(hLoad, heatRecovery, mHead);

      //熱処理可能の場合:処理熱量固定でヘッドを収束計算
      if (head < mHead)
      {
        Roots.ErrorFunction eFnc = delegate (double hd)
        {
          double hRcv = hd / getCompressionEfficienty
          (Math.Max(MINIMUM_PARTIALLOAD_RATE, hd / nominalHeatingHead)) - hd;
          return getHead_heatingMode(hLoad, hRcv, hd) - hd;
        };
        head = Roots.Brent(0, mHead, 0.0001, eFnc) + 0.0001;
        //室内機の状態を更新
        foreach (IndoorUnit iu in iUnits)
          if (iu.IsOperating && !iu.IsCooling) iu.UpdateState(CondensingTemperature);
      }
      //過負荷の場合:ヘッド固定で凝縮温度を収束計算
      else
      {
        head = mHead;
        Roots.ErrorFunction eFnc = delegate (double cTemp)
        {
          CondensingTemperature = cTemp;
          return getHead_heatingMode(heatRecovery, head) - head;
        };
        CondensingTemperature = Roots.Brent(EvaporatingTemperature, CondensingTemperature, 0.001, eFnc);
      }

      //ガス消費量[kW]を更新
      GasConsumption = heatingAdjustmentFactor * head 
        / getCompressionEfficienty(Math.Max(MINIMUM_PARTIALLOAD_RATE, head / mHead));
      PartialLoadRate = head / mHead;
      Head = head;
    }

    /// <summary>圧縮ヘッド[kW]を計算する</summary>
    /// <param name="heatRecovery">排熱回収[kW]</param>
    /// <param name="head">圧縮ヘッド（仮定値）[kW]</param>
    /// <returns>圧縮ヘッド[kW]</returns>
    private double getHead_heatingMode
      (double heatRecovery, double head)
    {
      double hLoad = 0;
      foreach (IndoorUnit iu in iUnits)
      {
        if (iu.IsOperating && !iu.IsCooling)
        {
          iu.UpdateState(CondensingTemperature);
          hLoad += iu.HeatLoad - iu.OverLoad;
        }
      }
      return getHead_heatingMode(hLoad, heatRecovery, head);
    }

    /// <summary>圧縮ヘッド[kW]を計算する</summary>
    /// <param name="hLoad">暖房負荷[kW]</param>
    /// <param name="heatRecovery">排熱回収[kW]</param>
    /// <param name="head">圧縮ヘッド（仮定値）[kW]</param>
    /// <returns>圧縮ヘッド[kW]</returns>
    private double getHead_heatingMode
      (double hLoad, double heatRecovery, double head)
    {
      double qEvp;
      if (HAS_HEATRECOVERY_SYSTEM) qEvp = hLoad - (heatRecovery + head);
      else qEvp = hLoad - head;
      if (qEvp < 0) return 0;
      else
        EvaporatingTemperature =
       Math.Min(MAXIMUM_EVAPORATING_TEMP, oHex.GetRefrigerantTemperature(qEvp));
      return getHead(EvaporatingTemperature, CondensingTemperature, hLoad - head, r410a) + oHex.DefrostLoad;
    }

    #endregion

    #region private staticメソッド

    /// <summary>圧縮ヘッド効率[-]を計算する</summary>
    /// <param name="partialLoad">負荷率[-]</param>
    /// <returns>圧縮ヘッド効率[-]</returns>
    private static double getCompressionEfficienty(double partialLoad)
    {
      return Math.Max(0.0001, 0.0719 * Math.Log(partialLoad) + 0.3539);
    }

    /// <summary>理論圧縮ヘッド[kW]を計算する</summary>
    /// <param name="tEvp">蒸発温度[C]</param>
    /// <param name="tCnd">凝縮温度[C]</param>
    /// <param name="qEvp">蒸発器負荷[kW]</param>
    /// <param name="refrg">冷媒</param>
    /// <returns>理論圧縮ヘッド[kW]</returns>
    /// <remarks>過熱・過冷却の計算が必要な場合はこのメソッドを調整</remarks>
    private static double getHead
      (double tEvp, double tCnd, double qEvp, Refrigerant refrg)
    {
      const double MIN_EVP = -25;
      const double MAX_CND = 65;

      //飽和状態の計算
      double dslCnd, dsvCnd, pCnd, dslEvp, dsvEvp, pEvp;

      refrg.GetSaturatedPropertyFromTemperature
        (Math.Min(MAX_CND, tCnd) + 273.15, out dslCnd, out dsvCnd, out pCnd);
      refrg.GetSaturatedPropertyFromTemperature
        (Math.Max(MIN_EVP, tEvp) + 273.15, out dslEvp, out dsvEvp, out pEvp);
      double hslCnd = refrg.GetEnthalpyFromTemperatureAndDensity(tCnd + 273.15, dslCnd);
      double hsvEvp = refrg.GetEnthalpyFromTemperatureAndDensity(tEvp + 273.15, dsvEvp);

      //冷媒流量[m3/s]の計算
      double vRate = qEvp / (dsvEvp * (hsvEvp - hslCnd));
      //圧縮仕事[kW]の計算
      double kappa = refrg.GetSpecificHeatRatioFromTemperatureAndDensity(tEvp + 273.15, dsvEvp);
      kappa = (kappa - 1) / kappa;
      return vRate / kappa * pEvp * (Math.Pow(pCnd / pEvp, kappa) - 1);
    }

    #endregion

  }
}
