﻿/* FlatPlateSolarCollector.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.SolarEnergy
{
  /// <summary>平板式集熱器</summary>
  public static class FlatPlateSolarCollector
  {

    #region 定数宣言

    /// <summary>黒体の放射定数[W/(m2K4)]</summary>
    private const double BLACK_CONSTANT = 5.67e-8;

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>ガラスの透過率の入射角特性近似係数</summary>
    private static readonly double[] MI_TAU = new double[] { 2.552, 1.364, -11.388, 13.617, -5.146 };

    /// <summary>ガラスの反射率の入射角特性近似係数</summary>
    private static readonly double[] MI_RHO = new double[] { 5.189, -12.392, 16.593, -11.851, 3.461 };

    /// <summary>天空日射に対する透過率の係数</summary>
    private static readonly double DIFFUSE_TRANSMITTANCE_COEF;

    /// <summary>天空日射に対する反射率の係数</summary>
    private static readonly double DIFFUSE_REFLECTANCE_COEF;

    #endregion

    #region staticコンストラクタ

    /// <summary>staticコンストラクタ</summary>
    static FlatPlateSolarCollector()
    {
      for (int i = 0; i < 5; i++)
      {
        DIFFUSE_TRANSMITTANCE_COEF += MI_TAU[i] / (i + 3);
        DIFFUSE_REFLECTANCE_COEF += MI_RHO[i] / (i + 3);
      }
      DIFFUSE_TRANSMITTANCE_COEF *= 2;
      DIFFUSE_REFLECTANCE_COEF *= 2;
    }

    #endregion

    #region staticメソッド

    /// <summary>太陽熱集熱器の集熱量を計算する</summary>
    /// <param name="skyTemperature">天空の温度[C]</param>
    /// <param name="airTemperature">空気温度[C]</param>
    /// <param name="directNormalRadiation">法線面直達日射量[W/m2]</param>
    /// <param name="diffuseRadiation">天空日射量[W/m2]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="waterInletTemperature">入口水温[C]</param>
    /// <param name="airThickness">空気層の厚み[m]</param>
    /// <param name="glassEmissivity">ガラスの放射率[-]</param>
    /// <param name="panelEmissivity">パネルの放射率[-]</param>
    /// <param name="windSpeed">外部風速[m/s]</param>
    /// <param name="insulatorThickness">断熱材厚み[m]</param>
    /// <param name="insulatorThermalConductivity">断熱材熱伝導率[W/(mK)]</param>
    /// <param name="surfaceArea">パネル面積[m2]</param>
    /// <param name="tubePitch">集熱媒体管の設置間隔[m]</param>
    /// <param name="innerDiameter">集熱媒体管の内径[m]</param>
    /// <param name="outerDiameter">集熱媒体管の外径[m]</param>
    /// <param name="panelThickness">パネル厚み[m]</param>
    /// <param name="panelThermalConductivity">パネルの熱伝導率[W/(mK)]</param>
    /// <param name="cosTheta">入射角の余弦</param>
    /// <param name="glassTransmittance">ガラスの日射透過率[-]</param>
    /// <param name="glassReflectance">ガラスの日射反射率[-]</param>
    /// <param name="panelAbsorptance">パネルの日射吸収率[-]</param>
    /// <param name="panelTemperature">パネルの温度[C]</param>
    /// <param name="glassTemperature">ガラスの温度[C]</param>
    /// <param name="waterOutletTemperature">出口水温[C]</param>
    /// <param name="meanWaterTemperature">平均水温[C]</param>
    /// <param name="efficiency">集熱効率[-]</param>
    /// <returns>集熱量[W]</returns>
    public static double GetHeatTransfer
      (double skyTemperature, double airTemperature, double directNormalRadiation, double diffuseRadiation, 
      double waterFlowRate, double waterInletTemperature, double airThickness, double glassEmissivity,
      double panelEmissivity, double windSpeed, double insulatorThickness, double insulatorThermalConductivity,
      double surfaceArea, double tubePitch, double innerDiameter, double outerDiameter, double panelThickness,
      double panelThermalConductivity, double cosTheta, double glassTransmittance, double glassReflectance, 
      double panelAbsorptance, out double panelTemperature, out double glassTemperature,
      out double waterOutletTemperature, out double meanWaterTemperature, out double efficiency)
    {
      double heatTransfer = 0; //集熱量[W]
      glassTemperature = 0;
      meanWaterTemperature = 0;

      //収束計算に無関係な数値を確定する//////////////////////////////
      //ガラスとパネルの日射吸収量[W/m2]の計算
      double sagDN, sapDN, sagDF, sapDF;
      getSolarAbsorptanceOfGlassAndPanel
        (cosTheta, glassTransmittance, glassReflectance, panelAbsorptance, out sagDN, out sapDN, out sagDF, out sapDF);
      double glassSolarAbsorption = sagDN * directNormalRadiation + sagDF * diffuseRadiation;
      double panelSolarAbsorption = sapDN * directNormalRadiation + sapDF * diffuseRadiation;

      //ガラスと外気との間の対流熱伝達率[W/m2K]の計算
      double convectiveHeatLoss_ga = windSpeed * 7.6 + 4.7;

      //ガラスとパネルとの間の熱通過率[W/m2K]の計算
      double airThermalConductivity = 0.0241 + 0.000077 * airTemperature;
      double convectiveHeatLoss_pg = airThermalConductivity / airThickness;   //Nu=1.0とする
      //ガラスとパネルとの間の有効放射定数の計算
      double effectiveEmissivity_pg = BLACK_CONSTANT / (1 / panelEmissivity + 1 / glassEmissivity - 1);

      //パネルと背面との間の熱通過率[W/m2K]の計算
      double heatLoss_pb = insulatorThermalConductivity / insulatorThickness;

      //パネル温度の初期値設定
      panelTemperature = airTemperature + 5;
      //1タイムステップ前のパネル温度
      double ptL = panelTemperature + 1;
      //パネル温度の収束計算
      int iterNumP = 0;
      while (0.001 < Math.Abs(panelTemperature - ptL))
      {
        if (100 < iterNumP) throw new Exception("FlatPlateSolarCollector iteration error");

        double heatLoss_pg = 0; //ガラスとパネルとの間の熱通過率[W/m2K]
        double radiativeHeatLoss_ga = 0; //ガラスと天空との間の放射熱伝達率[W/m2K]

        //1ステップ前のパネル温度を保存
        ptL = panelTemperature;

        //誤差関数を定義
        Roots.ErrorFunction efncG = delegate (double gTemp)
        {
          //ガラスと天空との放射熱伝達率[W/m2K]の計算
          radiativeHeatLoss_ga = 4 * BLACK_CONSTANT * glassEmissivity 
          * Math.Pow((gTemp + skyTemperature) / 2 + CONVERT_C_TO_K, 3);

          //ガラスとパネルとの間の熱通過率[W/m2K]の計算
          double radiativeHeatLoss_pg = 4 * effectiveEmissivity_pg * Math.Pow((ptL + gTemp) / 2 + CONVERT_C_TO_K, 3);
          heatLoss_pg = convectiveHeatLoss_pg + radiativeHeatLoss_pg;

          //更新されたガラス温度と仮定値との間の誤差
          return gTemp - (glassSolarAbsorption + heatLoss_pg * ptL + radiativeHeatLoss_ga * skyTemperature 
          + convectiveHeatLoss_ga * airTemperature) / (heatLoss_pg + radiativeHeatLoss_ga + convectiveHeatLoss_ga);
        };

        //ニュートン・ラプソン法でガラス温度を収束計算
        glassTemperature = Roots.Newton(efncG, 0.5 * (panelTemperature + airTemperature), 0.0001, 0.01, 0.01, 10);

        //熱損失係数UL[W/m2K]の計算
        double heatLoss_ga = convectiveHeatLoss_ga 
          + radiativeHeatLoss_ga * (glassTemperature - skyTemperature) / (glassTemperature - airTemperature);
        double heatLossCoefficient = (heatLoss_pg * heatLoss_ga) / (heatLoss_pg + heatLoss_ga) + heatLoss_pb;

        //四角フィンのフィン効率[-]の計算
        double wd2 = Math.Sqrt(heatLossCoefficient / (panelThermalConductivity * panelThickness)) 
          * (tubePitch - innerDiameter) / 2;
        double rectFinEfficiency = Math.Tanh(wd2) / wd2;

        //集熱器のフィン効率[-]の計算
        double tubeHeatTransfer = 
          getConvectiveHeatTransferCoefficientOfTube(waterInletTemperature, innerDiameter, waterFlowRate);
        double fe1 = 1 / (heatLossCoefficient * (outerDiameter + (tubePitch - outerDiameter) * rectFinEfficiency));
        double fe2 = 1 / (Math.PI * innerDiameter * tubeHeatTransfer);
        double finEfficiency = (1 / heatLossCoefficient) / (tubePitch * (fe1 + fe2));

        //集熱器除去因子FR[-]の計算
        double bf = waterFlowRate * Water.GetLiquidIsobaricSpecificHeat(waterInletTemperature) 
          * 1000 / (surfaceArea * heatLossCoefficient);
        double heatRemovalFactor = bf * (1 - Math.Exp(-finEfficiency / bf));

        //熱移動量[W]の計算
        heatTransfer = panelSolarAbsorption + glassSolarAbsorption * heatLoss_pg / (heatLoss_pg + heatLoss_ga);
        heatTransfer -= heatLossCoefficient * (waterInletTemperature - airTemperature);
        heatTransfer *= surfaceArea * heatRemovalFactor;

        //パネルと熱媒の平均温度を更新
        double buff = (heatTransfer / surfaceArea) / (heatLossCoefficient * heatRemovalFactor);
        panelTemperature = waterInletTemperature + buff * (1 - heatRemovalFactor);
        meanWaterTemperature = waterInletTemperature + buff * (1 - heatRemovalFactor / finEfficiency);

        iterNumP++;
      }
      //出口水温の計算
      waterOutletTemperature = 0.001 * waterInletTemperature + heatTransfer
        / (waterFlowRate * Water.GetLiquidIsobaricSpecificHeat(waterInletTemperature));
      //集熱効率の計算
      efficiency = heatTransfer / (cosTheta * directNormalRadiation + diffuseRadiation) / surfaceArea;

      //集熱量を出力
      return heatTransfer;
    }

    /// <summary>ガラスとパネルの総合日射吸収率[-]を計算する</summary>
    /// <param name="cosineTheta">日射入射角の余弦cosθ</param>
    /// <param name="glassTransmittance">ガラスの日射透過率[-]</param>
    /// <param name="glassReflectance">ガラスの日射反射率[-]</param>
    /// <param name="panelAbsorptance">パネルの日射吸収率[-]</param>
    /// <param name="glassDNSolarAbsorptance">法線面直達日射に対するパネルの総合日射吸収率[-]</param>
    /// <param name="panelDNSolarAbsorptance">法線面直達日射に対するガラスの総合日射吸収率[-]</param>
    /// <param name="glassDFSolarAbsorptance">天空日射に対するパネルの総合日射吸収率[-]</param>
    /// <param name="panelDFSolarAbsorptance">天空日射に対するガラスの総合日射吸収率[-]</param>
    private static void getSolarAbsorptanceOfGlassAndPanel
      (double cosineTheta, double glassTransmittance, double glassReflectance, double panelAbsorptance, 
      out double glassDNSolarAbsorptance, out double panelDNSolarAbsorptance,
      out double glassDFSolarAbsorptance, out double panelDFSolarAbsorptance)
    {
      //ガラスの規準化透過率と規準化反射率の計算
      double tn = 0;
      double rn = 0;
      for (int i = 4; 0 <= i; i--)
      {
        tn = cosineTheta * (tn + MI_TAU[i]);
        rn = cosineTheta * (rn + MI_RHO[i]);
      }

      //法線面直達日射に対するガラス単体の透過率・反射率・吸収率の計算
      double tauThetaG_DN = glassTransmittance * tn;
      double rhoThetaG_DN = 1 - (1 - glassReflectance) * rn;
      double alphaThetaG_DN = 1 - (tauThetaG_DN + rhoThetaG_DN);

      //天空日射に対するガラス単体の透過率・反射率・吸収率の計算
      double tauThetaG_DF = glassTransmittance * DIFFUSE_TRANSMITTANCE_COEF;
      double rhoThetaG_DF = glassReflectance * DIFFUSE_REFLECTANCE_COEF;
      double alphaThetaG_DF = 1 - (tauThetaG_DF + rhoThetaG_DF);

      //パネル単体の吸収率・反射率の計算
      double alphaThetaP = cosineTheta * panelAbsorptance;
      double rhoThetaP = 1 - alphaThetaP;

      //ガラスの総合吸収率の計算
      glassDNSolarAbsorptance = alphaThetaG_DN * (1 + tauThetaG_DN * rhoThetaP / (1 - rhoThetaG_DN * rhoThetaP));
      glassDFSolarAbsorptance = alphaThetaG_DF * (1 + tauThetaG_DF * rhoThetaP / (1 - rhoThetaG_DF * rhoThetaP));

      //パネルの総合吸収率の計算
      panelDNSolarAbsorptance = tauThetaG_DN * alphaThetaP / (1 - rhoThetaG_DN * rhoThetaP);
      panelDFSolarAbsorptance = tauThetaG_DF * alphaThetaP / (1 - rhoThetaG_DF * rhoThetaP);
    }

    /// <summary>水温・直径・水量から配管内対流熱伝達率を計算する</summary>
    /// <param name="waterTemperature">水温[C]</param>
    /// <param name="diameter">直径[m]</param>
    /// <param name="waterFlowRate">水量[L/min]</param>
    /// <returns>対流熱伝達率[W/(m2·K)]</returns>
    private static double getConvectiveHeatTransferCoefficientOfTube
      (double waterTemperature, double diameter, double waterFlowRate)
    {
      //動粘性係数[m2/s]・熱拡散率[m2/s]・熱伝導率[W/(m·K)]を計算
      double v = Water.GetLiquidDynamicViscosity(waterTemperature);
      double a = Water.GetLiquidThermalDiffusivity(waterTemperature);
      double lambda = Water.GetLiquidThermalConductivity(waterTemperature);

      //配管内流速[m/s]を計算
      double u = (waterFlowRate / 60 / 1000) / (Math.Pow(diameter / 2, 2) * Math.PI);

      //ヌセルト数を計算
      double reNumber = u * diameter / v;
      double prNumber = v / a;
      double nuNumber = 0.023 * Math.Pow(reNumber, 0.8) * Math.Pow(prNumber, 0.4);

      //ヌセルト数から対流熱伝達率を計算
      return nuNumber * lambda / diameter;
    }

    #endregion

  }
}
