﻿/* BuildingThermalModel.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading.Tasks;

using Popolo.Weather;

namespace Popolo.ThermalLoad
{
  /// <summary>建物の熱負荷計算モデル</summary>
  [Serializable]
  public class BuildingThermalModel: ImmutableBuildingThermalModel
  {

    #region プロパティ

    /// <summary>並列計算を行うか否か</summary>
    public static bool EnableParallelComputing { get; set; } = true;

    /// <summary>最大並列計算数を設定・取得する</summary>
    public static int MaxDegreeOfParallelism { get; set; } = 4;
    
    #endregion

    #region インスタンス変数

    /// <summary>初回の予測計算か否か</summary>
    private bool isFirstForecast = true;

    /// <summary>顕熱流に関する境界条件変更の真偽</summary>
    private Dictionary<MultiRooms, bool> hasHTChgd = new Dictionary<MultiRooms, bool>();

    /// <summary>水分流に関する境界条件変更の真偽</summary>
    private Dictionary<MultiRooms, bool> hasWTChgd = new Dictionary<MultiRooms, bool>();

    /// <summary>壁リスト</summary>
    private List<Wall> walls;

    /// <summary>多数室</summary>
    private MultiRooms[] mRooms;    

    /// <summary>計算時間間隔[sec]</summary>
    private double timeStep = 3600;

    /// <summary>ゾーン間換気リスト</summary>
    private interZoneAirFlowCollection[][] zoneVent;

    #endregion

    #region プロパティ

    /// <summary>多数室を取得する</summary>
    public ImmutableMultiRooms[] MultiRoom { get { return mRooms; } }

    /// <summary>現在の日時を取得する</summary>
    public DateTime CurrentDateTime { get; private set; }

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set
      {
        for (int i = 0; i < mRooms.Length; i++) mRooms[i].TimeStep = value;
        timeStep = value;
      }
    }

    /// <summary>太陽を取得する</summary>
    public ImmutableSun Sun { get; private set; }

    /// <summary>外気乾球温度[C]を取得する</summary>
    public double OutdoorTemperature { get; private set; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    public double OutdoorHumidityRatio { get; private set; }

    /// <summary>夜間放射[W/m2]を取得する</summary>
    public double NocturnalRadiation { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="mRooms">多数室</param>
    public BuildingThermalModel(MultiRooms[] mRooms)
    {
      this.mRooms = mRooms;
      zoneVent = new interZoneAirFlowCollection[mRooms.Length][];
      walls = new List<Wall>();
      List<ImmutableZone> zns = new List<ImmutableZone>();
      List<ImmutableWindow> wins = new List<ImmutableWindow>();
      for (int i = 0; i < mRooms.Length; i++)
      {
        zoneVent[i] = new interZoneAirFlowCollection[mRooms[i].ZoneNumber];
        for (int j = 0; j < zoneVent[i].Length; j++) zoneVent[i][j] = new interZoneAirFlowCollection();
        hasHTChgd.Add(mRooms[i], true);
        hasWTChgd.Add(mRooms[i], true);

        foreach (Wall wl in mRooms[i].Walls)
          if (!walls.Contains(wl)) walls.Add(wl);

        //ゾーンの重複確認
        foreach (ImmutableZone zn in mRooms[i].Zones)
          if (zns.Contains(zn)) throw new Exception("Zone belongs to more than 1 MultiRooms");
        zns.AddRange(mRooms[i].Zones);

        //窓の重複確認
        foreach (ImmutableWindow win in mRooms[i].Windows)
          if (wins.Contains(win)) throw new Exception("Window belongs to more than 1 MultiRooms");
        wins.AddRange(mRooms[i].Windows);
      }
    }
    
    #endregion

    #region 熱平衡更新処理

    /// <summary>将来の熱平衡状態を予測する</summary>
    /// <remarks>
    /// 繰り返し呼び出すことが可能。
    /// 最後にFixHeatTransferを呼び出して確定させる
    /// </remarks>
    public void ForecastHeatTransfer()
    {
      //初回計算時に他のゾーン温湿度を境界条件に設定
      if (isFirstForecast)
      {
        setInterZoneAirFlow();
        isFirstForecast = false;
      }

      if (EnableParallelComputing)
      {
        ParallelOptions options = new ParallelOptions();
        options.MaxDegreeOfParallelism = MaxDegreeOfParallelism;
        Parallel.ForEach(mRooms, options, mr =>
        {
          if (hasHTChgd[mr])
          {
            mr.ForecastHeatTransfer();
            hasHTChgd[mr] = false;
          }
        });
      }
      else
      {
        foreach (MultiRooms mr in mRooms)
        {
          if (hasHTChgd[mr])
          {
            mr.ForecastHeatTransfer();
            hasHTChgd[mr] = false;
          }
        }
      }
    }

    /// <summary>将来の水分衡状態を予測する</summary>
    /// <remarks>
    /// 繰り返し呼び出すことが可能。
    /// 最後にFixWaterTransferを呼び出して確定させる
    /// </remarks>
    public void ForecastWaterTransfer()
    {
      //初回計算時に他のゾーン温湿度を境界条件に設定
      if (isFirstForecast)
      {
        setInterZoneAirFlow();
        isFirstForecast = false;
      }

      foreach (MultiRooms mr in mRooms)
      {
        if (!mr.SolveMoistureTransferSimultaneously)
        {
          if (hasWTChgd[mr])
          {
            mr.ForecastWaterTransfer();
            hasWTChgd[mr] = false;
          }
        }
      }
    }

    /// <summary>状態を確定させる</summary>
    public void FixState()
    {
      foreach (MultiRooms mr in mRooms)
      {
        mr.FixHeatTransfer();
        mr.FixWaterTransfer();
        hasHTChgd[mr] = hasWTChgd[mr] = true;
      }
      //壁の熱流を更新
      foreach (Wall wl in walls)
      {
        wl.invMatrixUpdated = false;
        wl.Update();
      }
      isFirstForecast = true;
    }

    /// <summary>ゾーンの温湿度を将来予測値から現在値に戻す</summary>
    public void ResetAirState()
    {
      foreach (MultiRooms mr in mRooms)
      {
        mr.ResetAirState();
        hasHTChgd[mr] = hasWTChgd[mr] = true;
      }
    }

    /// <summary>ゾーン間換気条件を設定する</summary>
    private void setInterZoneAirFlow()
    {
      for (int i = 0; i < zoneVent.Length; i++)
      {
        for (int j = 0; j < zoneVent[i].Length; j++)
        {
          double afsum, tsa, wsa;
          afsum = tsa = wsa = 0;
          foreach (interZoneAirFlow zaf in zoneVent[i][j])
          {
            afsum += zaf.aFlow;
            ImmutableZone zn = mRooms[zaf.rmIndex].Zones[zaf.znIndex];
            tsa += zn.Temperature * zaf.aFlow;
            wsa += zn.HumidityRatio * zaf.aFlow;
          }
          if (afsum != 0)
          {
            tsa /= afsum;
            wsa /= afsum;
          }
          mRooms[i].setSupplyAir(j, tsa, wsa, afsum);
        }
      }
    }

    #endregion

    #region 境界条件設定処理

    /// <summary>壁とゾーンの温湿度を初期化する</summary>
    /// <param name="temperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    public void InitializeAirState(double temperature, double humidityRatio)
    { foreach (MultiRooms ml in mRooms) ml.InitializeAirState(temperature, humidityRatio); }

    /// <summary>外気条件を更新する</summary>
    /// <param name="dTime">現在の日時</param>
    /// <param name="sun">太陽</param>
    /// <param name="temperature">外気乾球温度[C]</param>
    /// <param name="humidityRatio">外気絶対湿度[kg/kg]</param>
    /// <param name="nocRadiation">夜間放射[W/m2]</param>
    public void UpdateOutdoorCondition
      (DateTime dTime, ImmutableSun sun, double temperature, double humidityRatio, double nocRadiation)
    {
      CurrentDateTime = dTime;
      this.Sun = sun;
      OutdoorTemperature = temperature;
      OutdoorHumidityRatio = humidityRatio;
      NocturnalRadiation = nocRadiation;
      foreach (MultiRooms mr in mRooms)
      {
        mr.UpdateOutdoorCondition(dTime, sun, temperature, humidityRatio, nocRadiation);
        hasHTChgd[mr] = hasWTChgd[mr] = true;
      }
    }

    /// <summary>地中温度[C]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="groundTemperature">地中温度[C]</param>
    public void SetGroundTemperature(int mRoomIndex, int wallIndex, bool isSideF, double groundTemperature)
    {
      mRooms[mRoomIndex].SetGroundTemperature(wallIndex, isSideF, groundTemperature);
      hasHTChgd[mRooms[mRoomIndex]] = hasWTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>ゾーン間換気量[kg/s]を設定する</summary>
    /// <param name="rmIndex">多数室番号</param>
    /// <param name="znIndex1">ゾーン番号1</param>
    /// <param name="znIndex2">ゾーン番号2</param>
    /// <param name="airFlowRate">ゾーン間換気量[kg/s]</param>
    public void SetCrossVentilation(int rmIndex, int znIndex1, int znIndex2, double airFlowRate)
    {
      mRooms[rmIndex].SetCrossVentilation (znIndex1, znIndex2, airFlowRate);
      hasHTChgd[mRooms[rmIndex]] = hasWTChgd[mRooms[rmIndex]] = true;
    }

    /// <summary>ゾーン間換気量[kg/s]を設定する</summary>
    /// <param name="rmIndex1">多数室番号1</param>
    /// <param name="znIndex1">ゾーン番号1</param>
    /// <param name="rmIndex2">多数室番号2</param>
    /// <param name="znIndex2">ゾーン番号2</param>
    /// <param name="airFlowRate">ゾーン間換気量[kg/s]</param>
    public void SetCrossVentilation(int rmIndex1, int znIndex1, int rmIndex2, int znIndex2, double airFlowRate)
    {
      //同一の多数室の場合には連成計算
      if (rmIndex1 == rmIndex2) SetCrossVentilation(rmIndex1, znIndex1, znIndex2, airFlowRate);
      //他の多数室の場合には境界条件として計算
      else
      {
        zoneVent[rmIndex2][znIndex2].AddAirFlow(rmIndex1, znIndex1, airFlowRate);
        zoneVent[rmIndex1][znIndex1].AddAirFlow(rmIndex2, znIndex2, airFlowRate);
      }
    }

    /// <summary>ゾーン1からゾーン2へ向かう換気量[kg/s]を設定する</summary>
    /// <param name="rmIndex1">多数室番号1</param>
    /// <param name="znIndex1">ゾーン番号1</param>
    /// <param name="rmIndex2">多数室番号2</param>
    /// <param name="znIndex2">ゾーン番号2</param>
    /// <param name="airFlowRate">換気量[kg/s]</param>
    public void SetAirFlow(int rmIndex1, int znIndex1, int rmIndex2, int znIndex2, double airFlowRate)
    {
      //同一の多数室の場合には連成計算
      if (rmIndex1 == rmIndex2) SetAirFlow(rmIndex1, znIndex1, znIndex2, airFlowRate);
      //他の多数室の場合には境界条件として計算
      else zoneVent[rmIndex2][znIndex2].AddAirFlow(rmIndex1, znIndex1, airFlowRate);
    }

    /// <summary>ゾーン1からゾーン2へ向かう換気量[kg/s]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex1">ゾーン番号1</param>
    /// <param name="zoneIndex2">ゾーン番号2</param>
    /// <param name="airFlowRate">換気量[kg/s]</param>
    /// <remarks>他のゾーンを含め風量収支を合わせる必要あり</remarks>
    public void SetAirFlow(int mRoomIndex, int zoneIndex1, int zoneIndex2, double airFlowRate)
    {
      mRooms[mRoomIndex].SetAirFlow(zoneIndex1, zoneIndex2, airFlowRate);
      hasHTChgd[mRooms[mRoomIndex]] = hasWTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>送水条件を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="wallIndex">壁（床）番号</param>
    /// <param name="mNumber">質点番号</param>
    /// <param name="flowRate">通水量[kg/s]</param>
    /// <param name="temperature">水温[C]</param>
    public void SetBuriedPipeWaterState
      (int mRoomIndex, int wallIndex, int mNumber, double flowRate, double temperature)
    {
      mRooms[mRoomIndex].SetBuriedPipeWaterState(wallIndex, mNumber, flowRate, temperature);
      hasHTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>給気条件を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="saTemperature">給気温度[C]</param>
    /// <param name="saHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="saFlowRate">給気量[kg/s]</param>
    public void SetSupplyAir
      (int mRoomIndex, int zoneIndex, double saTemperature, double saHumidityRatio, double saFlowRate)
    {
      mRooms[mRoomIndex].SetSupplyAir(zoneIndex, saTemperature, saHumidityRatio, saFlowRate);
      hasHTChgd[mRooms[mRoomIndex]] = hasWTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>換気量[kg/s]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="ventilationRate">換気量[kg/s]</param>
    public void SetVentilationRate(int mRoomIndex, int zoneIndex, double ventilationRate)
    { mRooms[mRoomIndex].SetVentilationRate(zoneIndex, ventilationRate); }

    /// <summary>発熱要素を追加する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatGain">発熱要素</param>
    public void AddHeatGain(int mRoomIndex, int zoneIndex, IHeatGain heatGain)
    { mRooms[mRoomIndex].AddHeatGain(zoneIndex, heatGain); }

    /// <summary>発熱要素を削除する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatGain">発熱要素</param>
    public void RemoveHeatGain(int mRoomIndex, int zoneIndex, IHeatGain heatGain)
    { mRooms[mRoomIndex].RemoveHeatGain(zoneIndex, heatGain); }

    #endregion

    #region 制御関連の処理

    /// <summary>乾球温度[C]を制御する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="setpoint">乾球温度設定値[C]</param>
    public void ControlDrybulbTemperature(int mRoomIndex, int zoneIndex, double setpoint)
    {
      mRooms[mRoomIndex].ControlDrybulbTemperature(zoneIndex, setpoint);
      hasHTChgd[mRooms[mRoomIndex]] = true;
      if (mRooms[mRoomIndex].SolveMoistureTransferSimultaneously) hasWTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>顕熱供給[W]を制御する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatSupply">顕熱供給[W]</param>
    public void ControlHeatSupply(int mRoomIndex, int zoneIndex, double heatSupply)
    {
      mRooms[mRoomIndex].ControlHeatSupply(zoneIndex, heatSupply);
      hasHTChgd[mRooms[mRoomIndex]] = true;
      if (mRooms[mRoomIndex].SolveMoistureTransferSimultaneously) hasWTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>絶対湿度[kg/kg]を制御する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="setpoint">絶対湿度設定値[kg/kg]</param>
    public void ControlHumidityRatio(int mRoomIndex, int zoneIndex, double setpoint)
    {
      mRooms[mRoomIndex].ControlHumidityRatio(zoneIndex, setpoint);
      hasWTChgd[mRooms[mRoomIndex]] = true;
      if (mRooms[mRoomIndex].SolveMoistureTransferSimultaneously) hasHTChgd[mRooms[mRoomIndex]] = true;
    }

    /// <summary>水分供給[kg/s]を制御する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="waterSupply">水分供給[kg/s]</param>
    public void ControlWaterSupply(int mRoomIndex, int zoneIndex, double waterSupply)
    {
      mRooms[mRoomIndex].ControlWaterSupply(zoneIndex, waterSupply);
      hasWTChgd[mRooms[mRoomIndex]] = true;
      if (mRooms[mRoomIndex].SolveMoistureTransferSimultaneously) hasHTChgd[mRooms[mRoomIndex]] = true;
    }

    #endregion

    #region 空調能力設定関連の処理

    /// <summary>加熱能力[W]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatingCapacity">加熱能力[W]</param>
    public void SetHeatingCapacity(int mRoomIndex, int zoneIndex, double heatingCapacity)
    { mRooms[mRoomIndex].SetHeatingCapacity(zoneIndex, heatingCapacity); }

    /// <summary>冷却能力[W]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="coolingCapacity">冷却能力[W]</param>
    public void SetCoolingCapacity(int mRoomIndex, int zoneIndex, double coolingCapacity)
    { mRooms[mRoomIndex].SetCoolingCapacity(zoneIndex, coolingCapacity); }

    /// <summary>加湿能力[kg/s]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="humidifyingCapacity">加湿能力[kg/s]</param>
    public void SetHumidifyingCapacity(int mRoomIndex, int zoneIndex, double humidifyingCapacity)
    { mRooms[mRoomIndex].SetHumidifyingCapacity(zoneIndex, humidifyingCapacity); }

    /// <summary>除湿能力[kg/s]を設定する</summary>
    /// <param name="mRoomIndex">多数室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="dehumidifyingCapacity">除湿能力[kg/s]</param>
    public void SetDehumidifyingCapacity(int mRoomIndex, int zoneIndex, double dehumidifyingCapacity)
    { mRooms[mRoomIndex].SetDehumidifyingCapacity(zoneIndex, dehumidifyingCapacity); }

    #endregion

    #region インナークラス定義

    /// <summary>ゾーン間換気情報集合クラス</summary>
    [Serializable]
    private class interZoneAirFlowCollection : IEnumerable<interZoneAirFlow>
    {

      /// <summary>rmIndex番の多数室のznIndex番のゾーンからの流入空気量[kg/s]</summary>
      private Dictionary<int, Dictionary<int, double>> aFlows 
        = new Dictionary<int, Dictionary<int, double>>();

      /// <summary>流入空気を追加する</summary>
      /// <param name="rmIndex">多数室番号</param>
      /// <param name="znIndex">ゾーン番号</param>
      /// <param name="aFlow">空気流入量[kg/s]</param>
      public void AddAirFlow(int rmIndex, int znIndex, double aFlow)
      {
        if (!aFlows.ContainsKey(rmIndex))
          aFlows.Add(rmIndex, new Dictionary<int, double>());
        aFlows[rmIndex][znIndex] = aFlow;
      }

      /// <summary>流入空気を解除する</summary>
      /// <param name="rmIndex">多数室番号</param>
      /// <param name="znIndex">ゾーン番号</param>
      public void RemoveAirFlow(int rmIndex, int znIndex)
      {
        if (aFlows.ContainsKey(rmIndex))
        {
          if (aFlows[rmIndex].ContainsKey(znIndex)) aFlows[rmIndex].Remove(znIndex);
          if (aFlows[rmIndex].Count == 0) aFlows.Remove(rmIndex);
        }
      }

      /// <summary>換気情報を出力する</summary>
      /// <returns>換気情報</returns>
      public IEnumerator<interZoneAirFlow> GetEnumerator()
      {
        foreach (int key1 in aFlows.Keys)
          foreach (int key2 in aFlows[key1].Keys)
            yield return new interZoneAirFlow(key1, key2, aFlows[key1][key2]);
      }

      IEnumerator IEnumerable.GetEnumerator()
      { throw new NotImplementedException(); }
    }

    /// <summary>ゾーン間換気情報クラス</summary>
    private class interZoneAirFlow
    {
      /// <summary>室番号を設定・取得する</summary>
      public int rmIndex { get; set; }

      /// <summary>ゾーン番号を設定・取得する</summary>
      public int znIndex { get; set; }

      /// <summary>換気量[kg/s]を設定・取得する</summary>
      public double aFlow { get; set; }

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="rmIndex">室番号</param>
      /// <param name="znIndex">ゾーン番号</param>
      /// <param name="aFlow">換気量</param>
      public interZoneAirFlow(int rmIndex, int znIndex, double aFlow)
      {
        this.rmIndex = rmIndex;
        this.znIndex = znIndex;
        this.aFlow = aFlow;
      }
    }

    #endregion

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用の建物熱負荷計算モデル</summary>
  public interface ImmutableBuildingThermalModel
  {
    /// <summary>多数室を取得する</summary>
    ImmutableMultiRooms[] MultiRoom { get; }

    /// <summary>現在の日時を取得する</summary>
    DateTime CurrentDateTime { get; }

    /// <summary>計算時間間隔[sec]を取得する</summary>
    double TimeStep { get; }

    /// <summary>太陽を取得する</summary>
    ImmutableSun Sun { get; }

    /// <summary>外気乾球温度[C]を取得する</summary>
    double OutdoorTemperature { get; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    double OutdoorHumidityRatio { get; }

    /// <summary>夜間放射[W/m2]を取得する</summary>
    double NocturnalRadiation { get; }
  }

  #endregion

}
