﻿/* BuriedPipe.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.ThermalLoad
{
  /// <summary>埋設配管</summary>
  [Serializable]
  public class BuriedPipe
  {

    #region 定数宣言

    /// <summary>水の密度[kg/m3]</summary>
    private const double WATER_DENSITY = 1000;

    /// <summary>水の比熱[J/kg]</summary>
    private const double WATER_SPECIFICHEAT = 4186;

    #endregion

    #region プロパティ

    /// <summary>水温[C]を取得する</summary>
    /// <remarks>配管内対流熱伝達率の計算に使用</remarks>
    public double InletWaterTemperature { get; private set; } = 25;

    /// <summary>水量[kg/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>分岐の数を取得する</summary>
    public int BranchNumber { get; private set; }

    /// <summary>配管敷設ピッチ[m]を取得する</summary>
    public double Pitch { get; private set; }

    /// <summary>配管総延長[m]を取得する</summary>
    public double Length { get; private set; }

    /// <summary>内径[m]を取得する</summary>
    public double InnerDiameter { get; private set; }

    /// <summary>外径[m]を取得する</summary>
    public double OuterDiameter { get; private set; }

    /// <summary>配管材の熱伝導率[W/(mK)]を取得する</summary>
    public double ThermalConductivityOfTube { get; private set; }

    /// <summary>上側フィン効率[-]を取得する</summary>
    public double UpperFinEfficiency { get; private set; }

    /// <summary>下側フィン効率[-]を取得する</summary>
    public double LowerFinEfficiency { get; private set; }

    /// <summary>熱通過有効度[-]を取得する</summary>
    public double Effectiveness { get; private set; }

    #endregion

    #region インスタンスメソッド

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="pitch">配管敷設ピッチ[m]</param>
    /// <param name="length">総配管長[m]</param>
    /// <param name="branchNumber">分岐の数[本]</param>
    /// <param name="iDiameter">内径[m]</param>
    /// <param name="oDiameter">外径[m]</param>
    /// <param name="tubeConductivity">配管熱伝導率[W/(mK)]</param>
    /// <param name="upperFinConductivity">上側フィン熱伝導率[W/mK]</param>
    /// <param name="lowerFinConductivity">下側フィン熱伝導率[W/mK]</param>
    /// <param name="upperFinResistance">上側フィン熱抵抗[(m2K)/W]</param>
    /// <param name="lowerFinResistance">下側フィン熱抵抗[(m2K)/W]</param>
    /// <param name="upperFinThickness">上側フィン厚み[m]</param>
    /// <param name="lowerFinThickness">下側フィン厚み[m]</param>
    public BuriedPipe(double pitch, double length, int branchNumber, double iDiameter, double oDiameter,
      double tubeConductivity, double upperFinConductivity, double lowerFinConductivity,
      double upperFinResistance, double lowerFinResistance, double upperFinThickness, double lowerFinThickness)
    {
      Pitch = pitch;
      Length = length;
      BranchNumber = branchNumber;
      InnerDiameter = iDiameter;
      OuterDiameter = oDiameter;
      ThermalConductivityOfTube = tubeConductivity;

      double bf1 = (pitch - oDiameter) / 2;
      double bf2 = oDiameter + (pitch - oDiameter);
      double ubL = bf1 * Math.Sqrt(1 / (upperFinResistance * upperFinConductivity * upperFinThickness));
      double ubR = bf1 * Math.Sqrt(1 / (lowerFinResistance * lowerFinConductivity * lowerFinThickness));
      UpperFinEfficiency = 1 / pitch * (bf2 * Math.Tanh(ubL) / ubL);
      LowerFinEfficiency = 1 / pitch * (bf2 * Math.Tanh(ubR) / ubR);
    }

    /// <summary>体積流量[kg/s]を設定する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    public void SetFlowRate(double flowRate)
    {
      WaterFlowRate = flowRate;
      updateEffectiveness();
    }

    /// <summary>水温[C]を設定する</summary>
    /// <param name="temperature">水温[C]</param>
    public void SetWaterTemperature(double temperature)
    {
      InletWaterTemperature = temperature;
      updateEffectiveness();
    }

    /// <summary>熱通過率[-]を更新する</summary>
    private void updateEffectiveness()
    {
      if (WaterFlowRate <= 0)
      {
        Effectiveness = 0;
        return;
      }

      //対流熱伝達率[W/m2K]の計算//////////////////////////
      //動粘性係数[m2/s]・熱拡散率[m2/s]・熱伝導率[W/(m·K)]を計算
      double v = Water.GetLiquidDynamicViscosity(InletWaterTemperature);
      double a = Water.GetLiquidThermalDiffusivity(InletWaterTemperature);
      double lambda = Water.GetLiquidThermalConductivity(InletWaterTemperature);

      //配管内流速[m/s]を計算
      double vFlow = WaterFlowRate / (WATER_DENSITY * BranchNumber);
      double u = vFlow / (Math.Pow(InnerDiameter / 2, 2) * Math.PI);

      //ヌセルト数を計算
      double reNumber = u * InnerDiameter / v;
      double prNumber = v / a;
      double nuNumber = 0.023 * Math.Pow(reNumber, 0.8) * Math.Pow(prNumber, 0.4);

      //ヌセルト数から対流熱伝達率を計算
      double hi = nuNumber * lambda / InnerDiameter;

      //伝熱係数KA, 移動単位数NTU, 熱通過率εの計算/////////
      double ka = 1 / (InnerDiameter * hi) 
        + 1 / (2 * ThermalConductivityOfTube) * Math.Log(OuterDiameter / InnerDiameter);
      ka = (Math.PI * Length) / ka;
      double ntu = ka / (WaterFlowRate * WATER_SPECIFICHEAT);
      Effectiveness = 1 - Math.Exp(-ntu);
    }

    #endregion

  }
}
