﻿/* Zone.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections.Generic;

namespace Popolo.ThermalLoad
{
  /// <summary>1質点を持つ空間</summary>
  [Serializable]
  public class Zone : ImmutableZone
  {

    #region インスタンス変数・プロパティ

    /// <summary>発熱要素リスト</summary>
    private List<IHeatGain> heatGains = new List<IHeatGain>();

    /// <summary>名称を設定・取得する</summary>
    public string Name { get; private set; }

    /// <summary>換気量[kg/s]を設定・取得する</summary>
    public double VentilationRate { get; set; }

    /// <summary>給気量[kg/s]を設定・取得する</summary>
    public double SupplyAirFlowRate { get; set; }

    /// <summary>空気の質量[kg]を取得する</summary>
    public double AirMass { get; private set; }

    /// <summary>床面積[m2]を取得する</summary>
    public double FloorArea { get; private set; }

    /// <summary>所属する室番号を取得する</summary>
    /// <remarks>いずれにも所属しない場合には-1</remarks>
    public int RoomIndex { get; internal set; } = -1;

    /// <summary>ゾーン番号を取得する</summary>
    public int Index { get; internal set; }

    /// <summary>所属する多数室を取得する</summary>
    public ImmutableMultiRooms MultiRoom { get; internal set; }

    /// <summary>空気以外の顕熱容量[J/K]を設定・取得する</summary>
    /// <remarks>
    /// 事務所の場合、室容積あたり12000[J/m3-K]程度
    /// 木村健一, 事務所建築の家具の熱容量,
    /// 日本建築学会関東支部第29会研究発表会, 1961.01
    /// 住宅の場合は事務所の半分程度
    /// </remarks>
    public double HeatCapacity { get; set; }

    /// <summary>空気以外の水分容量[kg]を設定・取得する</summary>
    public double WaterCapacity { get; set; }

    /// <summary>乾球温度[C]を取得する</summary>
    public double Temperature { get; internal set; } = 24;

    /// <summary>乾球温度設定値[C]を取得する</summary>
    public double TemperatureSetpoint { get; private set; } = 24;

    /// <summary>供給顕熱[W]を取得する</summary>
    public double HeatSupply { get; internal set; }

    /// <summary>乾球温度[C]を制御するか否か</summary>
    public bool TemperatureControlled { get; internal set; }

    /// <summary>給気温度[C]を設定・取得する</summary>
    public double SupplyAirTemperature { get; set; }

    /// <summary>絶対湿度[kg/kg]を取得する</summary>
    public double HumidityRatio { get; internal set; } = 0.0095;

    /// <summary>絶対湿度設定値[kg/kg]を取得する</summary>
    public double HumidityRatioSetpoint { get; private set; } = 0.0095;

    /// <summary>供給水分[kg/s]を取得する</summary>
    public double WaterSupply { get; internal set; }

    /// <summary>絶対湿度[kg/kg]を制御するか否か</summary>
    public bool HumidityControlled { get; internal set; }

    /// <summary>給気絶対湿度[kg/kg]を設定・取得する</summary>
    public double SupplyAirHumidityRatio { get; set; }

    /// <summary>加熱能力[W]を設定・取得する</summary>
    public double HeatingCapacity { get; set; } = double.PositiveInfinity;

    /// <summary>冷却能力[W]を設定・取得する</summary>
    public double CoolingCapacity { get; set; } = double.PositiveInfinity;

    /// <summary>加湿能力[kg/s]を設定・取得する</summary>
    public double HumidifyingCapacity { get; set; } = double.PositiveInfinity;

    /// <summary>除湿能力[kg/s]を設定・取得する</summary>
    public double DehumidifyingCapacity { get; set; } = double.PositiveInfinity;

    #endregion

    #region internalプロパティ

    /// <summary>ゾーンに面する壁・窓表面リスト</summary>
    internal List<wallWindowSurface> Surfaces { get; set; }

    /// <summary>他のゾーンからの給気量[kg/s]を設定・取得する</summary>
    internal double supplyAirFlowRate2 { get; set; }

    /// <summary>他のゾーンからの給気温度[C]を設定・取得する</summary>
    internal double supplyAirTemperature2 { get; set; }

    /// <summary>他のゾーンからの給気絶対湿度[kg/kg]を設定・取得する</summary>
    internal double supplyAirHumidityRatio2 { get; set; }

    #endregion

    #region コンストラクタ

    /// <summary>新しいインスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="airMass">空気の質量[kg]</param>
    public Zone(string name, double airMass)
    {
      Name = name;
      AirMass = airMass;
      Surfaces = new List<wallWindowSurface>();
    }

    /// <summary>新しいインスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="airMass">空気の質量[kg]</param>
    /// <param name="floorArea">床面積[m2]</param>
    public Zone(string name, double airMass, double floorArea)
    {
      Name = name;
      AirMass = airMass;
      FloorArea = floorArea;
      Surfaces = new List<wallWindowSurface>();
    }

    #endregion

    #region 温湿度設定・制御関連の処理

    /// <summary>温湿度を初期化する</summary>
    /// <param name="temperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    public void InitializeAirState(double temperature, double humidityRatio)
    {
      Temperature = temperature;
      HumidityRatio = humidityRatio;
    }

    /// <summary>乾球温度[C]を制御する</summary>
    /// <param name="setpoint">乾球温度設定値[C]</param>
    public void ControlDrybulbTemperature(double setpoint)
    {
      TemperatureControlled = true;
      TemperatureSetpoint = setpoint;
    }

    /// <summary>顕熱供給[W]を制御する</summary>
    /// <param name="heatSupply">顕熱供給[W]</param>
    public void ControlHeatSupply(double heatSupply)
    {
      TemperatureControlled = false;
      HeatSupply = heatSupply;
    }

    /// <summary>絶対湿度[kg/kg]を制御する</summary>
    /// <param name="setpoint">絶対湿度設定値[kg/kg]</param>
    public void ControlHumidityRatio(double setpoint)
    {
      HumidityControlled = true;
      HumidityRatioSetpoint = setpoint;
    }

    /// <summary>水分供給[kg/s]を制御する</summary>
    /// <param name="waterSupply">水分供給[kg/s]</param>
    public void ControlWaterSupply(double waterSupply)
    {
      HumidityControlled = false;
      WaterSupply = waterSupply;
    }

    /// <summary>平均表面温度を計算する</summary>
    /// <returns>平均表面温度[C]</returns>
    public double GetMeanSurfaceTemperature()
    {
      double tSum = 0;
      double sSum = 0;
      foreach (wallWindowSurface ws in Surfaces)
      {
        tSum += ws.SurfaceTemperature * ws.Area;
        sSum += ws.Area;
      }
      return tSum / sSum;
    }

    #endregion

    #region 発熱関連の処理

    /// <summary>発熱要素を追加する</summary>
    /// <param name="heatGain">発熱要素</param>
    public void AddHeatGain(IHeatGain heatGain) { heatGains.Add(heatGain); }

    /// <summary>発熱要素を削除する</summary>
    /// <param name="heatGain">発熱要素</param>
    public void RemoveHeatGain(IHeatGain heatGain) { heatGains.Remove(heatGain); }

    /// <summary>顕熱取得の対流成分[W]を積算する</summary>
    /// <returns>顕熱取得の対流成分[W]</returns>
    public double IntegrateConvectiveHeatgains()
    {
      double sum = 0;
      foreach (IHeatGain hg in heatGains) sum += hg.GetConvectiveHeatGain(this);
      return sum;
    }

    /// <summary>顕熱取得の放射成分[W]を積算する</summary>
    /// <returns>顕熱取得の放射成分[W]</returns>
    public double IntegrateRadiativeHeatGains()
    {
      double sum = 0;
      foreach (IHeatGain hg in heatGains) sum += hg.GetRadiativeHeatGain(this);
      return sum;
    }

    /// <summary>発生水分[kg/s]を積算する</summary>
    /// <returns>発生水分[kg/s]</returns>
    public double IntegrateWaterGains()
    {
      double sum = 0;
      foreach (IHeatGain hg in heatGains) sum += hg.GetWaterGain(this);
      return sum;
    }

    #endregion

  }

  #region 読み取り専用ゾーン

  /// <summary>読み取り専用Zone</summary>
  public interface ImmutableZone
  {
    /// <summary>名称を取得する</summary>
    string Name { get; }

    /// <summary>換気量[kg/s]を取得する</summary>
    double VentilationRate { get; }

    /// <summary>給気量[kg/s]を取得する</summary>
    double SupplyAirFlowRate { get; }

    /// <summary>空気の質量[kg]を取得する</summary>
    double AirMass { get; }

    /// <summary>床面積[m2]を取得する</summary>
    double FloorArea { get; }

    /// <summary>所属する室番号を取得する</summary>
    int RoomIndex { get; }

    /// <summary>所属する多数室を取得する</summary>
    ImmutableMultiRooms MultiRoom { get; }

    /// <summary>空気以外の顕熱容量[J/K]を取得する</summary>
    double HeatCapacity { get; }

    /// <summary>空気以外の水分容量[kg]を取得する</summary>
    double WaterCapacity { get; }

    /// <summary>乾球温度[C]を取得する</summary>
    double Temperature { get; }

    /// <summary>供給顕熱[W]を取得する</summary>
    double HeatSupply { get; }

    /// <summary>乾球温度[C]を制御するか否か</summary>
    bool TemperatureControlled { get; }

    /// <summary>給気温度[C]を取得する</summary>
    double SupplyAirTemperature { get; }

    /// <summary>絶対湿度[kg/kg]を取得する</summary>
    double HumidityRatio { get; }

    /// <summary>供給水分[kg/s]を取得する</summary>
    double WaterSupply { get; }

    /// <summary>絶対湿度[kg/kg]を制御するか否か</summary>
    bool HumidityControlled { get; }

    /// <summary>給気絶対湿度[kg/kg]を取得する</summary>
    double SupplyAirHumidityRatio { get; }

    /// <summary>顕熱取得の対流成分[kW]を積算する</summary>
    /// <returns>顕熱取得の対流成分[kW]</returns>
    double IntegrateConvectiveHeatgains();

    /// <summary>顕熱取得の放射成分[kW]を積算する</summary>
    /// <returns>顕熱取得の放射成分[kW]</returns>
    double IntegrateRadiativeHeatGains();

    /// <summary>発生水分[kg/s]を積算する</summary>
    /// <returns>発生水分[kg/s]</returns>
    double IntegrateWaterGains();

    /// <summary>平均表面温度を計算する</summary>
    /// <returns>平均表面温度[C]</returns>
    double GetMeanSurfaceTemperature();

    /// <summary>加熱能力[W]を取得する</summary>
    double HeatingCapacity { get; }

    /// <summary>冷却能力[W]を取得する</summary>
    double CoolingCapacity { get; }

    /// <summary>加湿能力[kg/s]を取得する</summary>
    double HumidifyingCapacity { get; }

    /// <summary>除湿能力[kg/s]を取得する</summary>
    double DehumidifyingCapacity { get; }
  }

  #endregion
  
}
