﻿/* LithiumBromide.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.ThermophysicalProperty
{
  /// <summary>臭化リチウム</summary>
  [Serializable]
  public class LithiumBromide
  {

    #region プロパティ

    /// <summary>比エンタルピーを取得する</summary>
    public double Enthalpy { get; private set; }

    /// <summary>水溶液の温度[K]を取得する</summary>
    public double LiquidTemperature { get; private set; }

    /// <summary>飽和蒸気の温度[K]を取得する</summary>
    public double VaporTemperature { get; private set; }

    /// <summary>質量分率[-]を取得する</summary>
    public double MassFraction { get; private set; }

    /// <summary>比熱[kJ/(kgK)]を取得する</summary>
    public double SpecificHeat { get; private set; }
    #endregion

    #region static変数

    /// <summary>比エンタルピー近似係数</summary>
    readonly static double[] C_LBN = 
      new double[] { -2024.33, 16330.9, -48816.1, 6.302948e4, -2.913705e4 };
    readonly static double[] D_LBN = 
      new double[] { 18.2829, -116.91757, 3.248041e2, -4.034184e2, 1.8520569e2 };
    readonly static double[] E_LBN =
      new double[] { -3.7008214e-2, 2.8877666e-1, -8.1313015e-1, 9.9116628e-1, -4.4441207e-1 };

    #endregion

    #region 飽和温度・水溶液温度・質量分率の計算

    /// <summary>溶液温度と質量分率から飽和温度を計算する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>平衡蒸気の飽和温度[K]</returns>
    public static double GetVaporTemperatureFromLiquidTemperatureAndMassFraction
      (double liquidTemperature, double massFraction)
    {
      double alb, blb;
      getCoefficient(massFraction, out alb, out blb);
      return alb + blb * liquidTemperature;
    }

    /// <summary>飽和温度と質量分率から溶液温度を計算する</summary>
    /// <param name="vaporTemperature">水蒸気の飽和温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>水溶液の温度[K]</returns>
    public static double GetLiquidTemperatureFromVaporTemperatureAndMassFraction
      (double vaporTemperature, double massFraction)
    {
      double alb, blb;
      getCoefficient(massFraction, out alb, out blb);
      return (vaporTemperature - alb) / blb;
    }

    /// <summary>溶液温度と飽和温度から質量分率を計算する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="vaporTemperature">平衡蒸気の飽和温度[K]</param>
    /// <returns>水溶液の質量分率[-]</returns>
    public static double GetMassFractionFromLiquidTemperatureAndVaporTemperature
      (double liquidTemperature, double vaporTemperature)
    {
      Roots.ErrorFunction eFnc = delegate (double mf)
      {
        return vaporTemperature - GetVaporTemperatureFromLiquidTemperatureAndMassFraction(liquidTemperature, mf);
      };
      return Roots.Newton(eFnc, 0.5, 0.001, 0.00001, 0.00001, 20);
    }

    /// <summary>物性値近似係数を計算する</summary>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <param name="ca">出力：係数a</param>
    /// <param name="cb">出力：係数b</param>
    private static void getCoefficient(double massFraction, out double ca, out double cb)
    {
      const double a0 = -22.8937; const double a1 = 152.554;
      const double a2 = -254.786; const double a3 = 152.949;
      const double a4 = -171.599;
      const double b0 = 1.09851;  const double b1 = -0.394508;

      double mf = Math.Min(1, Math.Max(0, massFraction));
      ca = a0 + mf * (a1 + mf * (a2 + mf * (a3 + mf * mf * mf * mf * a4)));
      cb = b0 + mf * b1;
    }

    #endregion

    #region 比エンタルピー関連の計算

    /// <summary>溶液温度と質量分率から比エンタルピーを計算する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>水溶液の比エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromLiquidTemperatureAndMassFraction
      (double liquidTemperature, double massFraction)
    {
      double clb, dlb, elb;
      getCoefficient(massFraction, out clb, out dlb, out elb);
      double ltk = liquidTemperature - 273.15;
      return clb + ltk * (dlb + ltk * elb);
    }

    /// <summary>溶液温度と飽和温度から比エンタルピーを計算する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="vaporTemperature">平衡蒸気の温度[K]</param>
    /// <returns>水溶液の比エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromLiquidTemperatureAndVaporTemperature
      (double liquidTemperature, double vaporTemperature)
    {
      double mf = GetMassFractionFromLiquidTemperatureAndVaporTemperature(liquidTemperature, vaporTemperature);
      return GetEnthalpyFromLiquidTemperatureAndMassFraction(liquidTemperature, mf);
    }

    /// <summary>比エンタルピーと質量分率から溶液温度を計算する</summary>
    /// <param name="enthalpy">水溶液の比エンタルピー[kJ/kg]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>水溶液の温度[K]</returns>
    public static double GetLiquidTemperatureFromEnthalpyAndMassFraction(double enthalpy, double massFraction)
    {
      double clb, dlb, elb;
      getCoefficient(massFraction, out clb, out dlb, out elb);

      //ニュートン法で収束計算
      Roots.ErrorFunction eFnc = delegate (double lTemp)
      {
        return enthalpy - (clb + lTemp * (dlb + lTemp * elb));
      };
      return Roots.Newton(eFnc, 40, 0.001, 0.00001, 0.00001, 20) + 273.15;
    }

    /// <summary>比エンタルピーと飽和温度から溶液温度を計算する</summary>
    /// <param name="enthalpy">水溶液の比エンタルピー[kJ/kg]</param>
    /// <param name="vaporTemperature">平衡蒸気の飽和温度[K]</param>
    /// <returns>水溶液の温度[K]</returns>
    public static double GetMassFractionFromEnthalpyAndVaporTemperature
      (double enthalpy, double vaporTemperature)
    {
      Roots.ErrorFunction eFnc = delegate (double mf)
      {
        return GetLiquidTemperatureFromEnthalpyAndMassFraction(enthalpy, mf)
        - GetLiquidTemperatureFromVaporTemperatureAndMassFraction(vaporTemperature, mf);
      };
      return Roots.Newton(eFnc, 0.5, 0.001, 0.00001, 0.00001, 20);
    }

    /// <summary>物性値近似係数を計算する</summary>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <param name="cc">出力：係数a</param>
    /// <param name="cd">出力：係数b</param>
    /// <param name="ce">出力：係数b</param>
    private static void getCoefficient(double massFraction, out double cc, out double cd, out double ce)
    {
      massFraction = Math.Min(1.0, massFraction);
      cc = C_LBN[4];
      cd = D_LBN[4];
      ce = E_LBN[4];
      for (int i = 3; 0 <= i; i--)
      {
        cc = cc * massFraction + C_LBN[i];
        cd = cd * massFraction + D_LBN[i];
        ce = ce * massFraction + E_LBN[i];
      }
    }

    #endregion

    #region その他の物性の計算

    /// <summary>密度[kg/m3]を計算する</summary>
    /// <param name="liquidTemperature">臭化リチウム水溶液の温度[K]</param>
    /// <param name="massFraction">臭化リチウム水溶液の質量分率[-]</param>
    /// <returns>密度[kg/m3]</returns>
    public static double GetDensity(double liquidTemperature, double massFraction)
    {      
      const double LBD = 3460;  //臭化リチウムの密度[kg/m3]
      double wd = Water.GetLiquidDensity(liquidTemperature);
      return LBD * massFraction + wd * (1 - massFraction);
    }

    /// <summary>溶液温度と質量分率から比熱を計算する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>水溶液の比熱[kJ/(kgK)]</returns>
    public static double GetSpecificHeat(double liquidTemperature, double massFraction)
    {
      double dlb = D_LBN[4];
      double elb = E_LBN[4];
      for (int i = 3; 0 <= i; i--)
      {
        dlb = dlb * massFraction + D_LBN[i];
        elb = elb * massFraction + E_LBN[i];
      }
      return dlb + 2 * elb * (liquidTemperature - 273.15);
    }

    #endregion

    #region staticメソッド（インスタンス生成）

    /// <summary>溶液温度と質量分率からインスタンスを生成する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>臭化リチウム水溶液のインスタンス</returns>
    public static LithiumBromide MakeFromLiquidTemperatureAndMassFraction
      (double liquidTemperature, double massFraction)
    {
      LithiumBromide lb = new LithiumBromide();
      lb.LiquidTemperature = liquidTemperature;
      lb.MassFraction = massFraction;
      lb.VaporTemperature = GetVaporTemperatureFromLiquidTemperatureAndMassFraction
        (liquidTemperature, massFraction);
      lb.Enthalpy = GetEnthalpyFromLiquidTemperatureAndMassFraction(liquidTemperature, massFraction);
      lb.SpecificHeat = GetSpecificHeat(liquidTemperature, massFraction);
      return lb;
    }

    /// <summary>飽和温度と質量分率からインスタンスを生成する</summary>
    /// <param name="vaporTemperature">平衡蒸気の飽和温度[K]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>臭化リチウム水溶液のインスタンス</returns>
    public static LithiumBromide MakeFromVaporTemperatureAndMassFraction
      (double vaporTemperature, double massFraction)
    {
      LithiumBromide lb = new LithiumBromide();
      lb.VaporTemperature = vaporTemperature;
      lb.MassFraction = massFraction;
      lb.LiquidTemperature = GetLiquidTemperatureFromVaporTemperatureAndMassFraction
        (vaporTemperature, massFraction);
      lb.Enthalpy = GetEnthalpyFromLiquidTemperatureAndMassFraction(lb.LiquidTemperature, massFraction);
      lb.SpecificHeat = GetSpecificHeat(lb.LiquidTemperature, massFraction);
      return lb;
    }

    /// <summary>溶液温度と飽和温度からインスタンスを生成する</summary>
    /// <param name="liquidTemperature">水溶液の温度[K]</param>
    /// <param name="vaporTemperature">平衡蒸気の飽和温度[K]</param>
    /// <returns>臭化リチウム水溶液のインスタンス</returns>
    public static LithiumBromide MakeFromLiquidTemperatureAndVaporTemperature
      (double liquidTemperature, double vaporTemperature)
    {
      double mf = GetMassFractionFromLiquidTemperatureAndVaporTemperature(liquidTemperature, vaporTemperature);
      return MakeFromLiquidTemperatureAndMassFraction(liquidTemperature, mf);
    }

    /// <summary>比エンタルピーと質量分率からインスタンスを生成する</summary>
    /// <param name="enthalpy">水溶液の比エンタルピー[kJ/kg]</param>
    /// <param name="massFraction">水溶液の質量分率[-]</param>
    /// <returns>臭化リチウム水溶液のインスタンス</returns>
    public static LithiumBromide MakeFromEnthalpyAndMassFraction(double enthalpy, double massFraction)
    {
      double lt = GetLiquidTemperatureFromEnthalpyAndMassFraction(enthalpy, massFraction);
      return MakeFromLiquidTemperatureAndMassFraction(lt, massFraction);
    }

    /// <summary>水溶液温度と飽和温度からインスタンスを生成する</summary>
    /// <param name="enthalpy">水溶液の比エンタルピー[kJ/kg]</param>
    /// <param name="vaporTemperature">平衡蒸気の飽和温度[K]</param>
    /// <returns>臭化リチウム水溶液のインスタンス</returns>
    public static LithiumBromide MakeFromEnthalpyAndVaporTemperature(double enthalpy, double vaporTemperature)
    {
      double mf = GetMassFractionFromEnthalpyAndVaporTemperature(enthalpy, vaporTemperature);
      return MakeFromVaporTemperatureAndMassFraction(vaporTemperature, mf);
    }

    #endregion

    #region コンストラクタを隠蔽

    /// <summary>コンストラクタ</summary>
    private LithiumBromide() { }

    #endregion

  }
}
